/* $Id: crires_model_kernel.c,v 1.54 2012-10-15 09:42:22 yjung Exp $
 *
 * This file is part of the CRIRES Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: yjung $
 * $Date: 2012-10-15 09:42:22 $
 * $Revision: 1.54 $
 * $Name: not supported by cvs2svn $
 */
 
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                                                   Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>

#include "crires_model_kernel.h"
#include "crires_model_qmetric.h"
#include "crires_model_io.h"
#include "irplib_utils.h"
#include "crires_load.h"
#include "crires_dfs.h"
#include "crires_utils.h"
#include "crires_pfits.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup crires_model_kernel Functions for the wl model computation
 */
/*----------------------------------------------------------------------------*/

/**@{*/

static const int vectordim=4;

/*----------------------------------------------------------------------------*/
/**
 * Model init sets up the matrix transformations for the current parameter
 * configuration. This does not include any transformations which are dependent
 * upon wavelength or order (refraction, difraction). Keeping this separate
 * from the main eval function speeds up simulation or wavemap generation. Note
 * however that for the annealing process, because many different prism and
 * grating angles can be present in a given set of anneal data, init must be
 * called before every eval.
 *
 * The function takes the p_crires structure containing all model parameters as
 * input.
 * 0 is returned on succes, and the p_crires structure itself is modified.
 */
/*----------------------------------------------------------------------------*/
int crires_model_init(struct crires * p_crires)
{
  int ii;
  double jaw_off;
  double buff;
  double slit_shift=0.0;
  mat ingrat, inaxis;
  mat tras_ingrat;
  mat unity_mat={ {1,0,0,0},
		  {0,1,0,0}, 
		  {0,0,1,0},
		  {0,0,0,1}};
 /*-----------------------------------------------------------------*/
  jaw_off=fix_jaw_off;
  buff=is_buff;

 // p_crires->es_x=0.0;
 // p_crires->es_y=0.0;

 /*******************  Rotation of entrance slit********************/
  crires_model_rotin(p_crires->e_slit,p_crires->mues,p_crires->nues,
		     p_crires->taues);

  crires_model_p_rot(p_crires);

 /*******************  Rotation into into Grating ********************/
 /*First perform an additional rotation that allows for the axis of rotation
   of the grating being offset*/
  crires_model_rotin(inaxis,
		     p_crires->axis_mug,
		     p_crires->axis_nug,
		     p_crires->axis_taug);

 /*define matrix camera (mcam see above)
   define matrix rotation in the detector indtc
   make the transformations:back into reference frame , camera
   reflexion , rotation into detector */
  crires_model_rotin(ingrat,
		     p_crires->mug,
		     p_crires->nug,
		     p_crires->taug);
  crires_model_multiplymatrix(p_crires->tograt,
			      inaxis,
			      ingrat); 

 /*Slit limits */
 /* This is no longer used. Slit rotation should only be done with
    tauis. Below is a check that slitrot=90.0 is always used
p_crires->slitxup=p_crires->slitx+
   (0.5*p_crires->slitw*cos(p_crires->slitrot)+
    0.5*p_crires->slith*sin(p_crires->slitrot));
 p_crires->slitxdown=p_crires->slitx-
   (0.5*p_crires->slitw*cos(p_crires->slitrot)+
    0.5*p_crires->slith*sin(p_crires->slitrot));
 p_crires->slityup=p_crires->slity+
   (0.5*p_crires->slitw*sin(p_crires->slitrot)+
    0.5*p_crires->slith*cos(p_crires->slitrot));
 p_crires->slitydown=p_crires->slity-
   (0.5*p_crires->slitw*sin(p_crires->slitrot)+
    0.5*p_crires->slith*cos(p_crires->slitrot));
 */
  if (p_crires->slitrot!=90.0*DEG2RAD) {
    cpl_msg_error(__func__, "slitrot!=90.0. Slit rotation should only be done with tauis.");
    return(1);
  }
  /*Note slityup is the jaw that defines the LOWER wavelength limit*/
  p_crires->slitxup=p_crires->slitx+0.5*p_crires->slith;
  p_crires->slitxdown=p_crires->slitx-0.5*p_crires->slith;

  p_crires->slitydown=p_crires->slity-jaw_off+slit_shift-buff;//-(buff*p_crires->slitw);
  p_crires->slityup=p_crires->slity-jaw_off+slit_shift+buff+p_crires->slitw;
  /* below would keep other jaw fixed*/
/*   p_crires->slitydown=p_crires->slity+jaw_off+slit_shift-buff-p_crires->slitw;//-(buff*p_crires->slitw); */
/*   p_crires->slityup=p_crires->slity+jaw_off+slit_shift+buff; */


 /*Chip limits */
 for (ii=0; ii < 4; ii++)
 {
   p_crires->chipxup[ii]=
     p_crires->chipx[ii]+p_crires->det_array_x+
     (p_crires->pix*(0.5*chipxpix*cos(p_crires->chiprot[ii])+
                     0.5*chipypix*sin(p_crires->chiprot[ii])+20.0));
   p_crires->chipxdown[ii]=
     p_crires->chipx[ii]+p_crires->det_array_x-
     (p_crires->pix*(0.5*chipxpix*cos(p_crires->chiprot[ii])+
                     0.5*chipypix*sin(p_crires->chiprot[ii])+20.0));
   p_crires->chipyup[ii]=
     p_crires->chipy[ii]+p_crires->det_array_y+
     (p_crires->pix*(0.5*chipxpix*sin(p_crires->chiprot[ii])+
                     0.5*chipypix*cos(p_crires->chiprot[ii])));
   p_crires->chipydown[ii]=
     p_crires->chipy[ii]+p_crires->det_array_y-
     (p_crires->pix*(0.5*chipxpix*sin(p_crires->chiprot[ii])+
                     0.5*chipypix*cos(p_crires->chiprot[ii])));
 }

 /*******************  Rotation into Detector ********************/

 /*THIS SHOULD BE: crires_model_transpose(tras_ingrat,p_crires->tograt);*/
 crires_model_transpose(tras_ingrat,ingrat);
 /*We do not rotate to the detector FoR, but we still have
   to rotate out of the grating FoR*/
 crires_model_multiplymatrix(p_crires->todetector,unity_mat,tras_ingrat);
 return 0;
}

/*----------------------------------------------------------------------------*/
/**
 * The transformation matrices for the prism are here separate from the rest of
 * init for "historical" reasons (earlier use of the code to compute optimum
 * prism angles in which the prism angle was being iterated but the rest of the
 * configuration remained constant. For current use (wavemap and trace
 * generation and annealing) it does not need to be separate.
 *
 * The function takes the p_crires structure containing all model parameters as
 * input.
 * Nothing is returned, but the p_crires structure itself is modified.
 */
/*----------------------------------------------------------------------------*/
void crires_model_p_rot(struct crires *p_crires)
{
  mat inprism1, inprism2,tras_inprism1,tras_inprism2,tras_inprism3, toaxisprism;
  mat reflect={ { 1,0,0,0}, 
		{ 0,1,0,0}, 
		{ 0,0,1,0}, 
		{ 0,0,0,-1} };
  /*-1 for z-component to change direction of vector*/;

  /*preliminary rotation to allow for prism axis offset*/

  crires_model_rotin(toaxisprism,
                    p_crires->axis_mup,
                    p_crires->axis_nup,
                    p_crires->axis_taup);

  /*******************  Rotation into Prism1********************/
 crires_model_rotin(inprism1,
                    p_crires->mup1,
                    p_crires->nup1,
                    p_crires->taup1);
 crires_model_multiplymatrix(p_crires->toprism1, toaxisprism, inprism1);

 /*******************  Rotation into Prism2********************/
 crires_model_rotin(inprism2,p_crires->mup2,p_crires->nup2,p_crires->taup2);
 crires_model_multiplymatrix(p_crires->toprism2,reflect,inprism2);

 /*******************  Rotation into Prism3********************/
 /*use the same rotation again, after the reflection it is as though the 
   exit surface has been rotated a further 15deg in the same sense*/
 crires_model_rotin(p_crires->toprism3,
                    -p_crires->mup2,
                    -p_crires->nup2,
                    p_crires->taup2);

 /*******************  Rotation back to ref********************/
 crires_model_transpose(tras_inprism1,p_crires->toprism1);
 crires_model_transpose(tras_inprism2,inprism2);
 crires_model_transpose(tras_inprism3,p_crires->toprism3);
 crires_model_multiplythreematrix(p_crires->prism_out,
                                  tras_inprism3,
                                  tras_inprism2,
                                  tras_inprism1);
}


/*----------------------------------------------------------------------------*/
/**
 * The eval function computes the physical location on the detector
 * (mm x and y) for a given wavelength, order and parameter confgration. All
 * transformations which are not dependent upon wavelength and order are
 * pre-computed by the init function.
 *
 * As input the function takes:
 * + double    lambda   - Wavelength (mm)
 * + int       morder   - Spectral order
 * + double ** znse_ref - Zinc selenide refractive index coefficients
 * + the p_crires structure containing all model parameters as input.
 * Nothing is returned, but the p_crires structure itself is modified.
 */
/*----------------------------------------------------------------------------*/
void crires_model_eval(DOUBLE  lambda,
                       int morder,
                       DOUBLE ** znse_ref,
                       struct crires *p_crires)
{
  DOUBLE lambda_um,lamf,rind_up,rind_down,xx,yy,hyp;
  vec vres, vres_is,slitpos2,slitpos,detproj;
  double is1, is2,del_mup,del_mug;
  double xdet_c, xdet_c2,ydet_c,ydet_c2, xydet_c;

/*   mat slitcorr={{1,0,0,0}, */
/* 		{0,1,0,0}, */
/* 		{0,0,1,0}, */
/* 		{0,0,0,1}}; */

/*   mat slituncorr={{1,0,0,0}, */
/* 		{0,1,0,0}, */
/* 		{0,0,1,0}, */
/* 		{0,0,0,1}}; */

  mat mech={ {1,0,0,0},
             {0,1,0,0},
             {-morder*(p_crires->sg),0,1,0},
             {0,0,0,-1}   };

  /* init start vector */
  vec vres1={lambda,
             0,
             0,
             -p_crires->fcp};
  /*----------------------------------------------------------------------*/

  slitpos[0]=0.0;
  slitpos[1]=1.0*p_crires->es_x;
  slitpos[2]=1.0*p_crires->es_y;
  slitpos[3]=0.0;
  crires_model_matrixforvector(slitpos2,p_crires->e_slit,slitpos);
  crires_model_subtractvectors(vres1,slitpos2);
  crires_model_copyvector(vres,vres1);
  crires_model_normall(vres);

  /***********************Rotation into Prism, look in init *******************/
  //printf("a %lf %lf %lf \n",vres[1], vres[2], vres[3]);
  crires_model_matrixforvector(vres,p_crires->toprism1,vres);
  //printf("b %lf %lf %lf \n",vres[1], vres[2], vres[3]);
 /* Next bit has to be different for CRIRES because (to my knowledge)
    there is no approximation that allows implementation of refraction
    as a linear function of wavelength and incident angle so we cannot
    implement this as a matrix multiplication like the grating case*/
  lambda_um=lambda*1000.0;
  lamf=lambda_um*lambda_um;
  /* Calculate refractive index at temperature above using Sellmeier
    equation */
  rind_up=sqrt(znse_ref[1][0]*lamf/(lamf-znse_ref[1][3]*znse_ref[1][3])+
		 znse_ref[1][1]*lamf/(lamf-znse_ref[1][4]*znse_ref[1][4])+
		 znse_ref[1][2]*lamf/(lamf-znse_ref[1][5]*znse_ref[1][5])+1.0);

  /* Calculate refractive index at temperature below using Sellmeier
    equation */
  rind_down=sqrt(znse_ref[0][0]*lamf/(lamf-znse_ref[0][3]*znse_ref[0][3])+
		  znse_ref[0][1]*lamf/(lamf-znse_ref[0][4]*znse_ref[0][4])+
		  znse_ref[0][2]*lamf/(lamf-znse_ref[0][5]*znse_ref[0][5])+1.0);

  /* Linear interpolation to get ref ind at actual temperature */
  p_crires->rind=rind_down
    +((p_crires->temper-znse_ref[0][6])/(znse_ref[1][6]-znse_ref[0][6]))*
    (rind_up-rind_down);

  crires_model_refract(vres,p_crires->rind,vres);

  /* now rotate to the relfection surface and reflect */
 /*NOTE: There is no need to rotate back into the referential here
   because the angle of this surface is given (in the config file
   relative to the last one */
  crires_model_matrixforvector(vres,p_crires->toprism2,vres);

  /* now rotate to exit surface */
  crires_model_matrixforvector(vres,p_crires->toprism3,vres);

  /* refraction as above, except with inverse ref index */
  crires_model_refract(vres,1.0/(p_crires->rind),vres);

  /* Rotate back into referential */
  crires_model_matrixforvector(vres,p_crires->prism_out,vres);

  /* Rotate to intermediate slit
     !! DO NOT ROTATE THE COMPONENT AS ABOVE. THE FOCAL
     LENGTH IS INVOLVED HERE SO THE ROTATION IS IN THE
     PROJECTION, NOT THE REFERENCE FRAME OF THE RAY */
  crires_model_copyvector(vres_is,vres);

  xx=(p_crires->flipisx)*(p_crires->fcp)/
    ((vres[3]/vres[1])*cos(p_crires->nuis)-sin(p_crires->nuis));
  yy=(p_crires->flipisy)*(p_crires->fcp)/
    ((vres[3]/vres[2])*cos(p_crires->muis)-sin(p_crires->muis));
  detproj[0]=0.0;
  detproj[1]=xx*cos(-p_crires->tauis)+yy*sin(-p_crires->tauis);
  detproj[2]=-xx*sin(-p_crires->tauis)+yy*cos(-p_crires->tauis);
  detproj[3]=0.0;
    /* vector with projection */
  crires_model_copyvector(vres_is,detproj);

  p_crires->xis=vres_is[1];
  p_crires->yis=vres_is[2];

  /* On slit? */
  if (p_crires->slit_on==1) {
    p_crires->onslit=-1;
  }
  else {
    p_crires->onslit=0;
  }    
  if (vres_is[1]<p_crires->slitxup &&
      vres_is[1]>p_crires->slitxdown &&
      vres_is[2]<p_crires->slityup &&
      vres_is[2]>p_crires->slitydown) {
    p_crires->onslit=1;
  }

  /*DEPTH OF SLIT COULD BE MODELLED HERE BY MAKING A SECOND PROJECTION
    WITH fcp+0.2mm. NET EFFECT WOULD BE TO MAKE THE SLIT EFFECTIVELY
    NARROWER AT SOME WAVELENGTHS */

  del_mug=(p_crires->mug-p_crires->genc_co1)*100.0;
  del_mup=(p_crires->mup1-p_crires->penc_co1)*100.0;

  is1=vres_is[1];
  is2=p_crires->slity-vres_is[2]+p_crires->offx*del_mup+p_crires->offy*del_mug;
  //is1=11.515;
  //is2=-1.83+p_crires->offx*del_mup+p_crires->offy*del_mug;

  hyp=sqrt(is1*is1+
	   (is2+(p_crires->cam_dis/2.0))*
	   (is2+(p_crires->cam_dis/2.0))+
	   p_crires->fk*p_crires->fk);
  vres[1]=is1/(hyp);
  vres[2]=(is2+(p_crires->cam_dis/2.0))/(hyp);
  vres[3]=p_crires->fk/(hyp);

  vres[3]=1.0;
  crires_model_normz(vres);


  /* Rotate into grating */
  crires_model_matrixforvector(vres,p_crires->tograt,vres);

  /* Difraction */

  crires_model_normz(vres);
  p_crires->grat_alpha=atan(vres[2]/vres[3]);

  crires_model_matrixforvector(vres,mech,vres);

  crires_model_normz(vres);
  p_crires->grat_beta=atan(vres[2]/vres[3]);


  /* Rotation into detector plane (redefined in init to ONLY ROTATE
     OUT OF THE GRATING FoR)*/
  crires_model_matrixforvector(vres, p_crires->todetector,vres);

  //crires_model_transpose(slituncorr,slitcorr);
  //crires_model_matrixforvector(vres,slituncorr,vres);

  /*Note taud~90deg, hence the detproj calculation below essentially swiches
    x and y so that afterwards the x or [1] co-ord is dispersion*/
  xx=(p_crires->flipx)*(p_crires->fk)/
    ((vres[3]/vres[1])*cos(p_crires->nud)-sin(p_crires->nud));
  yy=(p_crires->flipy)*(p_crires->fk)/
    ((vres[3]/vres[2])*cos(p_crires->mud)-sin(p_crires->mud));
  detproj[0]=0.0;
  detproj[1]=xx*cos(-p_crires->taud)+yy*sin(-p_crires->taud);
  detproj[2]=-xx*sin(-p_crires->taud)+yy*cos(-p_crires->taud);
  detproj[3]=0.0; 
  /* vector with projection */
  crires_model_copyvector(vres,detproj);


  xdet_c=vres[1]-p_crires->det_array_x;
  xdet_c2=xdet_c*xdet_c;
  ydet_c=vres[2]-p_crires->det_array_y;
  ydet_c2=ydet_c*ydet_c;
  xydet_c=xdet_c*ydet_c;

  p_crires->xdet=p_crires->det_array_x+
    p_crires->pc_y_xx[0]*del_mug+p_crires->pc_y_xx[3]*del_mug*del_mug+p_crires->pc_y_yy[0]*del_mup+p_crires->pc_y_yy[3]*del_mup*del_mup+p_crires->pc_y_xy[0]*del_mup*del_mug+
    (1.0+p_crires->pc_x_xx[0]*del_mug+p_crires->pc_x_xx[3]*del_mug*del_mug+p_crires->pc_x_yy[0]*del_mup+p_crires->pc_x_yy[3]*del_mup*del_mup+p_crires->pc_x_xy[0]*del_mup*del_mug)*xdet_c+
    p_crires->pc_x_xx[1]*xdet_c2+
    p_crires->pc_x_yy[1]*ydet_c2+
    p_crires->pc_x_yy[2]*ydet_c+
    p_crires->pc_x_xy[1]*xydet_c+
    p_crires->pc_x_xy[3]*is2;
  p_crires->ydet=p_crires->det_array_y
    +p_crires->pc_y_xx[2]*xdet_c+
    p_crires->pc_y_xx[1]*xdet_c2+
    p_crires->pc_y_yy[1]*ydet_c2+
    p_crires->pc_y_yy[2]*ydet_c+
    p_crires->pc_y_xy[1]*xydet_c;

}

/*----------------------------------------------------------------------------*/
/**
 * The detpix function computes the detector co-ordinates corresponding to
 * the physical location (mm x,y)
 * computed by the eval function (p_crires->xdet, p_crires->ydet). The
 * detector co-ordinates are written to p_crires->xpospix and p_crires->ypospix
 * (float) and p_crires->chippix[N] where N=0 contains the chip number (0-3)
 * and N=1,2 contain the x,y co-ordinates (integer)
 *
 * As input the function takes the p_crires structure containing all model
 * parameters as input.
 * Nothing is returned, but the p_crires structure itself is modified.
 */
/*----------------------------------------------------------------------------*/
void crires_model_detpix(struct crires *p_crires)
{
  /*Assume that all chip surfaces lie in the same
    plane (i.e. mud, nud and taud describe the plane of
    all chips). However allow individual rotations of
    each chip within the plane: rot1 to rot4 */
  int ii;
  double pc_x, pc_y;
  double xdet_c, xdet_c2,ydet_c,ydet_c2, xydet_c;

  /* Find which chip (if any) */
  p_crires->chippix[0]=-1;
  p_crires->chippix[1]=0;
  p_crires->chippix[2]=0;
  if (p_crires->xdet<p_crires->chipxdown[0])
  {
    p_crires->chippix[0]=5;
  }
  else if (p_crires->xdet>p_crires->chipxup[3])
  {
    p_crires->chippix[0]=6;
  }
  else if (p_crires->ydet>p_crires->chipyup[0])
  {
    p_crires->chippix[0]=7;
  }
  else if (p_crires->ydet<p_crires->chipydown[0])
  {
    p_crires->chippix[0]=8;
  }
  p_crires->xpospix=0.0;
  p_crires->ypospix=0.0;
  for (ii=0; ii < 4; ii++) {
    if (p_crires->xdet<p_crires->chipxup[ii] &&
        p_crires->xdet>p_crires->chipxdown[ii] &&
        p_crires->ydet<p_crires->chipyup[ii] &&
        p_crires->ydet>p_crires->chipydown[ii] &&
        p_crires->onslit!=-1) {
      p_crires->chippix[0]=ii;

      xdet_c=p_crires->xdet-(p_crires->det_array_x+p_crires->chipx[ii]);
      xdet_c2=xdet_c*xdet_c;
      ydet_c=p_crires->ydet-(p_crires->det_array_y+p_crires->chipy[ii]);
      ydet_c2=ydet_c*ydet_c;
      xydet_c=xdet_c*ydet_c;

  /**********PIN CUSHION DISTORTION************/
      pc_x=p_crires->det_array_x+p_crires->chipx[ii]+p_crires->pc_x_x1[ii]*xdet_c+p_crires->pc_x_y1[ii]*ydet_c;

      pc_y=p_crires->det_array_y+p_crires->chipy[ii]+p_crires->pc_y_y1[ii]*ydet_c+p_crires->pc_y_x1[ii]*xdet_c;

      p_crires->xdet=pc_x;
      p_crires->ydet=pc_y;

     /* xpospix=0.5,ypospix=0.5 would be the middle of the bottom left pixel */
      p_crires->xpospix=
        0.5*chipxpix+
        ((p_crires->xdet-(p_crires->chipx[ii]+p_crires->det_array_x))*
         cos(p_crires->chiprot[ii])-
         (p_crires->ydet-(p_crires->chipy[ii]+p_crires->det_array_y))*
         sin(p_crires->chiprot[ii]))/p_crires->pix;
      p_crires->ypospix=
        0.5*chipypix+
        ((p_crires->ydet-(p_crires->chipy[ii]+p_crires->det_array_y))*
         cos(p_crires->chiprot[ii])+
         (p_crires->xdet-(p_crires->chipx[ii]+p_crires->det_array_x))*
         sin(p_crires->chiprot[ii]))/p_crires->pix;
      /*The limits in chipxup...chipydown are not
        correct away from the axis of the chip. A pixel
        could fall within these limits but not actually fall on the chip. So:*/
      if (p_crires->xpospix>1023) {
        p_crires->xpospix=-1.0;
        p_crires->ypospix=-1.0;
        if (p_crires->chippix[0]==3) {
          p_crires->chippix[0]=6;
        }
        else {
          p_crires->chippix[0]=4;
        }
      }
      if (p_crires->xpospix<0) {
        p_crires->xpospix=-1.0;
        p_crires->ypospix=-1.0;
        if (p_crires->chippix[0]==0) {
          p_crires->chippix[0]=5;
        }
        else {
          p_crires->chippix[0]=4;
        }
      }
      if (p_crires->ypospix<0) {
        p_crires->xpospix=-1.0;
        //p_crires->ypospix=-1.0;
        p_crires->chippix[0]=8;
      }
      if (p_crires->ypospix>511) {
        p_crires->xpospix=-1.0;
        p_crires->ypospix=-1.0;
        p_crires->chippix[0]=7;
      }
      /*N<xpospix<N+1 => chippix[1]=N*/
      p_crires->chippix[1]=(int)(p_crires->xpospix);
      p_crires->chippix[2]=(int)(p_crires->ypospix);
    }
  }
  /* if we get this far and xdet and ydet are within the
     overall detector area, but not on any chip and the
     int. slit was passed, then the light must arrive on 
     one of the gaps between the chips, we set chippix[0]=4*/
  if (p_crires->onslit!=-1 && p_crires->chippix[0]==-1) p_crires->chippix[0]=4;
} 

/*----------------------------------------------------------------------------*/
/**
 * The ref_ind_read function no longer reads the refractive index sellmeier
 * co-effs, instead they are hard coded below. The function takes an empty
 * znse_ref array and fills the two rows with the selmeier coefficients
 * bracketing the temperature given. Interpolation is performed in the eval
 * function.
 *
 * As input the function takes the znse_ref array and the prism temperature
 * (temper)
 * 0 is returned upon success, -1 if there is an error. The znse_ref array
 * itself is filled.
 */
/*----------------------------------------------------------------------------*/
int crires_model_ref_ind_read(DOUBLE ** znse_ref, DOUBLE temper)
{
  int ii,jj,flag;
  /* The ZnSe Sellmeier coeffs for 65-90K are hard coded here */
  DOUBLE znse_ref_temp[8][7]={{0.0,0.0,0.0,0.0,0.0,0.0,0.0},
              {0.0,0.0,0.0,0.0,0.0,0.0,0.0},
              {4.43225472568,0.436053754909,2.93894051156,0.199312044041,
               0.384261841485,47.790690406,65.0},
              {4.43273532483,0.43649702919,2.92549054805,0.199315036235,
               0.384358520608,47.698208287,70.0},
              {4.43324606413,0.43695072796,2.91292914202,0.199321428826,
               0.384462450375,47.6159570681,75.0},
              {4.43378545254,0.437413966565,2.90136392313,0.199330985656,
               0.384573434589,47.5429491974,80.0},
              {4.43435205486,0.437885881855,2.89088177296,0.199343477639,
               0.384691257373,47.4782846413,85.0},
              {4.43494449266,0.438365631134,2.88154865542,0.199358683469,
               0.384815688125,47.4211647405,90.0}};
  for (ii=0;ii<8;ii++) {
      for (jj=0;jj<7;jj++) {
          znse_ref[ii][jj]=znse_ref_temp[ii][jj];
      }
  }
  flag=0;
  for (ii=2;ii<8;ii++) {
    /*      znse_ref[*][6] holds the temperature of the fit */
    if (temper>=znse_ref[ii][6]) {
      /*fill row [0][*] with the current values, the
        last time that this will be overwritten is the
        last time that temper is below the temperature value read in */
      flag=1;
      for (jj=0; jj<7; jj++) {
        znse_ref[0][jj]=znse_ref[ii][jj];
      }
    }
    else if (temper<znse_ref[ii][6] && flag==1) {
      /*fill row [1][*] with the current values, this
        will only be written once */
      flag=2;
      for (jj=0; jj<7; jj++) {
        znse_ref[1][jj]=znse_ref[ii][jj];
      }
    }      
  }
    /*if the conditions are never met so that flag=2
      then the temperature is out of range */
  if (flag!=2) {
    cpl_msg_error(__func__, "Temperature out of range %lf", temper);
    return -1;
  }
  return 0;
}


DOUBLE * crires_model_alloc1Darray(int asize)
{  
  DOUBLE *array;
  /* get memory for the 1D array */
  if ((array=(DOUBLE *)cpl_malloc(asize*sizeof(DOUBLE))) == NULL)
    {
      cpl_msg_error(__func__, "Failed to allocate 1D array" );
      return NULL;
    }  
  return array;   
}

double** crires_model_alloc2Darray(int asize, int bsize)
{  
     int ii;
     double **ccdtemp;
     /* get memory for all the 2D arrays */
     if ((ccdtemp=(double **)cpl_malloc(asize*sizeof(double *))) == NULL)
     {
       cpl_msg_error(__func__, "Failed to allocate 2D array at initial \
pointer");
       return NULL;
     }

     for (ii = 0; ii < asize; ii++)
     {
       if ( (ccdtemp[ii]=(double *)cpl_malloc(bsize*sizeof(double))) == NULL)
       {
	 cpl_msg_error(__func__, "Failed to allocate 2D array at row=%d\n",ii);
	 return NULL;
       }
     }
  return ccdtemp;   
}

float** crires_model_alloc2Darray_f(int asize, int bsize)
{  
     int ii;
     float **ccdtemp;
     /* get memory for all the 2D arrays */
     if ((ccdtemp=(float **)cpl_malloc(asize*sizeof(float *))) == NULL)
     {
       cpl_msg_error(__func__, "Failed to allocate 2D array at initial \
pointer");
       return NULL;
     }
     for (ii = 0; ii < asize; ii++)
     {
       if ( (ccdtemp[ii]=(float *)cpl_malloc(bsize*sizeof(float))) == NULL)
       {
	 cpl_msg_error(__func__, "Failed to allocate 2D array at row=%d\n",ii);
	 return NULL;
       }
     }
  return ccdtemp;   
}

int** crires_model_alloc2Darray_i(int asize, int bsize)
{  
     int ii;
     int** ccdtemp;
     /* get memory for all the 2D arrays */
     if ((ccdtemp=(int**)cpl_malloc(asize*sizeof(int*))) == NULL)
     {
       cpl_msg_error(__func__, "Failed to allocate 2D array at initial \
pointer");
       return NULL;
     }

     for (ii = 0; ii < asize; ii++)
     {
       if ( (ccdtemp[ii]=(int*)cpl_malloc(bsize*sizeof(int))) == NULL)
       {
	 cpl_msg_error(__func__, "Failed to allocate 2D array at row=%d\n",ii);
	 return NULL;
       }
     }
  return ccdtemp;   
}

int crires_model_free2Darray(double ** ccdtemp, int asize) {
  int i;
  const int my_asize=asize;
  for (i=my_asize-1; i >=0; i--) {
      if (ccdtemp[i] !=NULL) cpl_free(ccdtemp[i]);   
     else {
	 cpl_msg_error(__func__, "Error freeing 2D array at row=%d\n",i);
	 return -1;
       }      
    }
  if( ccdtemp !=NULL) cpl_free(ccdtemp); 
  else {
    cpl_msg_error(__func__, "Error freeing 2D array at initial pointer");
    return -1;
  }
  return 0;   
}

int crires_model_free2Darray_i(int** ccdtemp, int asize) {
  int i;
  for (i=asize-1; i >=0; i--) {
     if (ccdtemp[i] !=NULL) cpl_free(ccdtemp[i]);   
     else {
	 cpl_msg_error(__func__, "Error freeing 2D array at row=%d\n",i);
	 return -1;
       }      
    }
  if (ccdtemp !=NULL) cpl_free(ccdtemp); 
  else {
    cpl_msg_error(__func__, "Error freeing 2D array at initial pointer");
    return -1;
  }
  return 0;   
}

int crires_model_free2Darray_f(float ** ccdtemp, int asize)
{
  int i;
  for (i=asize-1; i >=0; i--) 
    {
     if (ccdtemp[i] !=NULL) cpl_free(ccdtemp[i]);   
     else 
       {
	 cpl_msg_error(__func__, "Error freeing 2D array at row=%d\n",i);
	 return -1;
       }      
    }
  if( ccdtemp !=NULL) cpl_free(ccdtemp); 
  else 
  {
    cpl_msg_error(__func__, "Error freeing 2D array at initial pointer");
    return -1;
  }
  return 0;   
}

double * crires_model_copy2D_to_1D(double ** ccdtemp,int asize, int bsize)
{
  int ii;
  int jj;
  int kk=0;
  double *oneDccdtemp=NULL; 
  if ((oneDccdtemp=(double *)cpl_malloc(asize*bsize*sizeof(double))) == NULL)
  {
    cpl_msg_error(__func__, "Error allocating memory for array copy\n");
    return NULL;
  }
  for (jj = 0; jj < bsize; jj++) {
    for (ii = 0; ii < asize; ii++) {
      oneDccdtemp[kk]=ccdtemp[ii][jj];
      kk +=1;
    }
  }  
  return oneDccdtemp;
}

double ** crires_model_copy1D_to_2D(double * oneDccdtemp, int asize, int bsize)
{
  int ii;
  int jj;
  int kk=0;
  double **ccdtemp=NULL;
  if ((ccdtemp=crires_model_alloc2Darray(asize,bsize))==NULL) {
    cpl_msg_error(__func__, "Error allocating memory for array copy\n");
    return NULL;
  }
  for (ii = 0; ii < asize; ii++)
    { 
      for (jj = 0; jj < bsize; jj++)
	{
	  ccdtemp[ii][jj]=oneDccdtemp[kk];
	  kk +=1;
	}
    }  
  return ccdtemp;
}

void crires_model_addvectors(vec a, vec b)
{
  int i;
  for(i=0; i<vectordim; i++) a[i]+=b[i];
}

void crires_model_addvectors3D(vec a, vec b)
{
  int i;
  for(i=0; i<vectordim-1; i++) a[i]+=b[i+1];
}

void crires_model_subtractvectors(vec a, vec b)
{
  int i;
  for(i=0; i<vectordim; i++) a[i]-=b[i];
}

DOUBLE crires_model_scalarproduct(vec a, vec b)
{
  int i;
  DOUBLE scalar;
  scalar=0.0;
  for (i=0; i<vectordim; i++) {
    scalar += a[i]*b[i];
  }
  return scalar; 
}

void crires_model_multiply(vec a, DOUBLE k)
{
  int i;
  for(i=0; i<vectordim; i++) a[i] *=k;
}

void crires_model_showvector(vec a)
{
 int i;
 for(i=0; i<vectordim; i++) printf("%lf ", a[i]*1000.0); 
 printf("          \n"); 
}

void crires_model_nullvector(vec a)
{
  int i;
  for(i=0; i<vectordim; i++) a[i]=0.0;
}

void crires_model_copyvector(vec a, vec b)
{
 int i;
 for(i=0; i<vectordim; i++) a[i]=b[i];
}

/* normalize by z^2 = 1- (x^2 + y^2) */
void crires_model_normz(vec a)
{
     /*this is WRONG! 
 if (fabs(a[1]) > 1.0 || fabs(a[2]) > 1.0){
 double norm = sqrt (a[1]*a[1] +  a[2]*a[2] +  a[3]*a[3]);
 a[1] /=norm;
 a[2] /=norm;
 a[3] /=norm;
 }*/
/* should be */
  
  if (fabs(a[2]) > 1.0) {
    a[1]=0.0;
    a[2]=1.0;
    a[3]=0.0;
  }
  else if (fabs(a[1]) > 1.0) {
    a[1]=1.0;
    a[2]=0.0;
    a[3]=0.0;
  }
  else if ((a[1]*a[1] +  a[2]*a[2]) > 1.0) {
    double norm = sqrt (a[1]*a[1] +  a[2]*a[2]);
    a[1] /=norm;
    a[2] /=norm;
    a[3]=0.0;
  }
  else{
    /* as before*/
    if(a[3]>=0) a[3]=sqrt(1.0-(a[1]*a[1]+a[2]*a[2]));
    else a[3]=-sqrt(1.0-(a[1]*a[1]+a[2]*a[2]));
  }
}

/* normalize by  1/sqrt(x^2 + y^2 + z^2) */
void crires_model_normall(vec a)
{
   DOUBLE mag;
   mag=sqrt(a[1]*a[1]+a[2]*a[2]+a[3]*a[3]);
   a[1]=a[1]/mag;
   a[2]=a[2]/mag;
   a[3]=a[3]/mag;
}

void crires_model_multiplymatrix(mat A, mat B, mat C)
{
  int i,j,k; 
  crires_model_nullmatrix(A);
  for(i=0; i<vectordim; i++) { 
    for(j=0; j<vectordim; j++) {
      for(k=0; k<vectordim; k++){
          A[i][j] +=B[i][k]*C[k][j];
        }
    }
  }
}

void crires_model_multiplythreematrix(mat A, mat B, mat C,mat D)
{
 mat temp;
 crires_model_nullmatrix(temp);
/* multiply C*D and store the result in temp  */
 crires_model_multiplymatrix(temp,C,D); 
/* multiply B*temp=B*C*D and store the result in A */
 crires_model_multiplymatrix(A,B,temp); 
} 

void crires_model_showmatrix(mat A)
{
 int i;
 for(i=0; i<vectordim; i++) {
   printf("%lf  %lf  %lf  %lf\n", A[i][0],A[i][1],A[i][2],A[i][3]);  
 }
  printf("          \n");
}

void crires_model_initializematrix(mat A)
{
  int i,j;
   for(i=0; i<vectordim; i++) { 
    for(j=0; j<vectordim; j++) {
      A[i][j]=i+j; 
    }
   }
}

void crires_model_nullmatrix(mat A)
{
  int i,j;
   for(i=0; i<vectordim; i++) { 
    for(j=0; j<vectordim; j++) {
      A[i][j]=0.0; 
    }
   }
}
void crires_model_copymatrix(mat A, mat B)
{
  int i,j;
   for(i=0; i<vectordim; i++) { 
    for(j=0; j<vectordim; j++) {
      A[i][j]=B[i][j]; 
    }
   }
}

void crires_model_matrixforvector(vec a,mat B,vec c)
{
  int i,j;
  vec temp;
  crires_model_nullvector(temp);
  for(i=0; i<vectordim; i++) { 
    for(j=0; j<vectordim; j++) {
        temp[i] +=B[i][j]*c[j];
    }
  }
  crires_model_copyvector(a,temp);  
}

void crires_model_transpose(mat A ,mat B)
{
   int i,j;
   for(i=0; i<vectordim; i++) { 
    for(j=0; j<vectordim; j++) {
      A[i][j]=B[j][i]; 
    }
   }
}

void crires_model_rotationmatrix(mat A, const char axis, const DOUBLE angle)
{  
  crires_model_nullmatrix(A); /*all element to zero */
  switch(axis) {
  /*change only the element involved in the x rotation  */
    case 'x':{
    /*  1 0 0 0 , 0 1 0 0, 0 0 c s , 0 0 -s c   (left handed) */
      A[0][0]=A[1][1]=1;
      A[2][2]=A[3][3]=cos(angle);
      A[2][3]=sin(angle); 
      A[3][2]= -A[2][3]; 
      break;
    }
  /*change only the element involved in the y rotation  */
    case 'y':{
    /*  1 0 0 0 , 0 c 0 -s, 0 0 1 0 , 0 s 0 c   (left handed) */
      A[0][0]=A[2][2]=1;
      A[1][1]=A[3][3]=cos(angle);
      A[1][3]=-sin(angle); 
      A[3][1]=-A[1][3]; 
      break;
  }
  /*change only the element involved in the z rotation  */
    case 'z':{
      A[0][0]=A[3][3]=1;
      A[1][1]=A[2][2]=cos(angle);
      A[1][2]=sin(angle); 
      A[2][1]=-A[1][2]; 
      break;
    }
  }
}
        
void crires_model_rotin(mat A,const DOUBLE x_angle,
                        const DOUBLE y_angle,
                        const DOUBLE z_angle)
{
  mat temp;
  mat X,Y,Z;
  crires_model_rotationmatrix(X,'x',x_angle);
  crires_model_rotationmatrix(Y,'y',y_angle); 
  crires_model_rotationmatrix(Z,'z',z_angle);

  crires_model_multiplymatrix(temp,Y,X);  
/* multiply Y*X and store the result in temp */
  crires_model_multiplymatrix(A,Z,temp);  
/* multiply Z*temp=Z*Y*X and store the result in A */
}      
void crires_model_refract(vec b, DOUBLE rind, vec a)
{

  DOUBLE phi, pheta, signflag,refrangle;
  signflag=1.0;
  phi=atan(a[3]/sqrt(a[1]*a[1]+a[2]*a[2]));
  pheta=atan(a[1]/a[2]);
  signflag=1.0;
  if (phi<0.0)
  {
    signflag=-1.0;
    phi=-phi;
  }
  refrangle=asin(sin((CPL_MATH_PI/2.0)-phi)/rind);
  b[0]=a[0]; /*preserve wavelength */
  b[1]=signflag*sin(pheta)*cos((CPL_MATH_PI/2.0)-refrangle);
  b[2]=signflag*cos(pheta)*cos((CPL_MATH_PI/2.0)-refrangle);
  b[3]=signflag*sin((CPL_MATH_PI/2.0)-refrangle);
}

int crires_model_preset_angles(char* preset,
				struct crires * p_crires)
{
  int penc, genc, compa, order;
  order=-1;
  compa=strncmp(preset,"59/0/n",6);
  if (compa==0) {
    penc=54346;
    genc=209247;
    order=59;
  }
  compa=strncmp(preset,"59/0/i",6);
  if (compa==0) {
    penc=54105;
    genc=198217;
    order=59;
  }
  compa=strncmp(preset,"58/0/n",6);
  if (compa==0) {
    penc=52949;
    genc=209243;
    order=58;
  }
  compa=strncmp(preset,"58/0/i",6);
  if (compa==0) {
    penc=52717;
    genc=198217;
    order=58;
  }
  compa=strncmp(preset,"57/0/n",6);
  if (compa==0) {
    penc=51584;
    genc=209239;
    order=57;
  }
  compa=strncmp(preset,"57/0/i",6);
  if (compa==0) {
    penc=51361;
    genc=198217;
    order=57;
  }
  compa=strncmp(preset,"56/0/n",6);
  if (compa==0) {
    penc=50251;
    genc=209235;
    order=56;
  }
  compa=strncmp(preset,"56/0/i",6);
  if (compa==0) {
    penc=50037;
    genc=198218;
    order=56;
  }
  compa=strncmp(preset,"55/0/n",6);
  if (compa==0) {
    penc=48948;
    genc=209231;
    order=55;
  }
  compa=strncmp(preset,"55/0/i",6);
  if (compa==0) {
    penc=48743;
    genc=198218;
    order=55;
  }
  compa=strncmp(preset,"54/0/n",6);
  if (compa==0) {
    penc=47676;
    genc=209227;
    order=54;
  }
  compa=strncmp(preset,"54/0/i",6);
  if (compa==0) {
    penc=47479;
    genc=198218;
    order=54;
  }
  compa=strncmp(preset,"53/0/n",6);
  if (compa==0) {
    penc=46434;
    genc=209224;
    order=53;
  }
  compa=strncmp(preset,"53/0/i",6);
  if (compa==0) {
    penc=46245;
    genc=198218;
    order=53;
  }
  compa=strncmp(preset,"52/0/n",6);
  if (compa==0) {
    penc=45221;
    genc=209220;
    order=52;
  }
  compa=strncmp(preset,"52/0/i",6);
  if (compa==0) {
    penc=45041;
    genc=198218;
    order=52;
  }
  compa=strncmp(preset,"51/0/n",6);
  if (compa==0) {
    penc=44037;
    genc=209217;
    order=51;
  }
  compa=strncmp(preset,"51/0/i",6);
  if (compa==0) {
    penc=43864;
    genc=198218;
    order=51;
  }
  compa=strncmp(preset,"50/0/n",6);
  if (compa==0) {
    penc=42881;
    genc=209213;
    order=50;
  }
  compa=strncmp(preset,"50/0/i",6);
  if (compa==0) {
    penc=42716;
    genc=198219;
    order=50;
  }
  compa=strncmp(preset,"49/0/n",6);
  if (compa==0) {
    penc=41754;
    genc=209210;
    order=49;
  }
  compa=strncmp(preset,"49/0/i",6);
  if (compa==0) {
    penc=41596;
    genc=198219;
    order=49;
  }
  compa=strncmp(preset,"48/0/n",6);
  if (compa==0) {
    penc=40654;
    genc=209207;
    order=48;
  }
  compa=strncmp(preset,"48/0/i",6);
  if (compa==0) {
    penc=40503;
    genc=198219;
    order=48;
  }
  compa=strncmp(preset,"47/0/n",6);
  if (compa==0) {
    penc=39581;
    genc=209204;
    order=47;
  }
  compa=strncmp(preset,"47/0/i",6);
  if (compa==0) {
    penc=39437;
    genc=198219;
    order=47;
  }
  compa=strncmp(preset,"46/0/n",6);
  if (compa==0) {
    penc=38534;
    genc=209201;
    order=46;
  }
  compa=strncmp(preset,"46/0/i",6);
  if (compa==0) {
    penc=38397;
    genc=198219;
    order=46;
  }
  compa=strncmp(preset,"45/0/n",6);
  if (compa==0) {
    penc=37514;
    genc=209198;
    order=45;
  }
  compa=strncmp(preset,"45/0/i",6);
  if (compa==0) {
    penc=37383;
    genc=198219;
    order=45;
  }
  compa=strncmp(preset,"44/0/n",6);
  if (compa==0) {
    penc=36519;
    genc=209196;
    order=44;
  }
  compa=strncmp(preset,"44/0/i",6);
  if (compa==0) {
    penc=36395;
    genc=198219;
    order=44;
  }
  compa=strncmp(preset,"43/-1/n",7);
  if (compa==0) {
    penc=36037;
    genc=251895;
    order=43;
  }
  compa=strncmp(preset,"43/-1/i",7);
  if (compa==0) {
    penc=35908;
    genc=240917;
    order=43;
  }
  compa=strncmp(preset,"43/1/n",6);
  if (compa==0) {
    penc=35081;
    genc=164423;
    order=43;
  }
  compa=strncmp(preset,"43/1/i",6);
  if (compa==0) {
    penc=34972;
    genc=153453;
    order=43;
  }
  compa=strncmp(preset,"42/-1/n",7);
  if (compa==0) {
    penc=35069;
    genc=251870;
    order=42;
  }
  compa=strncmp(preset,"42/-1/i",7);
  if (compa==0) {
    penc=34946;
    genc=240895;
    order=42;
  }
  compa=strncmp(preset,"42/1/n",6);
  if (compa==0) {
    penc=34160;
    genc=164444;
    order=42;
  }
  compa=strncmp(preset,"42/1/i",6);
  if (compa==0) {
    penc=34056;
    genc=153477;
    order=42;
  }
  compa=strncmp(preset,"41/-1/n",7);
  if (compa==0) {
    penc=34126;
    genc=251845;
    order=41;
  }
  compa=strncmp(preset,"41/-1/i",7);
  if (compa==0) {
    penc=34009;
    genc=240873;
    order=41;
  }
  compa=strncmp(preset,"41/1/n",6);
  if (compa==0) {
    penc=33262;
    genc=164467;
    order=41;
  }
  compa=strncmp(preset,"41/1/i",6);
  if (compa==0) {
    penc=33163;
    genc=153502;
    order=41;
  }
  compa=strncmp(preset,"40/-1/n",7);
  if (compa==0) {
    penc=33208;
    genc=251820;
    order=40;
  }
  compa=strncmp(preset,"40/-1/i",7);
  if (compa==0) {
    penc=33097;
    genc=240850;
    order=40;
  }
  compa=strncmp(preset,"40/1/n",6);
  if (compa==0) {
    penc=32388;
    genc=164490;
    order=40;
  }
  compa=strncmp(preset,"40/1/i",6);
  if (compa==0) {
    penc=32294;
    genc=153527;
    order=40;
  }
  compa=strncmp(preset,"39/-1/n",7);
  if (compa==0) {
    penc=32315;
    genc=251794;
    order=39;
  }
  compa=strncmp(preset,"39/-1/i",7);
  if (compa==0) {
    penc=32210;
    genc=240827;
    order=39;
  }
  compa=strncmp(preset,"39/1/n",6);
  if (compa==0) {
    penc=31537;
    genc=164513;
    order=39;
  }
  compa=strncmp(preset,"39/1/i",6);
  if (compa==0) {
    penc=31448;
    genc=153553;
    order=39;
  }
  compa=strncmp(preset,"38/-1/n",7);
  if (compa==0) {
    penc=31445;
    genc=251767;
    order=38;
  }
  compa=strncmp(preset,"38/-1/i",7);
  if (compa==0) {
    penc=31346;
    genc=240802;
    order=38;
  }
  compa=strncmp(preset,"38/1/n",6);
  if (compa==0) {
    penc=30708;
    genc=164538;
    order=38;
  }
  compa=strncmp(preset,"38/1/i",6);
  if (compa==0) {
    penc=30643;
    genc=153580;
    order=38;
  }
  compa=strncmp(preset,"37/-1/n",7);
  if (compa==0) {
    penc=30600;
    genc=251739;
    order=37;
  }
  compa=strncmp(preset,"37/-1/i",7);
  if (compa==0) {
    penc=30506;
    genc=240777;
    order=37;
  }
  compa=strncmp(preset,"37/1/n",6);
  if (compa==0) {
    penc=29920;
    genc=164564;
    order=37;
  }
  compa=strncmp(preset,"37/1/i",6);
  if (compa==0) {
    penc=29840;
    genc=153608;
    order=37;
  }
  compa=strncmp(preset,"36/-1/n",7);
  if (compa==0) {
    penc=29777;
    genc=251710;
    order=36;
  }
  compa=strncmp(preset,"36/-1/i",7);
  if (compa==0) {
    penc=29688;
    genc=240751;
    order=36;
  }
  compa=strncmp(preset,"36/1/n",6);
  if (compa==0) {
    penc=29135;
    genc=164592;
    order=36;
  }
  compa=strncmp(preset,"36/1/i",6);
  if (compa==0) {
    penc=29059;
    genc=153637;
    order=36;
  }
  compa=strncmp(preset,"35/-1/n",7);
  if (compa==0) {
    penc=28978;
    genc=251680;
    order=35;
  }
  compa=strncmp(preset,"35/-1/i",7);
  if (compa==0) {
    penc=28894;
    genc=240723;
    order=35;
  }
  compa=strncmp(preset,"35/1/n",6);
  if (compa==0) {
    penc=28370;
    genc=164621;
    order=35;
  }
  compa=strncmp(preset,"35/1/i",6);
  if (compa==0) {
    penc=28299;
    genc=153668;
    order=35;
  }
  compa=strncmp(preset,"34/-1/n",7);
  if (compa==0) {
    penc=28201;
    genc=251649;
    order=34;
  }
  compa=strncmp(preset,"34/-1/i",7);
  if (compa==0) {
    penc=28138;
    genc=240693;
    order=34;
  }
  compa=strncmp(preset,"34/1/n",6);
  if (compa==0) {
    penc=27627;
    genc=164651;
    order=34;
  }
  compa=strncmp(preset,"34/1/i",6);
  if (compa==0) {
    penc=27559;
    genc=153700;
    order=34;
  }
  compa=strncmp(preset,"33/-1/n",7);
  if (compa==0) {
    penc=27461;
    genc=251616;
    order=33;
  }
  compa=strncmp(preset,"33/-1/i",7);
  if (compa==0) {
    penc=27386;
    genc=240662;
    order=33;
  }
  compa=strncmp(preset,"33/1/n",6);
  if (compa==0) {
    penc=26904;
    genc=164684;
    order=33;
  }
  compa=strncmp(preset,"33/1/i",6);
  if (compa==0) {
    penc=26840;
    genc=153735;
    order=33;
  }
  compa=strncmp(preset,"32/-1/n",7);
  if (compa==0) {
    penc=26726;
    genc=251580;
    order=32;
  }
  compa=strncmp(preset,"32/-1/i",7);
  if (compa==0) {
    penc=26655;
    genc=240629;
    order=32;
  }
  compa=strncmp(preset,"32/1/n",6);
  if (compa==0) {
    penc=26201;
    genc=164719;
    order=32;
  }
  compa=strncmp(preset,"32/1/i",6);
  if (compa==0) {
    penc=26141;
    genc=153771;
    order=32;
  }
  compa=strncmp(preset,"31/-1/n",7);
  if (compa==0) {
    penc=26011;
    genc=251543;
    order=31;
  }
  compa=strncmp(preset,"31/-1/i",7);
  if (compa==0) {
    penc=25944;
    genc=240594;
    order=31;
  }
  compa=strncmp(preset,"31/1/n",6);
  if (compa==0) {
    penc=25518;
    genc=164756;
    order=31;
  }
  compa=strncmp(preset,"31/1/i",6);
  if (compa==0) {
    penc=25461;
    genc=153811;
    order=31;
  }
  compa=strncmp(preset,"30/-1/n",7);
  if (compa==0) {
    penc=25316;
    genc=251503;
    order=30;
  }
  compa=strncmp(preset,"30/-1/i",7);
  if (compa==0) {
    penc=25254;
    genc=240555;
    order=30;
  }
  compa=strncmp(preset,"30/1/n",6);
  if (compa==0) {
    penc=24853;
    genc=164797;
    order=30;
  }
  compa=strncmp(preset,"30/1/i",6);
  if (compa==0) {
    penc=24799;
    genc=153853;
    order=30;
  }
  compa=strncmp(preset,"29/-1/n",7);
  if (compa==0) {
    penc=24641;
    genc=251460;
    order=29;
  }
  compa=strncmp(preset,"29/-1/i",7);
  if (compa==0) {
    penc=24583;
    genc=240514;
    order=29;
  }
  compa=strncmp(preset,"29/1/n",6);
  if (compa==0) {
    penc=24206;
    genc=164841;
    order=29;
  }
  compa=strncmp(preset,"29/1/i",6);
  if (compa==0) {
    penc=24156;
    genc=153899;
    order=29;
  }
  compa=strncmp(preset,"28/-1/n",7);
  if (compa==0) {
    penc=23985;
    genc=251413;
    order=28;
  }
  compa=strncmp(preset,"28/-1/i",7);
  if (compa==0) {
    penc=23930;
    genc=240469;
    order=28;
  }
  compa=strncmp(preset,"28/1/n",6);
  if (compa==0) {
    penc=23576;
    genc=164890;
    order=28;
  }
  compa=strncmp(preset,"28/1/i",6);
  if (compa==0) {
    penc=23529;
    genc=153948;
    order=28;
  }
  compa=strncmp(preset,"27/-1/n",7);
  if (compa==0) {
    penc=23347;
    genc=251362;
    order=27;
  }
  compa=strncmp(preset,"27/-1/i",7);
  if (compa==0) {
    penc=23295;
    genc=240419;
    order=27;
  }
  compa=strncmp(preset,"27/1/n",6);
  if (compa==0) {
    penc=22963;
    genc=164943;
    order=27;
  }
  compa=strncmp(preset,"27/1/i",6);
  if (compa==0) {
    penc=22919;
    genc=154003;
    order=27;
  }
  compa=strncmp(preset,"26/-1/n",7);
  if (compa==0) {
    penc=22725;
    genc=251305;
    order=26;
  }
  compa=strncmp(preset,"26/-1/i",7);
  if (compa==0) {
    penc=22676;
    genc=240364;
    order=26;
  }
  compa=strncmp(preset,"26/1/n",6);
  if (compa==0) {
    penc=22366;
    genc=165002;
    order=26;
  }
  compa=strncmp(preset,"26/1/i",6);
  if (compa==0) {
    penc=22324;
    genc=154063;
    order=26;
  }
  compa=strncmp(preset,"25/-1/n",7);
  if (compa==0) {
    penc=22119;
    genc=251243;
    order=25;
  }
  compa=strncmp(preset,"25/-1/i",7);
  if (compa==0) {
    penc=22073;
    genc=240303;
    order=25;
  }
  compa=strncmp(preset,"25/1/n",6);
  if (compa==0) {
    penc=21782;
    genc=165068;
    order=25;
  }
  compa=strncmp(preset,"25/1/i",6);
  if (compa==0) {
    penc=21743;
    genc=154130;
    order=25;
  }
  compa=strncmp(preset,"24/-1/n",7);
  if (compa==0) {
    penc=21527;
    genc=251174;
    order=24;
  }
  compa=strncmp(preset,"24/-1/i",7);
  if (compa==0) {
    penc=21485;
    genc=240235;
    order=24;
  }
  compa=strncmp(preset,"24/1/n",6);
  if (compa==0) {
    penc=21212;
    genc=165141;
    order=24;
  }
  compa=strncmp(preset,"24/1/i",6);
  if (compa==0) {
    penc=21175;
    genc=154205;
    order=24;
  }
  compa=strncmp(preset,"23/-1/n",7);
  if (compa==0) {
    penc=20949;
    genc=251096;
    order=23;
  }
  compa=strncmp(preset,"23/-1/i",7);
  if (compa==0) {
    penc=20908;
    genc=240159;
    order=23;
  }
  compa=strncmp(preset,"23/1/n",6);
  if (compa==0) {
    penc=20653;
    genc=165225;
    order=23;
  }
  compa=strncmp(preset,"23/1/i",6);
  if (compa==0) {
    penc=20618;
    genc=154289;
    order=23;
  }
  compa=strncmp(preset,"22/-1/n",7);
  if (compa==0) {
    penc=20525;
    genc=291140;
    order=22;
  }
  compa=strncmp(preset,"22/-1/i",7);
  if (compa==0) {
    penc=20485;
    genc=280203;
    order=22;
  }
  compa=strncmp(preset,"22/0/n",6);
  if (compa==0) {
    penc=20240;
    genc=209156;
    order=22;
  }
  compa=strncmp(preset,"22/0/i",6);
  if (compa==0) {
    penc=20205;
    genc=198221;
    order=22;
  }
  compa=strncmp(preset,"22/1/n",6);
  if (compa==0) {
    penc=19969;
    genc=119156;
    order=22;
  }
  compa=strncmp(preset,"22/1/i",6);
  if (compa==0) {
    penc=19938;
    genc=108222;
    order=22;
  }
  compa=strncmp(preset,"21/-1/n",7);
  if (compa==0) {
    penc=19957;
    genc=290947;
    order=21;
  }
  compa=strncmp(preset,"21/-1/i",7);
  if (compa==0) {
    penc=19919;
    genc=280011;
    order=21;
  }
  compa=strncmp(preset,"21/0/n",6);
  if (compa==0) {
    penc=19689;
    genc=209156;
    order=21;
  }
  compa=strncmp(preset,"21/0/i",6);
  if (compa==0) {
    penc=19655;
    genc=198221;
    order=21;
  }
  compa=strncmp(preset,"21/1/n",6);
  if (compa==0) {
    penc=19433;
    genc=119387;
    order=21;
  }
  compa=strncmp(preset,"21/1/i",6);
  if (compa==0) {
    penc=19404;
    genc=108453;
    order=21;
  }
  compa=strncmp(preset,"20/-1/n",7);
  if (compa==0) {
    penc=19395;
    genc=290724;
    order=20;
  }
  compa=strncmp(preset,"20/-1/i",7);
  if (compa==0) {
    penc=19359;
    genc=279789;
    order=20;
  }
  compa=strncmp(preset,"20/0/n",6);
  if (compa==0) {
    penc=19142;
    genc=209155;
    order=20;
  }
  compa=strncmp(preset,"20/0/i",6);
  if (compa==0) {
    penc=19111;
    genc=198221;
    order=20;
  }
  compa=strncmp(preset,"20/1/n",6);
  if (compa==0) {
    penc=18900;
    genc=119654;
    order=20;
  }
  compa=strncmp(preset,"20/1/i",6);
  if (compa==0) {
    penc=18873;
    genc=108721;
    order=20;
  }
  compa=strncmp(preset,"19/-1/n",7);
  if (compa==0) {
    penc=18836;
    genc=290464;
    order=19;
  }
  compa=strncmp(preset,"19/-1/i",7);
  if (compa==0) {
    penc=18802;
    genc=279530;
    order=19;
  }
  compa=strncmp(preset,"19/0/n",6);
  if (compa==0) {
    penc=18596;
    genc=209154;
    order=19;
  }
  compa=strncmp(preset,"19/0/i",6);
  if (compa==0) {
    penc=18566;
    genc=198221;
    order=19;
  }
  compa=strncmp(preset,"19/1/n",6);
  if (compa==0) {
    penc=18366;
    genc=119966;
    order=19;
  }
  compa=strncmp(preset,"19/1/i",6);
  if (compa==0) {
    penc=18340;
    genc=109034;
    order=19;
  }
  compa=strncmp(preset,"18/-1/n",7);
  if (compa==0) {
    penc=18275;
    genc=290157;
    order=18;
  }
  compa=strncmp(preset,"18/-1/i",7);
  if (compa==0) {
    penc=18243;
    genc=279223;
    order=18;
  }
  compa=strncmp(preset,"18/0/n",6);
  if (compa==0) {
    penc=18047;
    genc=209154;
    order=18;
  }
  compa=strncmp(preset,"18/0/i",6);
  if (compa==0) {
    penc=18018;
    genc=198221;
    order=18;
  }
  compa=strncmp(preset,"18/1/n",6);
  if (compa==0) {
    penc=17826;
    genc=120335;
    order=18;
  }
  compa=strncmp(preset,"18/1/i",6);
  if (compa==0) {
    penc=17801;
    genc=109403;
    order=18;
  }
  compa=strncmp(preset,"17/-1/n",7);
  if (compa==0) {
    penc=17706;
    genc=289789;
    order=17;
  }
  compa=strncmp(preset,"17/-1/i",7);
  if (compa==0) {
    penc=17675;
    genc=278856;
    order=17;
  }
  compa=strncmp(preset,"17/0/n",6);
  if (compa==0) {
    penc=17487;
    genc=209153;
    order=17;
  }
  compa=strncmp(preset,"17/0/i",6);
  if (compa==0) {
    penc=17459;
    genc=198221;
    order=17;
  }
  compa=strncmp(preset,"17/1/n",6);
  if (compa==0) {
    penc=17274;
    genc=120776;
    order=17;
  }
  compa=strncmp(preset,"17/1/i",6);
  if (compa==0) {
    penc=17250;
    genc=109845;
    order=17;
  }
  compa=strncmp(preset,"16/-1/n",7);
  if (compa==0) {
    penc=17122;
    genc=289342;
    order=16;
  }
  compa=strncmp(preset,"16/-1/i",7);
  if (compa==0) {
    penc=17091;
    genc=278409;
    order=16;
  }
  compa=strncmp(preset,"16/0/n",6);
  if (compa==0) {
    penc=16909;
    genc=209153;
    order=16;
  }
  compa=strncmp(preset,"16/0/i",6);
  if (compa==0) {
    penc=16881;
    genc=198221;
    order=16;
  }
  compa=strncmp(preset,"16/1/n",6);
  if (compa==0) {
    penc=16701;
    genc=121313;
    order=16;
  }
  compa=strncmp(preset,"16/1/i",6);
  if (compa==0) {
    penc=16677;
    genc=110382;
    order=16;
  }
  compa=strncmp(preset,"15/-2/n",7);
  if (compa==0) {
    penc=16618;
    genc=326365;
    order=15;
  }
  compa=strncmp(preset,"15/-2/i",7);
  if (compa==0) {
    penc=16586;
    genc=315432;
    order=15;
  }
  compa=strncmp(preset,"15/-1/n",7);
  if (compa==0) {
    penc=16405;
    genc=249783;
    order=15;
  }
  compa=strncmp(preset,"15/-1/i",7);
  if (compa==0) {
    penc=16377;
    genc=238851;
    order=15;
  }
  compa=strncmp(preset,"15/1/n",6);
  if (compa==0) {
    penc=16198;
    genc=166654;
    order=15;
  }
  compa=strncmp(preset,"15/1/i",6);
  if (compa==0) {
    penc=16172;
    genc=155722;
    order=15;
  }
  compa=strncmp(preset,"15/2/n",6);
  if (compa==0) {
    penc=15994;
    genc=74717;
    order=15;
  }
  compa=strncmp(preset,"15/2/i",6);
  if (compa==0) {
    penc=15971;
    genc=63786;
    order=15;
  }
  compa=strncmp(preset,"14/-2/n",7);
  if (compa==0) {
    penc=15965;
    genc=325353;
    order=14;
  }
  compa=strncmp(preset,"14/-2/i",7);
  if (compa==0) {
    penc=15934;
    genc=314420;
    order=14;
  }
  compa=strncmp(preset,"14/-1/n",7);
  if (compa==0) {
    penc=15754;
    genc=249420;
    order=14;
  }
  compa=strncmp(preset,"14/-1/i",7);
  if (compa==0) {
    penc=15725;
    genc=238488;
    order=14;
  }
  compa=strncmp(preset,"14/1/n",6);
  if (compa==0) {
    penc=15545;
    genc=167051;
    order=14;
  }
  compa=strncmp(preset,"14/1/i",6);
  if (compa==0) {
    penc=15519;
    genc=156120;
    order=14;
  }
  compa=strncmp(preset,"14/2/n",6);
  if (compa==0) {
    penc=15339;
    genc=76051;
    order=14;
  }
  compa=strncmp(preset,"14/2/i",6);
  if (compa==0) {
    penc=15316;
    genc=65119;
    order=14;
  }
  compa=strncmp(preset,"13/-2/n",7);
  if (compa==0) {
    penc=15253;
    genc=324040;
    order=13;
  }
  compa=strncmp(preset,"13/-2/i",7);
  if (compa==0) {
    penc=15221;
    genc=313107;
    order=13;
  }
  compa=strncmp(preset,"13/-1/n",7);
  if (compa==0) {
    penc=15038;
    genc=248949;
    order=13;
  }
  compa=strncmp(preset,"13/-1/i",7);
  if (compa==0) {
    penc=15008;
    genc=238017;
    order=13;
  }
  compa=strncmp(preset,"13/1/n",6);
  if (compa==0) {
    penc=14824;
    genc=167567;
    order=13;
  }
  compa=strncmp(preset,"13/1/i",6);
  if (compa==0) {
    penc=14797;
    genc=156635;
    order=13;
  }
  compa=strncmp(preset,"13/2/n",6);
  if (compa==0) {
    penc=14612;
    genc=77777;
    order=13;
  }
  compa=strncmp(preset,"13/2/i",6);
  if (compa==0) {
    penc=14588;
    genc=66844;
    order=13;
  }
  compa=strncmp(preset,"12/-2/n",7);
  if (compa==0) {
    penc=14453;
    genc=322277;
    order=12;
  }
  compa=strncmp(preset,"12/-2/i",7);
  if (compa==0) {
    penc=14418;
    genc=311341;
    order=12;
  }
  compa=strncmp(preset,"12/-1/n",7);
  if (compa==0) {
    penc=14229;
    genc=248317;
    order=12;
  }
  compa=strncmp(preset,"12/-1/i",7);
  if (compa==0) {
    penc=14197;
    genc=237384;
    order=12;
  }
  compa=strncmp(preset,"12/1/n",6);
  if (compa==0) {
    penc=14005;
    genc=168259;
    order=12;
  }
  compa=strncmp(preset,"12/1/i",6);
  if (compa==0) {
    penc=13976;
    genc=157325;
    order=12;
  }
  compa=strncmp(preset,"12/2/n",6);
  if (compa==0) {
    penc=13782;
    genc=80088;
    order=12;
  }
  compa=strncmp(preset,"12/2/i",6);
  if (compa==0) {
    penc=13756;
    genc=69155;
    order=12;
  }
  compa=strncmp(preset,"11/-2/n",7);
  if (compa==0) {
    penc=13638;
    genc=354175;
    order=11;
  }
  compa=strncmp(preset,"11/-2/i",7);
  if (compa==0) {
    penc=13600;
    genc=343240;
    order=11;
  }
  compa=strncmp(preset,"11/-1/n",7);
  if (compa==0) {
    penc=13400;
    genc=284247;
    order=11;
  }
  compa=strncmp(preset,"11/-1/i",7);
  if (compa==0) {
    penc=13364;
    genc=273311;
    order=11;
  }
  compa=strncmp(preset,"11/0/n",6);
  if (compa==0) {
    penc=13161;
    genc=209156;
    order=11;
  }
  compa=strncmp(preset,"11/0/i",6);
  if (compa==0) {
    penc=13128;
    genc=198221;
    order=11;
  }
  compa=strncmp(preset,"11/1/n",6);
  if (compa==0) {
    penc=12921;
    genc=127400;
    order=11;
  }
  compa=strncmp(preset,"11/1/i",6);
  if (compa==0) {
    penc=12891;
    genc=116466;
    order=11;
  }
  compa=strncmp(preset,"11/2/n",6);
  if (compa==0) {
    penc=12680;
    genc=36597;
    order=11;
  }
  compa=strncmp(preset,"11/2/i",6);
  if (compa==0) {
    penc=12653;
    genc=25662;
    order=11;
  }
  if (order==-1) {
    printf("ERROR\n");
  }
  p_crires->mup1=p_crires->penc_co1+
    p_crires->penc_co2*(float)(penc)+
    p_crires->penc_co3*(float)(penc*penc);
  p_crires->mug=p_crires->genc_co1+
    p_crires->genc_co2*(float)(genc)+
    p_crires->genc_co3*(float)(genc*genc);
  return order;
}
      
/* Hardcoded intermediate slit widths for given order (mm) taken from Jeff's
   table.
   Upon success the slitwidth is returned, upon failure -1.0 */
double crires_model_slitw_tab(int mm)
{
  double slitw;
  slitw=0.0;
  switch(mm) {
  case 59: slitw=1.009; break;
  case 58: slitw=0.984; break;
  case 57: slitw=0.959; break;
  case 56: slitw=0.934; break;
  case 55: slitw=0.910; break;
  case 54: slitw=0.886; break;
  case 53: slitw=0.863; break;
  case 52: slitw=0.840; break;
  case 51: slitw=0.818; break;
  case 50: slitw=0.796; break;
  case 49: slitw=0.774; break;
  case 48: slitw=0.753; break;
  case 47: slitw=0.732; break;
  case 46: slitw=0.711; break;
  case 45: slitw=0.690; break;
  case 44: slitw=0.670; break;
  case 43: slitw=0.632; break;
  case 42: slitw=0.614; break;
  case 41: slitw=0.595; break;
  case 40: slitw=0.577; break;
  case 39: slitw=0.559; break;
  case 38: slitw=0.541; break;
  case 37: slitw=0.614; break;
  case 36: slitw=0.586; break;
  case 35: slitw=0.560; break;
  case 34: slitw=0.534; break;
  case 33: slitw=0.510; break;
  case 32: slitw=0.486; break;
  case 31: slitw=0.464; break;
  case 30: slitw=0.443; break;
  case 29: slitw=0.422; break;
  case 28: slitw=0.403; break;
  case 27: slitw=0.400; break;
  case 26: slitw=0.400; break;
  case 25: slitw=0.400; break;
  case 24: slitw=0.400; break;
  case 23: slitw=0.400; break;
  case 22: slitw=0.400; break;
  case 21: slitw=0.400; break;
  case 20: slitw=0.400; break;
  case 19: slitw=0.400; break;
  case 18: slitw=0.400; break;
  case 17: slitw=0.400; break;
  case 16: slitw=0.400; break;
  case 15: slitw=0.400; break;
  case 14: slitw=0.400; break;
  case 13: slitw=0.400; break;
  case 12: slitw=0.400; break;
  case 11: slitw=0.400; break;
  return -1.0;
  }
  return slitw;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the wavelengths at every pixel centre in the detector array
  @param    sci_file        Science file to get values from the header
  @param    cfg_tab_file    Model configuration file
  @param    order           Order
  @return   CRIRES_NB_DETECTORS cpl_image_list containing wavemaps for each chip
  The returned pointer must be deallocated with:
  cpl_imagelist_delete(returned_ptr);
*/
/*----------------------------------------------------------------------------*/
cpl_imagelist * crires_model_wavpix(
        const char  *   sci_file,
        const char  *   cfg_tab_file,
        int             order)
{
  cpl_imagelist       *   wav_pix_imgl ;
  cpl_image           *   wav_pix_img ;
  float               *   pwav_pix_img ;
  int                     i, j, xpos, ypos ;
  DOUBLE lam_min, lam_max, lamstep ,lambda, pixxlast, pixwavlast,inout,pos1;
  int chipx_current,ii,jj;
  DOUBLE **detec=NULL; /*detector array */
  DOUBLE **znse_ref=NULL;
  struct crires crires_model;
  struct crires *p_crires; /*pointer to crires_model */
  p_crires=&crires_model;
  /* commented out are test fit values */
  //DOUBLE prism_enc_scale=0.0000292966;
  //DOUBLE grat_enc_scale=0.000030504;

  if ((order=crires_model_setup(sci_file,cfg_tab_file,order,p_crires))
      ==-1) {
    cpl_msg_error(__func__, "Failing model setup") ;
    return NULL ;
  }

  if ((znse_ref=crires_model_alloc2Darray(8,7))==NULL){
    cpl_msg_error(__func__, "Cannot allocate 2D array");
    return NULL;
  }
  if (crires_model_ref_ind_read(znse_ref,p_crires->temper)==-1) {
    cpl_msg_error(__func__, "error setting ref index at T= %lf", p_crires->temper);
    return NULL ;
  }

  /* Set slit_on to zero so that even pixels that would not have passed 
     the Intermediate slit will get to the detector (this way the shape
     of the IS is not seen on the wavemap*/
  p_crires->slit_on=0;
  if (crires_model_init(p_crires)!=0) {
    cpl_msg_error(__func__, "Failed to initialise model" );
    return NULL;
  }
  /*compute wavelength range for given configuration,
    order and prism and grating setting*/
  inout=2.0*atan(0.5*p_crires->cam_dis/p_crires->fk);
  lam_min=(sin(p_crires->mug-(inout/2.0))+sin(p_crires->mug+(inout/2.0)))/
    ((p_crires->sg)*((double)(order)+1.0));
  lam_max=(sin(p_crires->mug-(inout/2.0))+sin(p_crires->mug+(inout/2.0)))/
    ((p_crires->sg)*((double)(order)-1.0));
  /* Fix position across slit to centre (width) */
  p_crires->es_y=0.0;

  /* calculate lamstep*/
  lambda=(lam_max+lam_min)/2.0;
  lamstep=(lam_max-lam_min)/(40000.0);
  crires_model_eval(lambda,order,znse_ref, p_crires);
  pos1=p_crires->xdet/p_crires->pix;
  lambda+=lamstep;
  crires_model_eval(lambda,order,znse_ref, p_crires);
  lamstep=lamstep/(2.5*((p_crires->xdet/p_crires->pix)-pos1));

  if ((detec=crires_model_alloc2Darray(ASIZE, BSIZE))==NULL){
    cpl_msg_error(__func__, "Cannot allocate 2D array");
    return NULL;
  }
  for (i=0 ; i<ASIZE ; i++) {
    for (j=0 ; j<BSIZE ; j++) {
      detec[i][j]=0.0;
    }
  }

  for (p_crires->es_x=-p_crires->es_s;
       p_crires->es_x<=p_crires->es_s;
       p_crires->es_x+=0.025) {
    chipx_current=0;
    pixxlast=0.0;
    pixwavlast=0.0;
    /* Iterate over wavelength range */
    for (lambda=lam_min; lambda<lam_max; lambda+=lamstep) {
      crires_model_eval(lambda,order,znse_ref, p_crires);
      crires_model_detpix(p_crires);
      if (p_crires->chippix[0]>-1 && p_crires->chippix[0]<4) {
	/*reset chipx_current when starting new chip */
        if (chipx_current>p_crires->chippix[1]+1) {
          chipx_current=p_crires->chippix[1];
        }
        if (p_crires->xpospix>(double)(chipx_current)+0.5) {
          /* Linear interpolation for value at centre of pixel */
	  xpos=p_crires->chippix[0]*(int)(chipxpix+chipxspace)+
               p_crires->chippix[1];
	  ypos=p_crires->chippix[2];
	  if (xpos>=0 && xpos<ASIZE && ypos>=0 && ypos<BSIZE) {
	    detec[xpos][ypos]=(pixwavlast+(lambda-pixwavlast)*
	       ((double)(chipx_current)+0.5-pixxlast)/
	       (p_crires->xpospix-pixxlast))*1000000.0;
	  }
          chipx_current=p_crires->chippix[1]+1;
        }
        pixxlast=p_crires->xpospix;
        pixwavlast=lambda;
      }
      ii++;
    }
    jj++;
  }

  /* Create the output image list */
  wav_pix_imgl = cpl_imagelist_new() ;
 
  /* Loop on the 4 images */
  for (ii=0 ; ii<4 ; ii++) {
    wav_pix_img = cpl_image_new(ASIZE/4, BSIZE, CPL_TYPE_FLOAT) ;
    pwav_pix_img = cpl_image_get_data_float(wav_pix_img) ;
    for (i=0 ; i<ASIZE/4 ; i++) {
      for (j=0 ; j<BSIZE ; j++) {
	if (detec[(ASIZE/4)*ii+i][j]<10000.0){
	  pwav_pix_img[i+j*ASIZE/4] = (double)detec[(ASIZE/4)*ii+i][j] ;
	}
      }
    }
    cpl_imagelist_set(wav_pix_imgl, wav_pix_img, ii) ;
  }
  if ((crires_model_free2Darray(znse_ref,8))!=0){
    cpl_msg_error(__func__, "Cannot free 2D array");
    return NULL;
  }
  if ((crires_model_free2Darray(detec,ASIZE))!=0){
    cpl_msg_error(__func__, "Cannot free 2D array");
    return NULL;
  }

  return wav_pix_imgl;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the wavelengths at a given position
  @param    sci_file        Science file to get values from the header
  @param    cfg_tab_file    Model configuration file
  @param    order           Order 
  @param    ent_slit_pos    Position on the slit
  @param    wlmap           The wavelength map or NULL
  @return   wavelengths cpl_vectors for wavelengths

  The returned pointer must be deallocated with
  for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) cpl_vector_delete(returned_ptr[i]) ;
  cpl_free(returned_ptr) ;
 */
/*----------------------------------------------------------------------------*/
cpl_vector ** crires_model_wavelengths(
        const char      *   sci_file,
        const char      *   cfg_tab_file,
        int                 order,
        double              ent_slit_pos,
        cpl_imagelist   *   wlmap)
{
    cpl_vector      **      wavelengths ;
    cpl_vector      **      out_traces ;
    cpl_imagelist   *       wl_map_model ;
    cpl_image       *       ima ;     
    double                  wave_val ;
    int                     wave_y ;
    int                     val ;
    int                     i, j;
  
    if (ent_slit_pos < 0)  {
      cpl_msg_error(__func__, "Invalid entrance slit position") ;
      return NULL ; 
    }

    /* First compute the wavelength map */
    if (wlmap == NULL) {
        if ((wl_map_model = crires_model_wavpix(sci_file, cfg_tab_file, 
                        -1))==NULL){
            cpl_msg_error(__func__, "Cannot create wavelength map") ;
            return NULL ;
        }
    } else {
        wl_map_model = wlmap ;
    }
    /* Compute the trace */
    if ((out_traces = crires_model_profile(sci_file, cfg_tab_file,
                    order, ent_slit_pos)) == NULL) {
        cpl_msg_error(__func__, "Cannot get the spectrum trace") ;
        if (wl_map_model != wlmap) cpl_imagelist_delete(wl_map_model) ;
        return NULL ;
    }

    /* Deduce the wavelengths */
    if ((wavelengths=cpl_malloc(CRIRES_NB_DETECTORS*sizeof(cpl_vector *)))
            ==NULL) {
        cpl_msg_error(__func__, "Cannot allocate wavelength array");
        return NULL;
    }
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        wavelengths[i] = cpl_vector_new(1024) ;
        cpl_vector_fill(wavelengths[i], 0.0) ;
    }

    /* Loop on the chips */
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        ima = cpl_imagelist_get(wl_map_model, i) ;
        for (j=0 ; j<cpl_image_get_size_x(ima) ; j++) {
            wave_y = (int)(cpl_vector_get(out_traces[i], j)) ;
            /* To handle ZEROS in the profile */
            if (wave_y == 0) wave_y = (int)ent_slit_pos ;
            wave_val = cpl_image_get(ima, j+1, wave_y+1, &val) ;
            cpl_vector_set(wavelengths[i], j, wave_val);
        }
    }

    if (wl_map_model != wlmap) cpl_imagelist_delete(wl_map_model) ;
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) cpl_vector_delete(out_traces[i]) ;
    cpl_free(out_traces) ;
  
    return wavelengths ;
}


/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the trace of the spectrum 
  @param    sci_file        Science file to get values from the header
  @param    cfg_tab_file    Model configuration file
  @param    order           Order 
  @param    ent_slit_pos    Position on the slit
  @return   CRIRES_NB_DETECTORS cpl_vectors for the y positions

  The returned pointer must be deallocated with
  for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) cpl_vector_delete(returned_ptr[i]) ;
  cpl_free(returned_ptr) ;
 */
/*----------------------------------------------------------------------------*/
cpl_vector ** crires_model_profile(
        const char  *   sci_file,
        const char  *   cfg_tab_file,
        int             order,
        double          ent_slit_pos)
{
  cpl_propertylist    *   plist ;
  int                     i, j , warn,is_on_chip_flag;
  DOUBLE lam_min, lam_max, lamstep ,lambda, lamref, pixxlast, pixylast, inout,pos1;
  DOUBLE sens, es_x_old, yold, conv_test;
  int chipx_current,ii;
  double *trace=NULL; /*trace array */
  DOUBLE **znse_ref=NULL;
  cpl_vector        **      out_traces ;
  double                    val ;
  struct crires crires_model;
  struct crires *p_crires; /*pointer to crires_model */
  p_crires=&crires_model;

  if ((order=crires_model_setup(sci_file,cfg_tab_file,order,p_crires))
      ==-1) {
    cpl_msg_error(__func__, "Failing model setup") ;
    return NULL ;
  }
  /* Check entries */
  if (sci_file == NULL) {
    cpl_msg_error(__func__, "Cannot find %s ", sci_file) ;
    return NULL ;
  }

  /* Read the header informations */
  if ((plist=cpl_propertylist_load(sci_file, 0)) == NULL) {
    cpl_msg_error(__func__, "getting header from file %s", sci_file);
    cpl_propertylist_delete(plist) ;
    return NULL ;
  }

  /*ref wvln is average of chip3 start and end wvln in header, div to get mm*/ 
  lamref=(crires_pfits_get_wlens3(plist)+crires_pfits_get_wlene3(plist))/2000000.0;
  if (cpl_error_get_code() != CPL_ERROR_NONE) {
    cpl_msg_error(__func__, "Cannot find detector wavelength limits in header");
    cpl_propertylist_delete(plist) ;
    return NULL;
  }
  cpl_propertylist_delete(plist) ;

  if ((znse_ref=crires_model_alloc2Darray(8,7))==NULL){
    cpl_msg_error(__func__, "Cannot allocate 2D array");
    return NULL;
  }
  if (crires_model_ref_ind_read(znse_ref,p_crires->temper)==-1) {
    cpl_msg_error(__func__, "error setting ref index at T= %lf", p_crires->temper);
    if ((crires_model_free2Darray(znse_ref,8))!=0) {
      cpl_msg_error(__func__, "Cannot free 2D array");
      return NULL;
    }
    return NULL ;
  }

  if (crires_model_init(p_crires)!=0) {
    cpl_msg_error(__func__, "Failed to initialise model" );
    if ((crires_model_free2Darray(znse_ref,8))!=0) {
      cpl_msg_error(__func__, "Cannot free 2D array");
      return NULL;
    }
    return NULL;
  }

  /*compute wavelength range for given configuration,
    order and prism and grating setting*/
  inout=2.0*atan(0.5*p_crires->cam_dis/p_crires->fk);
  lam_min=(sin(p_crires->mug-(inout/2.0))+sin(p_crires->mug+(inout/2.0)))/
    ((p_crires->sg)*((double)(order)+1.0));
  lam_max=(sin(p_crires->mug-(inout/2.0))+sin(p_crires->mug+(inout/2.0)))/
    ((p_crires->sg)*((double)(order)-1.0));

  /* Fix position across slit to centre (width) */
  p_crires->es_y=0.0;

  p_crires->slit_on=0;

  /*Below was used for full slit wavecals to determine a locus on the detector that
    passed through a given y_posn at the centre of chip 3 by determining what the
    entrance slit posn would have to be for chippix[0]=2, xpospix=512, ypospix=y_posn.
    Now, the refine recipe assumes the fixed fibre positions, however it is still used
    for crires_model_fix*/
  conv_test=100.0;
  p_crires->es_x=0.0;
  if (crires_model_init(p_crires)!=0) {
    cpl_msg_error(__func__, "Failed to initialise model" );
    if ((crires_model_free2Darray(znse_ref,8))!=0) {
      cpl_msg_error(__func__, "Cannot free 2D array");
      return NULL;
    }
    return NULL;
  }
  sens=50.0;
  crires_model_eval(lamref,order,znse_ref, p_crires);
  crires_model_detpix(p_crires);
  es_x_old=p_crires->es_x;
  yold=p_crires->ypospix;
  p_crires->es_x=1.0;
  ii=0;
  warn=0;

  if (ent_slit_pos>0.0) {
    while (conv_test>0.1) {
      if (crires_model_init(p_crires)!=0) {
	cpl_msg_error(__func__, "Failed to initialise model" );
	if ((crires_model_free2Darray(znse_ref,8))!=0) {
	  cpl_msg_error(__func__, "Cannot free 2D array");
	  return NULL;
	}
	return NULL;
      }
      p_crires->slit[1]=10.0;
      crires_model_eval(lamref,order,znse_ref, p_crires);
      crires_model_detpix(p_crires);
      if (p_crires->chippix[0]==2) {
	sens=(p_crires->ypospix-yold)/(p_crires->es_x-es_x_old);
	conv_test=(p_crires->ypospix-ent_slit_pos);
	es_x_old=p_crires->es_x;
	yold=p_crires->ypospix;
	p_crires->es_x-=(p_crires->ypospix-ent_slit_pos)/sens;
      }
      else {
	if (p_crires->chippix[0]==7) {
	  /*Off the top of the chip*/
	  p_crires->es_x-=0.05;
	}
	else if (p_crires->chippix[0]==8) {
	  /*Off the bottom of the chip*/
	  p_crires->es_x+=0.05;
	}
	else if (p_crires->chippix[0]==3) {
	  if (warn!=1) {
	    cpl_msg_warning(__func__, "Large discrepancy between model and reference wavelengths a");
	    warn=1;
	  }
	  lamref-=0.000001;
	}
	else if (p_crires->chippix[0]==1) {
	  if (warn!=1) {
	    cpl_msg_warning(__func__, "Large discrepancy between model and reference wavelengths b");
	    warn=1;
	  }
	  lamref+=0.000002;
	}
	else if (p_crires->chippix[0]==4) {
	  if (warn!=1) {
	    cpl_msg_warning(__func__, "Large discrepancy between model and reference wavelengths c");
	    warn=1;
	  }
	  lamref-=0.000001;
	}

      }
      ii+=1;
      if (ii>10000) {
	cpl_msg_error(__func__, "Cannot determine the slit position");
	if ((crires_model_free2Darray(znse_ref,8))!=0) {
	  cpl_msg_error(__func__, "Cannot free 2D array");
	  return NULL;
	}
	return NULL;
      }
    }
    cpl_msg_info(__func__, "slit position: %lfmm", p_crires->es_x);
  }
  else {
    p_crires->es_x=p_crires->slit[(int)(-ent_slit_pos)];
  }

  /* calculate lamstep*/
  lambda=(lam_max+lam_min)/2.0;
  lamstep=(lam_max-lam_min)/(40000.0);
  crires_model_eval(lambda,order,znse_ref, p_crires);
  pos1=p_crires->xdet/p_crires->pix;
  lambda+=lamstep;
  crires_model_eval(lambda,order,znse_ref, p_crires);
  lamstep=lamstep/(2.5*((p_crires->xdet/p_crires->pix)-pos1));

  p_crires->slit_on=0;

  if ((trace=crires_model_alloc1Darray(ASIZE))==NULL){
    cpl_msg_error(__func__, "Cannot allocate 2D array");
    if ((crires_model_free2Darray(znse_ref,8))!=0) {
      cpl_msg_error(__func__, "Cannot free 2D array");
      return NULL;
    }
    return NULL;
  }

  chipx_current=0;
  pixxlast=0.0;
  pixylast=0.0;

  /* Create output vectors */
  if ((out_traces=cpl_malloc(CRIRES_NB_DETECTORS*sizeof(cpl_vector*)))==NULL) {
    cpl_msg_error(__func__, "Cannot allocate trace array");
    if ((crires_model_free2Darray(znse_ref,8))!=0) {
      cpl_msg_error(__func__, "Cannot free 2D array");
      return NULL;
    }
    return NULL;
  }

  for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
      out_traces[i] = cpl_vector_new(1024+1) ;
      /*the extra 1 posn in the vector is to hold the entrance slit posn
	in mm that corresponds to the pixel posn at centre of chip 3, as
	computed below*/
      cpl_vector_fill(out_traces[i], 0.0) ;
      cpl_vector_set(out_traces[i], 1024, p_crires->es_x) ;
  }

  is_on_chip_flag=0;
  for (lambda=lam_min; lambda<lam_max; lambda+=lamstep) {
    crires_model_eval(lambda,order,znse_ref, p_crires);
    crires_model_detpix(p_crires);
    if (p_crires->chippix[0]>-1 && p_crires->chippix[0]<4) {
      is_on_chip_flag=1;
      /*reset chipx_current when starting new chip */
      if (chipx_current>p_crires->chippix[1]+1) {
	chipx_current=p_crires->chippix[1];
      }
//trace is 1d, 4096 long, i.e. chips are concatenated
      trace[p_crires->chippix[0]*(int)(chipxpix+chipxspace)+
	   p_crires->chippix[1]]=pixylast+(p_crires->ypospix-pixylast)*
	((double)(chipx_current)+0.5-pixxlast)/(p_crires->xpospix-pixxlast);
      chipx_current=p_crires->chippix[1]+1;
    }
    pixxlast=p_crires->xpospix;
    pixylast=p_crires->ypospix;
    if (p_crires->chippix[1]>0) {
      val=trace[p_crires->chippix[0]*(int)(chipxpix+chipxspace)+
		p_crires->chippix[1]];
      i = p_crires->chippix[0] ;
      j = p_crires->chippix[1] ;
      if (i>=0 && i<CRIRES_NB_DETECTORS && j>0 && j<=1024) {
	cpl_vector_set(out_traces[i], j-1, val) ;
      }
    }
  }
  if (is_on_chip_flag==0) {
    cpl_msg_warning(__func__, "Can't compute profile (invalid slit position?)");
    cpl_free(trace);
    if ((crires_model_free2Darray(znse_ref,8))!=0) {
      cpl_msg_error(__func__, "Cannot free 2D array");
      return NULL;
    }
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) cpl_vector_delete(out_traces[i]) ;
    cpl_free(out_traces);
    return NULL;
  }

  cpl_free(trace);
  if ((crires_model_free2Darray(znse_ref,8))!=0) {
    cpl_msg_error(__func__, "Cannot free 2D array");
    return NULL;
  }

  return out_traces ;
}


/*The number of measured co-ordinates must equal the number of wavelengths in meas_coord_file (and they should also correspond)*/
int qget_meas_coordinates(int sizearraywavelengths,
                          coord* meas_coord,
                          char* datname)
{
  int feat_cnt=0;

  FILE* measfile=fopen(datname,"r");
  /*  FILE * matchfile=fopen("match.dat","r");*/
  if(measfile ==NULL ){
    cpl_msg_error(__func__, "Cannot open the measured coords file");
    return -1;
  }

  while(fscanf(measfile, "%d %lf %lf %d %lf %d %lf %d %d %d\n",
               &meas_coord[feat_cnt].counter,
               &meas_coord[feat_cnt].y,
               &meas_coord[feat_cnt].x,
               &meas_coord[feat_cnt].chip,
               &meas_coord[feat_cnt].flux,
               &meas_coord[feat_cnt].slit_pos,
               &meas_coord[feat_cnt].wave,
               &meas_coord[feat_cnt].mode,
               &meas_coord[feat_cnt].penc,
               &meas_coord[feat_cnt].genc) != EOF ){
    feat_cnt+=1;
  }

  if (feat_cnt != sizearraywavelengths) {
    cpl_msg_error(__func__, "Number of spectral features is not equal to the number of wavelengths and slit posn's supplied! \n %d != %d \n",feat_cnt, sizearraywavelengths);
    return -1;
  }
  return 0;
}

/*count lines of the calibration files in order to dimension the array's size.*/ 
int qcountlines (const char* filename)
  {

  /* count  lines  ignore Comments (lines start with a %)*/
  char oneline[200];
  char* c;
  char firstchar[]="";
  char comment[]="%";
  FILE* myfile;
  int linecounter=0;
  if((myfile=fopen((char*)filename,"r"))==NULL){
    cpl_msg_error(__func__, "Cannot open file %s for reading.\n",filename);
    return -1;
  }

  do
   {
      c = fgets(oneline, 200, myfile);  /* get one line from the file */
      if (c != NULL)
      {
       strncpy(firstchar,oneline,1);
       if(strncmp(firstchar,comment,1))
         /*if is not a comment increment linecounter*/
          linecounter++;
      }

   }
   while (c != NULL);               /* repeat until NULL          */
   fclose(myfile);
   return linecounter;
  }

int crires_model_setup(
        const char  *   sci_file,
        const char  *   cfg_tab_file,
        int             order,
	struct crires * p_crires)
{
  cpl_propertylist    *   plist ;
  DOUBLE temper;
  double grat_enc, prism_enc;
  /* Check entries */
  if (sci_file == NULL) {
    cpl_msg_error(__func__, "Cannot find sci_file") ;
    return -1 ;
  }
  if (cfg_tab_file== NULL) {
    cpl_msg_error(__func__, "Cannot find %s ", cfg_tab_file) ;
    return -1 ;
  }
  /* Load the config */
  if (crires_model_io_load_config_best(cfg_tab_file, p_crires) == -1) {
    cpl_msg_error(__func__, "Cannot load %s as a config", cfg_tab_file) ;
    return -1 ;
  }
  /* Read the header informations */
  if ((plist=cpl_propertylist_load(sci_file, 0)) == NULL) {
    cpl_msg_error(__func__, "getting header from file %s", sci_file);
    cpl_propertylist_delete(plist) ;
    return -1 ;
  }
  if (order < 0) {
    order = crires_pfits_get_order(plist) ;
    if (order==0) {
      if (cal_flag==0) {
	cpl_msg_error(__func__, "Cannot find order in header");
	cpl_propertylist_delete(plist) ;
	return -1;
      } else {
	printf("WARNING, FIXING FAULTY HEADER ORDER VALUE, SET cal_flag TO 0 IN crires_model_kernel.h FOR ROUTINE OPERATIONS!!!\n");
	cpl_msg_warning(__func__, "Order determined from prism setting");
	/*temporary fix for scan reduction because orders are missing from headers*/
	order=temp_order_fn(crires_pfits_get_prism_enc(plist),crires_pfits_get_grat_enc(plist));
      }
    }
  }
  temper=crires_pfits_get_temp_prism(plist);
  grat_enc=(double)(crires_pfits_get_grat_enc(plist)-200000);
  prism_enc=(double)(crires_pfits_get_prism_enc(plist)-36000);
  p_crires->slitw=crires_pfits_get_slit2_pos(plist);
  cpl_propertylist_delete(plist) ;
  if (cpl_error_get_code() != CPL_ERROR_NONE) {
      cpl_msg_error(__func__, "Cannot read keywords from the header") ;
      return -1 ;  
  }
  //printf("in setup function, grat: %lf prism: %lf temper: %lf order %d \n",grat_enc,prism_enc,temper,order);
  p_crires->mup1=p_crires->penc_co1+
    p_crires->penc_co2*(prism_enc)+
    p_crires->penc_co3*(prism_enc*prism_enc);
  p_crires->mug=p_crires->genc_co1+
    p_crires->genc_co2*(grat_enc)+
    p_crires->genc_co3*(grat_enc*grat_enc);
  //  printf("grat angle: %lf prism angle: %lf\n",p_crires->mug/DEG2RAD,p_crires->mup1/DEG2RAD);
  p_crires->temper=temper;
  return order;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Use the model to predict line positions for given setup
  @param    sci_file        the sci/cal file with header info (prism,grat etc)
  @param    wave_eg_file    FITS line list filename
  @param    cfg_tab_file    FITS model config file
  @param    morder          Spectral order (-1 means look in header of sci file)
  @return   lines           2D array: 6 properties, index

The returned pointers must be deallocated with:
  crires_model_free2Darray(lines,6)

 */
/*----------------------------------------------------------------------------*/
double ** crires_model_predict(
        const char  *   sci_file,
        const char  *   wave_eg_file,
        const char  *   cfg_tab_file,
        int             morder)
{
  DOUBLE ** znse_ref;
  DOUBLE ** lines;
  cpl_table *   lines_tab ;
  int line_no, ll, lines_number;
  DOUBLE lambda, lambda_nan, inten;
  struct crires crires_model;
  struct crires *p_crires; /*pointer to crires_model */
  p_crires=&crires_model;

  if ((lines=crires_model_alloc2Darray(6,200))==NULL) {
    cpl_msg_error(__func__, "Cannot allocate 2D array");
    return NULL;
  }
  for (line_no=0;line_no<200;line_no+=1) {
    lines[0][line_no]=0.0;
    lines[1][line_no]=0.0;
    lines[2][line_no]=0.0;
    lines[3][line_no]=0.0;
    lines[4][line_no]=0.0;
    lines[5][line_no]=0.0;
  }
  if ((morder=crires_model_setup(sci_file,cfg_tab_file,morder,p_crires))
      ==-1) {
    cpl_msg_error(__func__, "Failing model setup") ;
    return NULL ;
  }

  if ((znse_ref=crires_model_alloc2Darray(8,7))==NULL) {
    cpl_msg_error(__func__, "Cannot allocate 2D array");
    return NULL;
  };
  /*open and read refractive index data file */
  if (crires_model_ref_ind_read(znse_ref,p_crires->temper)==-1) {
    cpl_msg_error(__func__, "error setting ref index at T= %lf", p_crires->temper);
    return NULL ;
  }

  p_crires->es_x=0.0;
  p_crires->es_y=0.0;
  p_crires->slit_on=0;
  /*compute wavelength range for given configuration,
    order and prism and grating setting*/
  /*double inout, lam_min;
  inout=2.0*atan(0.5*p_crires->cam_dis/p_crires->fk);
  lam_min=(sin(p_crires->mug-(inout/2.0))+sin(p_crires->mug+(inout/2.0)))/
  ((p_crires->sg)*((double)(morder)+1.0));
  */

  if (crires_model_init(p_crires)!=0) {
    cpl_msg_error(__func__, "Failed to initialise model" );
    return NULL;
  }
  /* Read the input FITS lines list */
  lines_tab = cpl_table_load(wave_eg_file, 1, 0) ;
  lines_number = cpl_table_get_nrow(lines_tab) ;
  line_no=0;
  /* Loop on the lines and  call the crires model routine */
  for (ll=0 ; ll<lines_number ; ll++) {
    lambda_nan=cpl_table_get_double(lines_tab, CRIRES_COL_WAVELENGTH, ll, NULL);
    inten = cpl_table_get_double(lines_tab, CRIRES_COL_EMISSION, ll, NULL);

    lambda=lambda_nan*1e-6; /*lambda in mm */

    crires_model_eval(lambda,morder,znse_ref, p_crires);

    crires_model_detpix(p_crires);

    if (p_crires->chippix[0]>-1 && p_crires->chippix[0]<4) {
      lines[0][line_no]=lambda_nan;
      lines[1][line_no]=(double)(morder);
      lines[2][line_no]=inten;
      lines[3][line_no]=(double)(p_crires->chippix[0]);
      lines[4][line_no]=p_crires->xpospix;
      lines[5][line_no]=p_crires->ypospix;
      //printf("%lf %lf %lf %lf %lf %lf \n",lines[0][line_no],lines[1][line_no],lines[2][line_no],lines[3][line_no],lines[4][line_no],lines[5][line_no]);
      line_no+=1;
    }
  }
  if ((crires_model_free2Darray(znse_ref,8))!=0){
    cpl_msg_error(__func__, "Cannot free 2D array");
    return NULL;
  }
  return lines;
}


int temp_order_fn(int penc,
		  int genc)
{
  int order;
  if (penc+(170000-genc)/400>50500) order=58;
  if (penc+(170000-genc)/400>49100 && penc+(170000-genc)/400<50500) order=57;
  if (penc+(170000-genc)/400>47850 && penc+(170000-genc)/400<49100) order=56;
  if (penc+(170000-genc)/400>46450 && penc+(170000-genc)/400<47850) order=55;
  if (penc+(170000-genc)/400>45100 && penc+(170000-genc)/400<46450) order=54;
  if (penc+(170000-genc)/400>43950 && penc+(170000-genc)/400<45100) order=53;
  if (penc+(170000-genc)/400>43000 && penc+(170000-genc)/400<43950) order=52;
  if (penc+(170000-genc)/400>41850 && penc+(170000-genc)/400<43000) order=51;
  if (penc+(170000-genc)/400>40800 && penc+(170000-genc)/400<41850) order=50;
  if (penc+(170000-genc)/400>39700 && penc+(170000-genc)/400<40800) order=49;
  if (penc+(170000-genc)/400>38700 && penc+(170000-genc)/400<39700) order=48;
  if (penc+(170000-genc)/400>37700 && penc+(170000-genc)/400<38700) order=47;
  if (penc+(170000-genc)/400>36600 && penc+(170000-genc)/400<37700) order=46;
  if (penc+(170000-genc)/400>35650 && penc+(170000-genc)/400<36600) order=45;
  if (penc+(170000-genc)/400>34850 && penc+(170000-genc)/400<35650) order=44;
  if (penc+(170000-genc)/400>33800 && penc+(170000-genc)/400<34850) order=43;
  if (penc+(170000-genc)/400>32850 && penc+(170000-genc)/400<33800) order=42;
  if (penc+(170000-genc)/400>31950 && penc+(170000-genc)/400<32850) order=41;
  if (penc+(170000-genc)/400>31070 && penc+(170000-genc)/400<31950) order=40;
  if (penc+(170000-genc)/400>30230 && penc+(170000-genc)/400<31070) order=39;
  if (penc+(170000-genc)/400>29400 && penc+(170000-genc)/400<30230) order=38;
  if (penc+(170000-genc)/400>28600 && penc+(170000-genc)/400<29400) order=37;
  if (penc+(170000-genc)/400>27870 && penc+(170000-genc)/400<28600) order=36;
  if (penc+(170000-genc)/400>27090 && penc+(170000-genc)/400<27870) order=35;
  if (penc+(170000-genc)/400>26275 && penc+(170000-genc)/400<27090) order=34;
  if (penc+(170000-genc)/400>25550 && penc+(170000-genc)/400<26275) order=33;
  if (penc+(170000-genc)/400>24830 && penc+(170000-genc)/400<25550) order=32;
  if (penc+(170000-genc)/400>24160 && penc+(170000-genc)/400<24830) order=31;
  if (penc+(170000-genc)/400>23505 && penc+(170000-genc)/400<24160) order=30;
  if (penc+(170000-genc)/400>22865 && penc+(170000-genc)/400<23505) order=29;
  if (penc+(170000-genc)/400>22255 && penc+(170000-genc)/400<22865) order=28;
  if (penc+(170000-genc)/400>21655 && penc+(170000-genc)/400<22255) order=27;
  if (penc+(170000-genc)/400>21080 && penc+(170000-genc)/400<21655) order=26;
  if (penc+(170000-genc)/400>20525 && penc+(170000-genc)/400<21080) order=25;
  if (penc+(170000-genc)/400>19985 && penc+(170000-genc)/400<20525) order=24;
  if (penc+(170000-genc)/400>19490 && penc+(170000-genc)/400<19985) order=23;
  if (penc+(170000-genc)/400>18930 && penc+(170000-genc)/400<19490) order=22;
  if (penc+(170000-genc)/400>18315 && penc+(170000-genc)/400<18930) order=21;
  return order;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Check that the config is valid for the raw file
  @param    cri_config_file        the config file
  @param    raw_file               the raw file with header info
  @return   0 for valid, 1 for invalid, -1 for other errors (eg. files are
            missing)

 */
/*----------------------------------------------------------------------------*/

int crires_model_config_check(const char*   cri_config_file,
			      const char*   raw_file)
{
    cpl_propertylist    *   plist ;
    cpl_propertylist    *   plist_cfg ;
    double                  mjd_obs ;
    const char          *   cfg_id ;
    double                  val_start=0.0;
    double                  val_end=0.0;
 
    /* Check entries */
  
    if (raw_file == NULL) {
        cpl_msg_error(__func__, "Cannot find raw file %s", raw_file) ;
        return -1 ;
    }
    if (cri_config_file== NULL) {
        cpl_msg_error(__func__, "Cannot find %s ", cri_config_file) ;
        return -1 ;
    }

    /* Read the raw header informations */
    if ((plist=cpl_propertylist_load(raw_file, 0)) == NULL) {
        cpl_msg_error(__func__, "getting header from raw file %s", raw_file);
        return -1 ;
    }

    /* Read the config header informations */
    if ((plist_cfg=cpl_propertylist_load(cri_config_file, 0)) == NULL) {
        cpl_msg_error(__func__, "getting header from config file %s", 
                cri_config_file);
        cpl_propertylist_delete(plist) ;
        return -1 ;
    }
    mjd_obs=crires_pfits_get_mjd_obs(plist);
    cpl_propertylist_delete(plist) ;
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_error_reset() ;
        cpl_msg_error(__func__, "Missing MJD-OBS in %s", raw_file) ;
        return -1 ;
    }
  
    /* 
     * cfg_id is a unique identifier for each default config (it may be that
     * the config being used is a refinement of the default config, but it should
     *  still carry the identifier of the default from which it was derived).
     */
    cfg_id=crires_pfits_get_cfg_ident(plist_cfg);

    /* validity periods for updated configs are hard coded below in if table
        structure. Any new config should have an end date of mjd=99999*/
    if (cfg_id == NULL) {
        cpl_propertylist_delete(plist_cfg) ;
        cpl_error_reset() ;
        cpl_msg_error(__func__, "Missing CFG-IDENT in %s", cri_config_file) ;
        return -1 ;
    }
    if (!strcmp(cfg_id, "AlwaysOK")) {
        cpl_propertylist_delete(plist_cfg) ;
        return 0 ;
    } else if (!strcmp(cfg_id, "Apr 2010")) {
        val_start = 55185.0 ; 
        val_end = 55446.0 ;
    } else if (!strcmp(cfg_id, "Nov 2010")) {
        val_start = 55446.0 ; 
        val_end = 55640.0;
    } else if (!strcmp(cfg_id, "May 2011")) {
        val_start = 55641.0 ; 
        val_end = 56190.0;
    } else if (!strcmp(cfg_id, "Sep 2012")) {
        val_start = 56191.0 ; 
        val_end = 99999.0;
    } else {
        cpl_msg_error(__func__, 
                "Invalid physical model configuration identifier in %s. The config file is not compatible with this version of the pipeline", 
                cri_config_file) ;
        cpl_propertylist_delete(plist_cfg) ;
        return -1;
    }
    cpl_propertylist_delete(plist_cfg) ;
    
    /* Return  */
    if (mjd_obs<val_start || mjd_obs>val_end) {
        cpl_msg_error(__func__, 
                "The cfg file %s is not compatible with the science file %s",
                cri_config_file, raw_file);
        return 1 ;
    } else {
        return 0 ;
    }
}
