/* $Id: crires_pfits.c,v 1.33 2012-09-20 12:06:34 yjung Exp $
 *
 * This file is part of the CRIRES Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: yjung $
 * $Date: 2012-09-20 12:06:34 $
 * $Revision: 1.33 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
   								Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include <string.h>

#include "crires_pfits.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup crires_pfits     FITS header protected access
 *
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*-----------------------------------------------------------------------------
  							Function codes
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the PRO.CATG
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * crires_pfits_get_procatg(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist, CPL_DFS_PRO_CATG);
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the PRO.TYPE
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * crires_pfits_get_protype(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist, CPL_DFS_PRO_TYPE);
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the position angle
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double crires_pfits_get_drot_angle(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO INS1 DROT POSANG") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the temperature of the prism
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double crires_pfits_get_temp_prism(const cpl_propertylist * plist)
{
    const char  *   sval ;

    sval = cpl_propertylist_get_string(plist, "ESO INS TEMP30 NAME") ;
    if ((sval != NULL) && (!strcmp(sval, "Prism"))) {
        return cpl_propertylist_get_double(plist, "ESO INS TEMP30 VAL") ;
    } else {
        sval = cpl_propertylist_get_string(plist, "ESO INS TEMP34 NAME") ;
        if ((sval != NULL) && (!strcmp(sval, "Prism"))) {
            return cpl_propertylist_get_double(plist, "ESO INS TEMP34 VAL") ;
        } else {
            return -1.0 ;
        }
    } 
    return -1.0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the baffle position
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double crires_pfits_get_bafflepos(const cpl_propertylist * plist)
{
    double  val ;
    val = cpl_propertylist_get_double(plist, "ESO INS BAFFL POS") ;
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_error_reset() ;
        val = cpl_propertylist_get_double(plist, "ESO INS1 SHUT1 POS") ;
    }
    return val ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the slit2 position
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double crires_pfits_get_slit2_pos(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO INS SLIT2 POS") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the DEC
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double crires_pfits_get_dec(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "DEC") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the RA
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double crires_pfits_get_ra(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "RA") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the arcfile   
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * crires_pfits_get_arcfile(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist, "ARCFILE");
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the pipefile
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * crires_pfits_get_pipefile(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist, "PIPEFILE");
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the date of observation 
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * crires_pfits_get_date_obs(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist, "DATE-OBS") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the MJD of observation 
  @param    plist       property list to read from
  @return   double
 */
/*----------------------------------------------------------------------------*/
double crires_pfits_get_mjd_obs(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "MJD-OBS") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Unique identifier of default config
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * crires_pfits_get_cfg_ident(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist, "ESO PRO CFG-IDENT") ;
}


/*----------------------------------------------------------------------------*/
/**
  @brief    find out the reference wavelength
  @param    plist       property list to read from
  @return   the value
 */
/*----------------------------------------------------------------------------*/
double crires_pfits_get_refwlen(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO INS WLEN REF");
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the wavelength MIN
  @param    plist       property list to read from
  @return   the value
 */
/*----------------------------------------------------------------------------*/
double crires_pfits_get_wlen_min(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO INS WLEN MIN");
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the wavelength MAX
  @param    plist       property list to read from
  @return   the value
 */
/*----------------------------------------------------------------------------*/
double crires_pfits_get_wlen_max(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO INS WLEN MAX");
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the wavelength start of the current detector
  @param    plist       property list to read from
  @return   the value
 */
/*----------------------------------------------------------------------------*/
double crires_pfits_get_wlens(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO INS WLEN STRT");
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the wavelength start of detector 1
  @param    plist       property list to read from
  @return   the value
 */
/*----------------------------------------------------------------------------*/
double crires_pfits_get_wlens1(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO INS WLEN STRT1");
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the wavelength start of detector 2
  @param    plist       property list to read from
  @return   the value 
 */
/*----------------------------------------------------------------------------*/
double crires_pfits_get_wlens2(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO INS WLEN STRT2");
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the wavelength start of detector 3
  @param    plist       property list to read from
  @return   the value
 */
/*----------------------------------------------------------------------------*/
double crires_pfits_get_wlens3(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO INS WLEN STRT3");
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the wavelength start of detector 4
  @param    plist       property list to read from
  @return   the value
 */
/*----------------------------------------------------------------------------*/
double crires_pfits_get_wlens4(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO INS WLEN STRT4");
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the wavelength end of the current detector
  @param    plist       property list to read from
  @return   the value
 */
/*----------------------------------------------------------------------------*/
double crires_pfits_get_wlene(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO INS WLEN END");
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the wavelength end of detector 1
  @param    plist       property list to read from
  @return   the value
 */
/*----------------------------------------------------------------------------*/
double crires_pfits_get_wlene1(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO INS WLEN END1");
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the wavelength end of detector 2
  @param    plist       property list to read from
  @return   the value
 */
/*----------------------------------------------------------------------------*/
double crires_pfits_get_wlene2(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO INS WLEN END2");
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the wavelength end of detector 3
  @param    plist       property list to read from
  @return   the value
 */
/*----------------------------------------------------------------------------*/
double crires_pfits_get_wlene3(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO INS WLEN END3");
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the wavelength end of detector 4
  @param    plist       property list to read from
  @return   the value
 */
/*----------------------------------------------------------------------------*/
double crires_pfits_get_wlene4(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO INS WLEN END4");
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the PINHF.POS value 
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double crires_pfits_get_pinhf_pos(const cpl_propertylist * plist)
{
    double  val ;
    val = cpl_propertylist_get_double(plist, "ESO INS PINHF POS") ;
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_error_reset() ;
        val = cpl_propertylist_get_double(plist, "ESO INS1 FOCU1 POS") ;
    }
    return val ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the PRISM.ENC value 
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int crires_pfits_get_prism_enc(const cpl_propertylist * plist)
{
    int     val ;
    val = cpl_propertylist_get_int(plist, "ESO INS PRISM ENC") ;
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_error_reset() ;
        val = cpl_propertylist_get_int(plist, "ESO INS PRIS ENC") ;
    }
    return val ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the PRO DATANCOM value 
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int crires_pfits_get_datancom(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_int(plist, "ESO PRO DATANCOM")  ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the QC SPECPOS value
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int crires_pfits_get_specpos(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_int(plist, "ESO QC SPECPOS")  ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the ORDER value 
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int crires_pfits_get_order(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_int(plist, "ESO INS GRAT ORDER")  ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the GRAT.ENC value 
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int crires_pfits_get_grat_enc(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_int(plist, "ESO INS GRAT ENC")  ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the SLIT2.ENC value 
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int crires_pfits_get_slit2_enc(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_int(plist, "ESO INS SLIT2 ENC")  ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the CARRG.ENC value 
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int crires_pfits_get_carrg_enc(const cpl_propertylist * plist)
{
    int     val ;
    val = cpl_propertylist_get_int(plist, "ESO INS CARRG ENC") ;
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_error_reset() ;
        val = cpl_propertylist_get_int(plist, "ESO INS1 OPTI1 ENC") ;
    }
    return val ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the EXPNO value 
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int crires_pfits_get_expno(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_int(plist, "ESO TPL EXPNO")  ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the NAXIS2 value 
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int crires_pfits_get_naxis2(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_int(plist, "NAXIS2")  ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the NEXP value 
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int crires_pfits_get_nexp(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_int(plist, "ESO TPL NEXP")  ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the STRIPE CENTERY value 
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int crires_pfits_get_stripe_centery(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_int(plist, "ESO DET STRIPE CENTERY")  ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the STRIPE OFFSETY value 
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int crires_pfits_get_stripe_offsety(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_int(plist, "ESO DET STRIPE OFFSETY")  ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the NDIT value 
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int crires_pfits_get_ndit(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_int(plist, "ESO DET NDIT")  ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the DIT value 
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double crires_pfits_get_dit(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO DET DIT")  ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the cumulative offset in X
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double crires_pfits_get_cumoffsetx(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO SEQ CUMOFFSETX") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the cumulative offset in Y
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double crires_pfits_get_cumoffsety(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO SEQ CUMOFFSETY") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the CUBE status
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int crires_pfits_get_cube_status(const cpl_propertylist * plist)
{
    int     val ;

    cpl_error_reset() ;
    val = cpl_propertylist_get_bool(plist, "ESO DET FILE CUBE ST") ;
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_error_reset() ;
        val = cpl_propertylist_get_bool(plist, "ESO DET FILE CUBE S") ;
    }
    return val ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the NCORRS.NAME
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * crires_pfits_get_ncorrs(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist, "ESO DET NCORRS NAME");
}

/**@}*/
