/* $Id: crires_photom.c,v 1.13 2011-11-04 13:59:29 yjung Exp $
 *
 * This file is part of the CRIRES Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: yjung $
 * $Date: 2011-11-04 13:59:29 $
 * $Revision: 1.13 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
   								Includes
 -----------------------------------------------------------------------------*/

#include <math.h>
#include <cpl.h>

#include "crires_photom.h"
#include "crires_dfs.h"
#include "crires_pfits.h"
#include "crires_utils.h"

/*-----------------------------------------------------------------------------
   								Includes
 -----------------------------------------------------------------------------*/

static cpl_vector * crires_photom_conversion(const cpl_bivector *,
        const cpl_bivector *) ;
static cpl_vector * crires_photom_throughput(const cpl_vector *, 
        const cpl_vector *, int) ;
static cpl_vector * crires_photom_sensitivity(const cpl_vector *, 
        const cpl_vector *, double) ;
static double crires_photom_great_circle_dist(double, double, double, double) ;
static int crires_photom_star_find(const cpl_vector *, const cpl_vector *, 
        double, double, double, double *) ;

/*----------------------------------------------------------------------------*/
/**
 * @defgroup crires_photom      Photometry
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the conversion/throughput high level
  @param    extr        the extracted spectrum
  @param    std_flux    standard flux
  @param    chip        the chip number
  @param    display     flag to create displays
  @return   0 if ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
int crires_photom_conv_engine(
        cpl_table           *   extr,
        const cpl_bivector  *   std_flux,
        int                     chip,
        int                     display)
{
    cpl_vector      *   wl ;
    cpl_vector      *   extracted ;
    cpl_vector      *   extracted_opt ;
    cpl_bivector    *   star ;
    cpl_bivector    *   star_opt ;
    cpl_vector      *   conversion ;
    cpl_vector      *   conversion_opt ;
    cpl_vector      *   throughput ;

    /* Test entries */
    if (extr == NULL || std_flux == NULL) return -1 ;

    /* Create the extracted bivector */
    wl = cpl_vector_wrap(cpl_table_get_nrow(extr),
            cpl_table_get_data_double(extr, CRIRES_COL_WAVELENGTH)) ;
    extracted = cpl_vector_wrap(cpl_table_get_nrow(extr),
            cpl_table_get_data_double(extr, CRIRES_COL_EXTRACT_INT_RECT)) ;
    extracted_opt = cpl_vector_wrap(cpl_table_get_nrow(extr),
            cpl_table_get_data_double(extr, CRIRES_COL_EXTRACT_INT_OPT)) ;
    star = cpl_bivector_wrap_vectors(wl, extracted);
    star_opt = cpl_bivector_wrap_vectors(wl, extracted_opt);

    /* Compute the OPT conversion */
    cpl_msg_info(__func__, "Conversion computation OPT") ;
    cpl_msg_indent_more() ;
    if ((conversion_opt = crires_photom_conversion(star_opt, std_flux))==NULL) {
        cpl_msg_error(__func__, "Cannot compute the conversion factor OPT") ;
        cpl_msg_indent_less() ;
        cpl_bivector_unwrap_vectors(star) ;
        cpl_bivector_unwrap_vectors(star_opt) ;
        cpl_vector_unwrap(extracted) ;
        cpl_vector_unwrap(extracted_opt) ;
        cpl_vector_unwrap(wl) ;
        return -1 ;
    }
    cpl_msg_indent_less() ;

    /* Compute the RECT conversion */
    cpl_msg_info(__func__, "Conversion computation RECT") ;
    cpl_msg_indent_more() ;
    if ((conversion = crires_photom_conversion(star, std_flux)) == NULL) {
        cpl_msg_error(__func__, "Cannot compute the conversion factor RECT") ;
        cpl_msg_indent_less() ;
        cpl_bivector_unwrap_vectors(star) ;
        cpl_bivector_unwrap_vectors(star_opt) ;
        cpl_vector_unwrap(extracted) ;
        cpl_vector_unwrap(extracted_opt) ;
        cpl_vector_unwrap(wl) ;
        cpl_vector_delete(conversion_opt) ;
        return -1 ;
    }
    cpl_msg_indent_less() ;

    /* Compute the throughput */
    cpl_msg_info(__func__, "Throughput computation") ;
    cpl_msg_indent_more() ;
    if ((throughput = crires_photom_throughput(conversion, wl, chip)) == NULL) {
        cpl_msg_error(__func__, "Cannot compute the throughput") ;
        cpl_msg_indent_less() ;
        cpl_bivector_unwrap_vectors(star) ;
        cpl_bivector_unwrap_vectors(star_opt) ;
        cpl_vector_unwrap(extracted) ;
        cpl_vector_unwrap(extracted_opt) ;
        cpl_vector_unwrap(wl) ;
        cpl_vector_delete(conversion_opt) ;
        cpl_vector_delete(conversion) ;
        return -1 ;
    }
    cpl_msg_indent_less() ;

    /* Destroy the extracted bivector */
    cpl_bivector_unwrap_vectors(star) ;
    cpl_bivector_unwrap_vectors(star_opt) ;
    cpl_vector_unwrap(wl) ;
    cpl_vector_unwrap(extracted) ;
    cpl_vector_unwrap(extracted_opt) ;

    /* Create the output table */
    cpl_table_wrap_double(extr, cpl_vector_get_data(conversion),
            CRIRES_COL_CONVERSION_RECT) ;
    cpl_vector_unwrap(conversion) ;
    cpl_table_wrap_double(extr, cpl_vector_get_data(conversion_opt),
            CRIRES_COL_CONVERSION_OPT) ;
    cpl_vector_unwrap(conversion_opt) ;
    cpl_table_wrap_double(extr, cpl_vector_get_data(throughput),
            CRIRES_COL_THROUGHPUT) ;
    cpl_vector_unwrap(throughput) ;

    /* Plot if requested */
    if(display) {
        cpl_plot_column(
                "set grid;set xlabel 'Wavelength (nm)';set ylabel 'Conversion RECT (ADU/sec/Jy)';",
                "t 'Conversion factor' w lines", "", extr,
                CRIRES_COL_WAVELENGTH, CRIRES_COL_CONVERSION_RECT) ;
        cpl_plot_column(
                "set grid;set xlabel 'Wavelength (nm)';set ylabel 'Conversion OPT (ADU/sec/Jy)';",
                "t 'Conversion factor OPT' w lines", "", extr,
                CRIRES_COL_WAVELENGTH, CRIRES_COL_CONVERSION_OPT) ;
        cpl_plot_column(
                "set grid;set xlabel 'Wavelength (nm)';set ylabel 'Throughput (e-/photons)';",
                "t 'Throughput' w lines", "", extr,
                CRIRES_COL_WAVELENGTH, CRIRES_COL_THROUGHPUT) ;
    }
    return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the wished std star photospheric flux
  @param    tab     the std stars table
  @param    fname   the file to get the RA and DEC
  @return   the bivector with the flux or NULL in error case
 */
/*----------------------------------------------------------------------------*/
cpl_bivector * crires_photom_conv_get_star(
        const cpl_table     *   tab,
        const char          *   fname) 
{
    cpl_propertylist    *   plist ;
    double                  ra, dec ;
    cpl_vector          *   ras ;
    cpl_vector          *   decs ;
    double                  maxdist, dist ;
    int                     ind ; 
    cpl_array           *   wave ;
    cpl_array           *   flux ;
    cpl_vector          *   wave_vec ;
    cpl_vector          *   flux_vec ;
    cpl_bivector        *   flux_biv ;
    
    /* Test entries */
    if (tab == NULL || fname == NULL) return NULL ;

    /* Initialise CRIRES_PHOTOM_STARSDIST_ARCSEC  arcsec in degrees*/
    maxdist = CRIRES_PHOTOM_STARSDIST_ARCSEC / 3600.0 ;
    
    /* Get the RA and DEC observed */
    if ((plist=cpl_propertylist_load(fname, 0)) == NULL) {
        cpl_msg_error(__func__, "Cannot get header");
        return NULL ;
    }
    ra = crires_pfits_get_ra(plist) ;
    dec = crires_pfits_get_dec(plist) ;
    cpl_propertylist_delete(plist) ;
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "Cannot read RA / DEC") ;
        return NULL ;
    }

    /* Get the RAs and DECs */
    ras = cpl_vector_wrap(cpl_table_get_nrow(tab), 
            cpl_table_get_data_double((cpl_table*)tab, CRIRES_COL_RA)) ;
    decs = cpl_vector_wrap(cpl_table_get_nrow(tab), 
            cpl_table_get_data_double((cpl_table*)tab, CRIRES_COL_DEC)) ;

    /* Find the star */
    cpl_msg_info(__func__, "Search a star around RA=%g, DEC=%g in the catalog",
            ra, dec) ;
    if ((ind = crires_photom_star_find(ras, decs, ra, dec, maxdist,
                    NULL)) < 0) {
        cpl_msg_error(__func__, "No star found") ;
        cpl_vector_unwrap(ras) ;
        cpl_vector_unwrap(decs) ;
        return NULL ;
    }
    dist = 3600 * crires_photom_great_circle_dist(cpl_vector_get(ras,
                ind), cpl_vector_get(decs, ind), ra, dec) ;
    cpl_msg_info(__func__, "Star found: %s at RA=%g, DEC=%g at %g arcsec", 
            cpl_table_get_string(tab, CRIRES_COL_STDNAME, ind),
            cpl_vector_get(ras, ind), cpl_vector_get(decs, ind),
            dist) ;
    cpl_vector_unwrap(ras) ;
    cpl_vector_unwrap(decs) ;

    /* Load the star */
    wave=cpl_array_duplicate(cpl_table_get_array(tab,CRIRES_COL_PHOTOFLUX,0));
    flux=cpl_array_duplicate(cpl_table_get_array(tab,CRIRES_COL_PHOTOFLUX,ind));
    wave_vec = cpl_vector_wrap(cpl_array_get_size(wave), 
            cpl_array_get_data_double(wave)) ;
    flux_vec = cpl_vector_wrap(cpl_array_get_size(flux), 
            cpl_array_get_data_double(flux)) ;
    cpl_array_unwrap(wave) ;
    cpl_array_unwrap(flux) ;
    flux_biv=cpl_bivector_wrap_vectors(wave_vec, flux_vec) ;
    return flux_biv ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the sensitivity high level
  @return   the exposure time if ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
double crires_photom_sens_engine(
        cpl_table           *   extr,
        const char          *   file,
        double                  exptime,
        int                     display)
{
    cpl_propertylist    *   plist ;
    int                     ndit, nexp ;
    double                  dit, expt_loc ;
    cpl_vector          *   conversion ;
    cpl_vector          *   sigmas ;
    cpl_vector          *   sensitivity ;

    /* Test entries */
    if (extr == NULL || file == NULL) return -1.0 ;

    /* Check the exposure time */
    if (exptime < 0) {
        if ((plist=cpl_propertylist_load(file, 0)) == NULL) {
            cpl_msg_error(__func__, "Getting header");
            return -1.0 ;
        }
        dit = crires_pfits_get_dit(plist) ;
        ndit = crires_pfits_get_ndit(plist) ;
        nexp = crires_pfits_get_nexp(plist) ;
        cpl_propertylist_delete(plist) ;
        if (cpl_error_get_code() != CPL_ERROR_NONE) {
            cpl_msg_error(__func__, "Cannot read the needed DIT/NDIT/NEXP") ;
            return -1.0 ;
        }
        expt_loc = dit * ndit * nexp ;
    } else {
        expt_loc = exptime ;
    }
    cpl_msg_info(__func__, "Exposure time: %g sec.", expt_loc) ;

    /* Compute the sensitivity */
    cpl_msg_info(__func__, "Sensitivity computation") ;
    cpl_msg_indent_more() ;
    conversion = cpl_vector_wrap(cpl_table_get_nrow(extr),
            cpl_table_get_data_double(extr, CRIRES_COL_CONVERSION_OPT)) ;
    sigmas = cpl_vector_wrap(cpl_table_get_nrow(extr),
            cpl_table_get_data_double(extr, CRIRES_COL_EXTRACT_ERR_OPT)) ;
    sensitivity = crires_photom_sensitivity(conversion, sigmas, expt_loc) ;
    cpl_vector_unwrap(conversion) ;
    cpl_vector_unwrap(sigmas) ;
    if (sensitivity == NULL) {
        cpl_msg_indent_less() ;
        cpl_msg_error(__func__, "Cannot compute the sensitivity") ;
        return -1.0 ;
    }
    cpl_msg_indent_less() ;

    /* Create the output table */
    cpl_table_wrap_double(extr, cpl_vector_get_data(sensitivity),
            CRIRES_COL_SENSITIVITY) ;
    cpl_vector_unwrap(sensitivity) ;

    /* Plot if requested */
    if(display) {
        cpl_plot_column(
                "set grid;set xlabel 'Wavelength (nm)';set ylabel 'Sensitivity (Jy/10sig/hour)';",
                "t 'Sensitivity' w lines", "", extr,
                CRIRES_COL_WAVELENGTH, CRIRES_COL_SENSITIVITY) ;
    }
    return expt_loc ;
}
 
/**@}*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the conversion signal
  @param    star        the star spectrum
  @param    photflux    the star photospheric flux (from a catalog)
  @return   the conversion signal or NULL in error case

  The returned conversion signal is a vector with the size of the star
  bivector. It corresponds, for each wavelength of the star bivector, of
  the star value divided by the photflux value (interpolated).

  conversion = star_spec / model

  star_spec in ADU/sec
  model in Jy
  conversion in ADU/sec/Jy
 */
/*----------------------------------------------------------------------------*/
static cpl_vector * crires_photom_conversion(
        const cpl_bivector  *   star,
        const cpl_bivector  *   photflux)
{
    cpl_vector      *   conversion ;
    cpl_bivector    *   model ;
    double          *   pmodel ;
    int                 i ;

    /* Check entries */
    if (star == NULL) return NULL ;
    if (photflux == NULL) return NULL ;
    
    /* Resample the model */
    model = cpl_bivector_duplicate(star) ;
    if (cpl_bivector_interpolate_linear(model, photflux) != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "Cannot interpolate the signal") ;
        cpl_bivector_delete(model) ;
        return NULL ;
    }

    /* Set to 1 the values that are 0 */
    pmodel = cpl_vector_get_data(cpl_bivector_get_y(model)) ;
    for (i=0 ; i<cpl_bivector_get_size(model) ; i++) {
        if (fabs(pmodel[i]) == 0.0) pmodel[i] = 1.0 ;
    }

    /* Create the conversion */
    conversion = cpl_vector_duplicate(cpl_bivector_get_y_const(star)) ;
    if (cpl_vector_divide(conversion, 
                cpl_bivector_get_y(model)) != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "Cannot divide the star by the model") ;
        cpl_bivector_delete(model) ;
        return NULL ;
    }
    cpl_bivector_delete(model) ;
    return conversion ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the throughput signal
  @param    conversion  the conversion in ADU/sec/Jy
  @param    wl          the wavelength in nanometers
  @param    chip        the chip number
  @return   the throughput signal or NULL in error case

  The returned throughput signal is a vector with the size of the input
  vectors.

  throughput = (conversion*gain) / ((1e-32*S*Delta_lam*1e6)/(wl*h))

  Delta_lam is the wl/pix in m./pix
  wl is the wavelength in m.
  h = 6.626e-34
  S = PI * 4^2 m^2 
  gain = *** / 0.9  phot/sec/(ADU/sec) -> depend on the chip
 */
/*----------------------------------------------------------------------------*/
static cpl_vector * crires_photom_throughput(
        const cpl_vector    *   conversion,
        const cpl_vector    *   wl,
        int                     chip)
{
    double          h, S, gain ;
    cpl_vector  *   throughput ;
    cpl_vector  *   wave_m ;
    cpl_vector  *   dwave_m ;
    double      *   pdwave_m ;
    int             i ;

    /* Check entries */
    if (conversion == NULL) return NULL ;
    if (wl == NULL) return NULL ;
    if (chip < 1 || chip > CRIRES_NB_DETECTORS) return NULL ;
    
    /* Initialise  */
    h = 6.626e-34 ;
    S = 16 * 3.141592653589793238462 ;
    if (chip == 1)      gain = CRIRES_GAIN_CHIP1 ;
    else if (chip == 2) gain = CRIRES_GAIN_CHIP2 ;
    else if (chip == 3) gain = CRIRES_GAIN_CHIP3 ;
    else if (chip == 4) gain = CRIRES_GAIN_CHIP4 ;
    else return NULL ;
 
    /* Wavelength in meters */
    wave_m = cpl_vector_duplicate(wl) ;
    cpl_vector_multiply_scalar(wave_m, 1e-9) ;

    /* Wavelength per pixels */
    dwave_m = cpl_vector_duplicate(wave_m) ;
    pdwave_m = cpl_vector_get_data(dwave_m) ;
    for (i=cpl_vector_get_size(dwave_m)-1 ; i>0 ; i--) 
        pdwave_m[i] -= pdwave_m[i-1] ;
    pdwave_m[0] = pdwave_m[1] ;

    /* Compute the throughput */
    throughput = cpl_vector_duplicate(conversion) ;
    cpl_vector_multiply_scalar(throughput, gain) ;
    cpl_vector_multiply_scalar(throughput, h) ;
    cpl_vector_multiply(throughput, wave_m) ;
    cpl_vector_divide_scalar(throughput, 1e-32) ;
    cpl_vector_divide_scalar(throughput, S) ;
    cpl_vector_divide(throughput, dwave_m) ;
    cpl_vector_divide_scalar(throughput, 1e6) ;

    cpl_vector_delete(dwave_m) ;
    cpl_vector_delete(wave_m) ;
    return throughput ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the sensitivity
  @param    conversion  the conversion signal
  @param    sigma       the sigma (error) on the extraction
  @param    exptime     the exposure time
  @return   the sensitivity or NULL in error case

  The input vectors must have the same size.

  sensitivity = (10 * sigma * sqrt(exptime/3600)) / conversion
  
  sigma in ADU/sec
  conversion in ADU/sec/Jy
  exptime in sec
  sensitivity in Jy/10sigma/hour
 */
/*----------------------------------------------------------------------------*/
static cpl_vector * crires_photom_sensitivity(
        const cpl_vector    *   conversion,
        const cpl_vector    *   sigma,
        double                  exptime)
{
    cpl_vector      *   sensitivity ;
    double              factor ;
    const double    *   pconversion ;
    double          *   psensitivity ;
    int                 i ;

    /* Check entries */
    if (conversion == NULL) return NULL ;
    if (sigma == NULL) return NULL ;
    if (exptime < 0) return NULL ;

    /* Get access  to data */
    pconversion = cpl_vector_get_data_const(conversion) ;
    sensitivity = cpl_vector_duplicate(sigma) ;
    psensitivity = cpl_vector_get_data(sensitivity) ;

    /* Compute sensitivity */
    for (i=0 ; i<cpl_vector_get_size(conversion) ; i++) {
        if (fabs(pconversion[i])<1e-5) psensitivity[i] = 0.0 ;
        else psensitivity[i] /= pconversion[i] ;
    }
    for (i=0 ; i<11 ; i++) psensitivity[i] = 0 ;

    /* Compute the factor */
    factor = sqrt(exptime/3600) * 10 ;

    /* Multiply by the factor */
    if (cpl_vector_multiply_scalar(sensitivity, factor) != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "Cannot mupltiply by the factor") ;
        cpl_vector_delete(sensitivity) ;
        return NULL ;
    }
    return sensitivity ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Find the index of the star closest to (ra, dec)
  @param    v_ra     Vector of Right Ascensions (ra, dec) [degree]
  @param    v_dec    Vector of Declinations (ra, dec)  [degree]
  @param    ra       Right Ascension of position to search for
  @param    dec      Declination of ditto
  @param    maxdist  Maximum acceptable (non-negative) distance [degree]
  @param    pdist    Actual distance (if not NULL)
  @return   The index (starting from zero), or negative on error.

  The two vectors must be of identical length.
  pdist may be NULL.
  It is an error if no star is found.
 */
/*----------------------------------------------------------------------------*/
static int crires_photom_star_find(
        const cpl_vector    *   v_ra, 
        const cpl_vector    *   v_dec,
        double                  ra, 
        double                  dec, 
        double                  maxdist, 
        double              *   pdist)
{
    int                 nra, ndec, minind ;
    double              dmin, gdist ;
    const double    *   pv_ra ;
    const double    *   pv_dec ;
    int                 i ;

    /* Test inputs */
    if (v_ra == NULL || v_dec == NULL) return -1 ;

    /* Initialise */
    dmin = 0 ;
    minind = -1 ;
    nra = cpl_vector_get_size(v_ra) ;
    ndec = cpl_vector_get_size(v_dec) ;
    pv_ra = cpl_vector_get_data_const(v_ra) ;
    pv_dec = cpl_vector_get_data_const(v_dec) ;

    /* Test inputs */
    if (nra != ndec) return -1 ;
    if (maxdist < 0) return -1 ;
    
    /* Find the index of the star closest to the given coordinate */
    for (i=0 ; i < nra ; i++) {
        /* Get the distance */
        gdist = crires_photom_great_circle_dist(pv_ra[i], pv_dec[i], ra, dec);
        if (i == 0 || gdist < dmin) {
            minind = i;
            dmin = gdist;
        }
    }
    if (pdist != NULL) *pdist = dmin;

    /* Check that it is close enough */
    if (dmin > maxdist) return -1 ;
    return minind ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Compute the great-circle distance between two points on a sphere
  @param   ra1    Right ascension of first point [degrees]
  @param   dec1   Declination of first point [degrees]
  @param   ra2    Right ascension of second point [degrees]
  @param   dec2   Declination of second point [degrees]
  @return  Non-negative distance [degrees].
  @see     http://en.wikipedia.org/wiki/Great-circle_distance (on 2005-10-23)
 */
/*----------------------------------------------------------------------------*/
static double crires_photom_great_circle_dist(
        double      ra1, 
        double      dec1,
        double      ra2, 
        double      dec2)
{
  /* Convert all input from degrees to radian - and back for the result */
  const double dra  = sin( atan(1.0)/45.0 * (ra2  - ra1 )/2.0 );
  const double ddec = sin( atan(1.0)/45.0 * (dec2 - dec1)/2.0 );

  dec1 *= atan(1.0)/45.0;
  dec2 *= atan(1.0)/45.0;

  return 2.0 * asin(sqrt( ddec*ddec + cos(dec1)*cos(dec2)*dra*dra))
      * 45.0/atan(1.0);
}
