/* $Id: crires_util_profile.c,v 1.35 2012-09-19 14:10:27 yjung Exp $
 *
 * This file is part of the crires Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: yjung $
 * $Date: 2012-09-19 14:10:27 $
 * $Revision: 1.35 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include "crires_recipe.h"

#include "crires_extract.h"

/*-----------------------------------------------------------------------------
                                Define
 -----------------------------------------------------------------------------*/

#define RECIPE_STRING "crires_util_profile"

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int crires_util_profile_save(const cpl_table **, 
        const cpl_parameterlist *, cpl_frameset *) ;

static char crires_util_profile_description[] = 
"This recipe accepts 1 parameter:\n"
"First parameter:   the profile produced by the spectrum extraction.\n"
"                   (PRO TYPE = "CRIRES_PROTYPE_PROFILE")\n"
"\n"
"This recipe produces 1 file:\n"
"First product:     the table with the FWHMs values of the profile.\n"
"                   (PRO TYPE = "CRIRES_PROTYPE_FWHM")\n" ;

CRIRES_RECIPE_DEFINE(crires_util_profile,
        CRIRES_PARAM_DISPLAY,
        "Compute the FWHM of the spectrum along lambda",
        crires_util_profile_description) ;

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    int                 display ;
    /* Outputs */
    int                 win_mode ;
    crires_illum_period period ;
} crires_util_profile_config ;

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    frameset   the frames list
  @param    parlist     the parameters list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int crires_util_profile(
        cpl_frameset            *   frameset,
        const cpl_parameterlist *   parlist)
{
    cpl_frame           *   fr ;
    const char          *   fname ;
    cpl_imagelist       *   profile ;
    cpl_propertylist    *   plist ;
    const char          *   sval ;
    cpl_table           *   tab[CRIRES_NB_DETECTORS] ;
    cpl_bivector        *   centers_fwhms ;
    int                     nbrows, ly ;
    int                     i, j ;
    
    /* Initialise */
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        tab[i] = NULL ;
    }
    ly = -1 ;
    
    /* Retrieve input parameters */
    crires_util_profile_config.display = crires_parameterlist_get_int(parlist,
            RECIPE_STRING, CRIRES_PARAM_DISPLAY) ;
    
    /* Identify the RAW and CALIB frames in the input frameset */
    if (crires_dfs_set_groups(frameset, NULL)) {
        cpl_msg_error(__func__, "Cannot identify RAW and CALIB frames") ;
        return -1 ;
    }

    /* Get the detector illumination period */
    fr = cpl_frameset_get_position(frameset, 0);
    crires_util_profile_config.period =
        crires_get_detector_illum_period(cpl_frame_get_filename(fr)) ;
    if (crires_util_profile_config.period == CRIRES_ILLUM_UNKNOWN) {
        cpl_msg_error(__func__,
                "Cannot determine the detector illumination period") ;
        return -1 ;
    }

    /* Windowing mode ? */
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(fr), 0)) == NULL)
        return -1 ;
    sval = crires_pfits_get_ncorrs(plist) ;
    if (!strcmp(sval, "FowlerNsampGRstWin")) {
        crires_util_profile_config.period = CRIRES_ILLUM_FULL_DETECTOR ;
        crires_util_profile_config.win_mode = 1 ;
    } else {
        crires_util_profile_config.win_mode = 0 ;
    }
    cpl_propertylist_delete(plist) ;

    /* Display the Detector illumination */
    crires_display_detector_illum(crires_util_profile_config.period) ;

    /* The first frame must be the profile images */
    cpl_msg_info(__func__, "Input image loading") ;
    fname = cpl_frame_get_filename(fr) ;
    profile = crires_load_file(fname, crires_util_profile_config.period, 
            CPL_TYPE_FLOAT) ;

    /* Get the FWHMs */
    cpl_msg_info(__func__, "FWHM computation") ;
    cpl_msg_indent_more() ;
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        /* Skip some detectors in windowing mode */
        if ((i==0 || i==CRIRES_NB_DETECTORS-1) &&
                (crires_util_profile_config.win_mode == 1)) {
            continue ;
        }
        cpl_msg_info(__func__, "Chip number %d", i+1) ;
        cpl_msg_indent_more() ;
        if ((centers_fwhms = crires_extract_spatial_fwhm(
                        cpl_imagelist_get(profile, i))) == NULL) {
            cpl_msg_error(__func__, 
                    "Cannot get the FWHM for chip number %d", i+1) ;
            continue ;
        }
        cpl_msg_indent_less() ;

        /* Correct for the position */
        if (i+1 == 1) 
            ly = crires_get_detector_ly1(crires_util_profile_config.period) ;
        if (i+1 == 2) 
            ly = crires_get_detector_ly2(crires_util_profile_config.period) ;
        if (i+1 == 3) 
            ly = crires_get_detector_ly3(crires_util_profile_config.period) ;
        if (i+1 == 4) 
            ly = crires_get_detector_ly4(crires_util_profile_config.period) ;
        /* If ly not found, use the bottom of the image */
        if (ly < 0) ly = 1 ;
        cpl_vector_add_scalar(cpl_bivector_get_x(centers_fwhms),(double)(ly-1));

        /* Plot if requested */
        if(crires_util_profile_config.display == i+1) {
            cpl_plot_vector(
"set grid;set xlabel 'Position (pixels)';set ylabel 'Y positions (pixels)';",
            "t 'Y positions' w lines", "", cpl_bivector_get_x(centers_fwhms));
            cpl_plot_vector(
        "set grid;set xlabel 'Position (pixels)';set ylabel 'FWHM (pixels)';",
                "t 'FWHMs' w lines", "", cpl_bivector_get_y(centers_fwhms));
        }
       
        /* Create a table from the vector */
        nbrows = cpl_image_get_size_x(cpl_imagelist_get(profile, i)) ;
        tab[i] = cpl_table_new(nbrows) ;
        cpl_table_new_column(tab[i], CRIRES_COL_WL_DETPOSX, CPL_TYPE_INT) ;
        cpl_table_new_column(tab[i], CRIRES_COL_DETPOSY, CPL_TYPE_DOUBLE) ;
        cpl_table_new_column(tab[i], CRIRES_COL_FWHM, CPL_TYPE_DOUBLE) ;
        for (j=0 ; j<nbrows ; j++) {
            cpl_table_set_int(tab[i], CRIRES_COL_WL_DETPOSX, j, j+1) ;
            cpl_table_set_double(tab[i], CRIRES_COL_DETPOSY, j,
                    cpl_vector_get(cpl_bivector_get_x(centers_fwhms), j));
            cpl_table_set_double(tab[i], CRIRES_COL_FWHM, j,
                    cpl_vector_get(cpl_bivector_get_y(centers_fwhms), j));
        }
        cpl_bivector_delete(centers_fwhms) ;
    }
    cpl_msg_indent_less() ;
    cpl_imagelist_delete(profile) ;
        
    /* Reconstruct chips in windowing mode using detector 2 */
    if (crires_util_profile_config.win_mode == 1) {
        if (tab[1] != NULL) {
            tab[0] = cpl_table_duplicate(tab[1]) ;
            cpl_table_set_size(tab[0], 0) ;
            tab[CRIRES_NB_DETECTORS-1] = cpl_table_duplicate(tab[0]) ;
        }
    }

    /* Save the product */
    cpl_msg_info(__func__, "Save the product") ;
    cpl_msg_indent_more() ;
    if (crires_util_profile_save((const cpl_table **)tab, parlist, frameset)) {
        cpl_msg_error(__func__, "Cannot save the product") ;
        cpl_msg_indent_less() ;
        for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) cpl_table_delete(tab[i]) ;
        return -1 ;
    }
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) cpl_table_delete(tab[i]) ;
    cpl_msg_indent_less() ;

    /* Return */
    if (cpl_error_get_code()) return -1 ;
    else return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the products
  @param    out         the tables to save
  @param    parlist     the parameters list
  @param    set         the frames set
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int crires_util_profile_save(
        const cpl_table         **  out,
        const cpl_parameterlist *   parlist,
        cpl_frameset            *   set)
{
    const char          *   recipe_name = "crires_util_profile" ;

    /* Write the table */
    crires_table_save(set,
            parlist,
            set,
            out,
            recipe_name,
            CRIRES_FWHM_TAB,
            CRIRES_PROTYPE_FWHM,
            NULL,
            NULL,
            PACKAGE "/" PACKAGE_VERSION,
            "crires_util_profile.fits") ;

    return  0;
}
