/* $Id: crires_util_wlcalib.c,v 1.69 2012-09-20 12:06:34 yjung Exp $
 *
 * This file is part of the crires Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: yjung $
 * $Date: 2012-09-20 12:06:34 $
 * $Revision: 1.69 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <locale.h>
#include "crires_recipe.h"

#include "irplib_wlxcorr.h"
#include "irplib_utils.h"

#include "crires_wlcalib.h"
#include "crires_wlestimate.h"

/*-----------------------------------------------------------------------------
                                Define
 -----------------------------------------------------------------------------*/

#define RECIPE_STRING "crires_util_wlcalib"

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int crires_util_wlcalib_save(const cpl_imagelist *, const cpl_table **, 
        const cpl_table **, const cpl_parameterlist *, cpl_frameset *) ;

static char crires_util_wlcalib_description[] = 
"This recipe accepts 2 parameters:\n"
"First parameter:   the extracted spectrum in pixels.\n"
"                   (PRO TYPE = "CRIRES_PROTYPE_SPEC_PIX")\n"
"Second parameter:  the table with the lines catalog.\n"
"                   (PRO TYPE = "CRIRES_PROTYPE_CATALOG")\n"
"\n"
"This recipe produces 2 files:\n"
"First product:     the image with the wavelength values.\n"
"                   (PRO TYPE = "CRIRES_PROTYPE_WL_MAP")\n"
"Second product:    the table with the wavelength polynomial coefficients.\n"
"                   (PRO TYPE = "CRIRES_PROTYPE_WL_POLY")\n" ;

CRIRES_RECIPE_DEFINE(crires_util_wlcalib,
        CRIRES_PARAM_WAVES          |
        CRIRES_PARAM_DISPLAY        |
        CRIRES_PARAM_WL_LOG         |
        CRIRES_PARAM_WL_NOLIMIT     |
        CRIRES_PARAM_WL_ERROR       |
        CRIRES_PARAM_XC_LIMIT       |
        CRIRES_PARAM_WL_NBSAMPLES   |
        CRIRES_PARAM_DEGREE         |
        CRIRES_PARAM_WL_CLEAN       |
        CRIRES_PARAM_LINES          |
        CRIRES_PARAM_FIRST_PLOTTED  |
        CRIRES_PARAM_LAST_PLOTTED,
        "Wavelength calibration of a spectrum",
        crires_util_wlcalib_description) ;

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    double                  wstart[CRIRES_NB_DETECTORS] ;
    double                  wstop[CRIRES_NB_DETECTORS] ;
    int                     display ;
    int                     degree ;
    int                     wl_nolimit ;
    int                     wl_log ;
    double                  wl_err ;
    int                     wl_nsamples ;
    int                     wl_clean ;
    double                  wl_xclimit ;
    int                     use_ppm ;
    int                     slitw ;
    int                     fwhm ;
    const char          *   lines ;
    int                     first_plotted_line ;
    int                     last_plotted_line ;
    /* Output */
    int                     win_mode ;
    crires_illum_period     period ;
    double                  qc_wlcent[CRIRES_NB_DETECTORS] ;
    double                  qc_wldisp[CRIRES_NB_DETECTORS] ;
    double                  qc_wlxc[CRIRES_NB_DETECTORS] ;
} crires_util_wlcalib_config ;

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    frameset   the frames list
  @param    parlist     the parameters list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int crires_util_wlcalib(
        cpl_frameset            *   frameset,
        const cpl_parameterlist *   parlist)
{
    cpl_table           *   ext_spec ;
    const char          *   fname ;
    cpl_propertylist    *   plist ;
    double                  wmin, wmax ;
    cpl_vector          *   spec ;
    double              *   pspec ;
    cpl_polynomial      *   phdisp ;
    cpl_polynomial      *   poly_sol ;
    cpl_matrix          *   samppos ;
    cpl_size                mindeg, maxdeg ;
    cpl_bivector        *   lines_biv ;
    cpl_table           *   cat ;
    double                  val ;
    cpl_table           *   real_sol ;
    int                     nrows, spec_pos ;
    cpl_bivector        *   cat_biv ;
    double              *   cat_biv_x ;
    double              *   cat_biv_y ;
    const char          *   sval ;
    cpl_frame           *   fr ;
    cpl_polynomial      *   disp_poly ;
    cpl_table           *   wl_infos[CRIRES_NB_DETECTORS] ;
    cpl_table           *   wl_tab[CRIRES_NB_DETECTORS] ;
    cpl_imagelist       *   wl_map ;
    int                     i, j ;

    /* Needed for sscanf() */
    setlocale(LC_NUMERIC, "C");
 
    /* Initialise */
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        crires_util_wlcalib_config.qc_wlcent[i] = -1.0 ;
        crires_util_wlcalib_config.qc_wldisp[i] = -1.0 ;
        crires_util_wlcalib_config.qc_wlxc[i] = -1.0 ;
        wl_tab[i] = NULL ;
        wl_infos[i] = NULL ;
    }
    crires_util_wlcalib_config.lines = NULL ;
    crires_util_wlcalib_config.use_ppm = 0 ;
    crires_util_wlcalib_config.slitw = 2 ;
    crires_util_wlcalib_config.fwhm = 2 ;
    
    /* Retrieve input parameters */
    sval = crires_parameterlist_get_string(parlist, RECIPE_STRING,
            CRIRES_PARAM_WAVES) ;
    if (sscanf(sval, "%lg,%lg,%lg,%lg,%lg,%lg,%lg,%lg",
                    &crires_util_wlcalib_config.wstart[0],
                    &crires_util_wlcalib_config.wstop[0],
                    &crires_util_wlcalib_config.wstart[1],
                    &crires_util_wlcalib_config.wstop[1],
                    &crires_util_wlcalib_config.wstart[2],
                    &crires_util_wlcalib_config.wstop[2],
                    &crires_util_wlcalib_config.wstart[3],
                &crires_util_wlcalib_config.wstop[3])!=2*CRIRES_NB_DETECTORS){
        return -1 ;
    }
    crires_util_wlcalib_config.display = crires_parameterlist_get_int(parlist,
            RECIPE_STRING, CRIRES_PARAM_DISPLAY) ;
    crires_util_wlcalib_config.degree = crires_parameterlist_get_int(parlist,
            RECIPE_STRING, CRIRES_PARAM_DEGREE) ;
    crires_util_wlcalib_config.wl_log = crires_parameterlist_get_bool(parlist,
            RECIPE_STRING, CRIRES_PARAM_WL_LOG) ;
    crires_util_wlcalib_config.wl_nolimit = crires_parameterlist_get_bool(
            parlist, RECIPE_STRING, CRIRES_PARAM_WL_NOLIMIT) ;
    crires_util_wlcalib_config.wl_err = crires_parameterlist_get_double(parlist,
            RECIPE_STRING, CRIRES_PARAM_WL_ERROR) ;
    crires_util_wlcalib_config.wl_xclimit = crires_parameterlist_get_double(
            parlist, RECIPE_STRING, CRIRES_PARAM_XC_LIMIT) ;
    crires_util_wlcalib_config.wl_nsamples = crires_parameterlist_get_int(
            parlist, RECIPE_STRING, CRIRES_PARAM_WL_NBSAMPLES) ;
    crires_util_wlcalib_config.wl_clean = crires_parameterlist_get_bool(parlist,
            RECIPE_STRING, CRIRES_PARAM_WL_CLEAN) ;
    crires_util_wlcalib_config.lines = crires_parameterlist_get_string(parlist, 
            RECIPE_STRING, CRIRES_PARAM_LINES) ;
    crires_util_wlcalib_config.first_plotted_line=crires_parameterlist_get_int(
            parlist, RECIPE_STRING, CRIRES_PARAM_FIRST_PLOTTED) ;
    crires_util_wlcalib_config.last_plotted_line=crires_parameterlist_get_int(
            parlist, RECIPE_STRING, CRIRES_PARAM_LAST_PLOTTED) ;

    /* Identify the RAW and CALIB frames in the input frameset */
    if (crires_dfs_set_groups(frameset, NULL)) {
        cpl_msg_error(__func__, "Cannot identify RAW and CALIB frames") ;
        return -1 ;
    }
    
    /* Check that we have 2 files in input */
    if (cpl_frameset_get_size(frameset) != 2) {
        cpl_msg_error(__func__, "Expects 2 files in input") ;
        return -1 ;
    }

    /* Get the detector illumination period */
    fr = cpl_frameset_get_position(frameset, 0);
    crires_util_wlcalib_config.period =
        crires_get_detector_illum_period(cpl_frame_get_filename(fr)) ;
    if (crires_util_wlcalib_config.period == CRIRES_ILLUM_UNKNOWN) {
        cpl_msg_error(__func__,
                "Cannot determine the detector illumination period") ;
        return -1 ;
    }

    /* Windowing mode ? */
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(fr), 0)) == NULL)
        return -1 ;
    sval = crires_pfits_get_ncorrs(plist) ;
    if (!strcmp(sval, "FowlerNsampGRstWin")) {
        crires_util_wlcalib_config.period = CRIRES_ILLUM_FULL_DETECTOR ;
        crires_util_wlcalib_config.win_mode = 1 ;
    } else {
        crires_util_wlcalib_config.win_mode = 0 ;
    }
    cpl_propertylist_delete(plist) ;
    
    /* Display the Detector illumination */
    crires_display_detector_illum(crires_util_wlcalib_config.period) ;

    /* Load the second file */
    cpl_msg_info(__func__, "Second frame validity check") ;
    cpl_msg_indent_more() ;
    fr = cpl_frameset_get_position(frameset, 1);
    cat = crires_load_table_check(cpl_frame_get_filename(fr), 1,
            CRIRES_PROTYPE_CATALOG, -1, -1, 0) ;
    if (cat == NULL) {
        cpl_msg_error(__func__, "Second frame is incorrect") ;
        cpl_msg_indent_less() ;
        return -1 ;
    }
    cpl_msg_indent_less() ;
    
    /* Create the catalog spectrum in a bivector */
    nrows = cpl_table_get_nrow(cat) ;
    cat_biv = cpl_bivector_new(nrows) ;
    cat_biv_x = cpl_bivector_get_x_data(cat_biv) ;
    cat_biv_y = cpl_bivector_get_y_data(cat_biv) ;
    for (i=0 ; i<nrows ; i++) {
        cat_biv_x[i] = cpl_table_get(cat, CRIRES_COL_WAVELENGTH, i, NULL) ;
        val = cpl_table_get(cat, CRIRES_COL_EMISSION, i, NULL) ;
        if (crires_util_wlcalib_config.wl_log && val > 0)
            cat_biv_y[i] = log10(val) ;
        else 
            cat_biv_y[i] = val ;
    }
    cpl_table_delete(cat) ;

    /* Get the raw file name */
    fr = cpl_frameset_get_position(frameset, 0);
    fname = cpl_frame_get_filename(fr) ;

    /* Get the Minimum and Maximum wavelengths */
    if (crires_util_wlcalib_config.wl_nolimit == 0) {
        plist = cpl_propertylist_load(fname, 0) ;
        wmin = crires_pfits_get_wlen_min(plist) ;
        wmax = crires_pfits_get_wlen_max(plist) ;
        cpl_propertylist_delete(plist) ;
        if (cpl_error_get_code()) {
            wmin = wmax = -1.0 ;
            cpl_error_reset() ;
        }
    } else {
        wmin = wmax = -1.0 ;
    }

    /* Loop on the spectra to calibrate */
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        /* Skip some detectors in windowing mode */
        if ((i==0 || i==CRIRES_NB_DETECTORS-1) &&
                (crires_util_wlcalib_config.win_mode == 1)) {
            continue ;
        }

        cpl_msg_info(__func__, "Wavelength calibration for chip %d", i+1) ;
        /* Load the first file in table */
        cpl_msg_info(__func__, "Load the extracted table") ;
        cpl_msg_indent_more() ;
        if ((ext_spec=crires_load_table_check(fname, i+1, 
                    CRIRES_PROTYPE_SPEC_PIX, -1, -1, 0)) == NULL) {
            cpl_msg_warning(__func__, "Empty extension") ;
            cpl_msg_indent_less() ;
            continue ;
        }
        cpl_msg_indent_less() ;
        
        /* Get the Wavlength estimation */
        cpl_msg_info(__func__, "Wavelength estimation") ;
        cpl_msg_indent_more() ;
        if ((phdisp = crires_wlestimate_compute(
                        crires_util_wlcalib_config.wstart[i],
                        crires_util_wlcalib_config.wstop[i])) == NULL) {
            if ((phdisp = crires_wlestimate_get(fname, i+1)) == NULL) {
                cpl_msg_error(__func__, "Cannot get the wavelength estimate") ;
                cpl_msg_indent_less() ;
                cpl_table_delete(ext_spec) ;
                continue ;
            }
        }
        cpl_msg_indent_less() ;

        /* Convert to vector */
        nrows = cpl_table_get_nrow(ext_spec) ;
        spec = cpl_vector_new(nrows) ;
        pspec = cpl_vector_get_data(spec) ;
        for (j=0 ; j<nrows ; j++) {
            pspec[j] = cpl_table_get(ext_spec, CRIRES_COL_EXTRACT_INT_RECT, j, 
                    NULL);
        }
        cpl_table_delete(ext_spec) ;

        /* Wavelength calibration */
        cpl_msg_info(__func__, "Wavelength calibration computation") ;
        cpl_msg_indent_more() ;
        if ((disp_poly = crires_wlcalib_engine(spec, cat_biv, phdisp,
                        crires_util_wlcalib_config.slitw, 
                        crires_util_wlcalib_config.fwhm,
                        crires_util_wlcalib_config.degree,
                        wmin, wmax,
                        crires_util_wlcalib_config.wl_err,
                        crires_util_wlcalib_config.wl_nsamples,
                        crires_util_wlcalib_config.wl_clean, 
                        crires_util_wlcalib_config.wl_xclimit,
                        crires_util_wlcalib_config.use_ppm, 
                        (i+1==crires_util_wlcalib_config.display),
                        crires_util_wlcalib_config.first_plotted_line,
                        crires_util_wlcalib_config.last_plotted_line,
                        &((crires_util_wlcalib_config.qc_wlxc)[i]),
                        NULL)) == NULL) {
            cpl_msg_error(__func__, 
                    "Wavelength calibration failed - use estimate") ;
            disp_poly = cpl_polynomial_duplicate(phdisp) ;
            crires_util_wlcalib_config.qc_wlxc[i] = -1.0 ;
        }
        cpl_msg_indent_less() ;
        
        /* Compute the QC parameters */
        crires_util_wlcalib_config.qc_wlcent[i] = 
            cpl_polynomial_eval_1d(disp_poly, (double)512, NULL) ;
        crires_util_wlcalib_config.qc_wldisp[i] = 
            (cpl_polynomial_eval_1d(disp_poly, (double)1024, NULL) -
             cpl_polynomial_eval_1d(disp_poly, (double)1, NULL)) / 1023 ;
        
        /* Plot the result */
        if (crires_util_wlcalib_config.display==i+1) {
            /* Compute the solution from the passed associated lines if passed*/
            poly_sol = NULL ;
            if (crires_util_wlcalib_config.lines != NULL && 
                    crires_util_wlcalib_config.lines[0] != (char)0) {
                lines_biv = 
                    cpl_bivector_read((char*)crires_util_wlcalib_config.lines);
                if ((lines_biv != NULL) && 
        (cpl_bivector_get_size(lines_biv)>crires_util_wlcalib_config.degree)) {
                    poly_sol = cpl_polynomial_new(1);
                    samppos = cpl_matrix_wrap(1, 
                            cpl_bivector_get_size(lines_biv),
                            cpl_bivector_get_x_data(lines_biv)) ;
                    mindeg = 1 ;
                    maxdeg = (cpl_size)crires_util_wlcalib_config.degree ;
                    cpl_polynomial_fit(poly_sol,
                            samppos, NULL,
                            cpl_bivector_get_y(lines_biv),
                            NULL, CPL_FALSE, 
                            &mindeg,
                            &maxdeg) ;
                    cpl_matrix_unwrap(samppos) ;
                }
                if (lines_biv != NULL) cpl_bivector_delete(lines_biv) ;
            } 
            irplib_wlxcorr_plot_solution(phdisp, disp_poly, poly_sol, 1, 1024);
            if (poly_sol != NULL) {
                real_sol = irplib_wlxcorr_gen_spc_table(spec, cat_biv,
                        crires_util_wlcalib_config.slitw, 
                        crires_util_wlcalib_config.fwhm,
                        phdisp, poly_sol) ;
                irplib_wlxcorr_plot_spc_table(real_sol, "Real",
                        crires_util_wlcalib_config.first_plotted_line,
                        crires_util_wlcalib_config.last_plotted_line) ;
                cpl_table_delete(real_sol) ;
            }
            if (poly_sol != NULL) cpl_polynomial_delete(poly_sol) ;
        }

        /* Compute the products */
        wl_infos[i] = irplib_wlxcorr_gen_spc_table(spec, cat_biv, 
                crires_util_wlcalib_config.slitw, 
                crires_util_wlcalib_config.fwhm, 
                phdisp, disp_poly) ;

        /* Free */
        cpl_vector_delete(spec) ;
        cpl_polynomial_delete(phdisp) ;

        /* Create the table */
        if (crires_util_wlcalib_config.qc_wlxc[i] > 0)
            spec_pos = crires_get_y_spec_position(fname, i+1) ;
        else 
            spec_pos = -1 ;
        wl_tab[i] = crires_wlcalib_gen_wltab_one(
                (const cpl_polynomial *)disp_poly, 
                spec_pos, 
                crires_util_wlcalib_config.qc_wlxc[i]) ;
        cpl_polynomial_delete(disp_poly) ;
    }
    cpl_bivector_delete(cat_biv) ;
      
    /* Reconstruct chips in windowing mode using detector 2 */
    if (crires_util_wlcalib_config.win_mode == 1) {
        if (wl_tab[1] != NULL) {
            wl_tab[0] = cpl_table_duplicate(wl_tab[1]) ;
            cpl_table_set_size(wl_tab[0], 0) ;
            wl_tab[CRIRES_NB_DETECTORS-1] = cpl_table_duplicate(wl_tab[0]) ;
        }
        if (wl_infos[1] != NULL) {
            wl_infos[0] = cpl_table_duplicate(wl_infos[1]) ;
            cpl_table_set_size(wl_infos[0], 0) ;
            wl_infos[CRIRES_NB_DETECTORS-1] = cpl_table_duplicate(wl_infos[0]) ;
        }
    }

    /* Create the Map */
    if ((wl_map = crires_wlcalib_gen_wlmap((const cpl_table **)wl_tab)) 
            == NULL) {
        cpl_msg_error(__func__, "Cannot create WL Map") ;
        for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
            if (wl_tab[i] != NULL) cpl_table_delete(wl_tab[i]) ;
            if (wl_infos[i] != NULL) cpl_table_delete(wl_infos[i]) ;
        }
        return -1 ;
    }

    /* Save the result */
    cpl_msg_info(__func__, "Save the products") ;
    cpl_msg_indent_more() ;
    if (crires_util_wlcalib_save(wl_map, (const cpl_table **)wl_tab, 
                (const cpl_table **)wl_infos, parlist, frameset) == -1) {
        cpl_msg_error(__func__, "Cannot save products");
        cpl_msg_indent_less() ;
        cpl_imagelist_delete(wl_map) ;
        for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
            if (wl_tab[i] != NULL) cpl_table_delete(wl_tab[i]) ;
            if (wl_infos[i] != NULL) cpl_table_delete(wl_infos[i]) ;
        }
        return -1 ;
    }
    cpl_msg_indent_less() ;
    cpl_imagelist_delete(wl_map) ;
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        if (wl_tab[i] != NULL) cpl_table_delete(wl_tab[i]) ;
        if (wl_infos[i] != NULL) cpl_table_delete(wl_infos[i]) ;
    }

    /* Return */
    if (cpl_error_get_code()) return -1 ;
    else return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the products
  @param    ilist       the wavelength images
  @param    tab         the wavelength tables
  @param    xctab       the wavelength cross correlation tables
  @param    parlist     the parameters list
  @param    set         the frames set
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int crires_util_wlcalib_save(
        const cpl_imagelist     *   ilist,
        const cpl_table         **  tab,
        const cpl_table         **  xctab,
        const cpl_parameterlist *   parlist,
        cpl_frameset            *   set)
{
    cpl_propertylist    **  qclists ;
    const cpl_frame     *   ref_frame ;
    cpl_propertylist    *   inputlist ;
    const char          *   recipe_name = "crires_util_wlcalib" ;
    int                     i ;

    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW) ;

    /* Create the QC lists */
    qclists = cpl_malloc(CRIRES_NB_DETECTORS * sizeof(cpl_propertylist*)) ;
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        qclists[i] = cpl_propertylist_new() ;
        cpl_propertylist_append_double(qclists[i], "ESO QC CENTWL", 
                crires_util_wlcalib_config.qc_wlcent[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC DISPWL", 
                crires_util_wlcalib_config.qc_wldisp[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC XCORR", 
                crires_util_wlcalib_config.qc_wlxc[i]) ;

        /* Propagate some keywords from input raw frame extensions */
        inputlist = cpl_propertylist_load_regexp(
                cpl_frame_get_filename(ref_frame), i+1,
                CRIRES_HEADER_EXT_FORWARD, 0) ;
        cpl_propertylist_copy_property_regexp(qclists[i], inputlist, 
                CRIRES_HEADER_EXT_FORWARD, 0) ;
        cpl_propertylist_delete(inputlist) ;
    }

    /* Write the image */
    crires_image_save(set,
            parlist,
            set,
            ilist,
            recipe_name,
            CRIRES_WL_MAP_IMA,
            CRIRES_PROTYPE_WL_MAP,
            crires_util_wlcalib_config.period,
            NULL,
            (const cpl_propertylist **)qclists,
            PACKAGE "/" PACKAGE_VERSION,
            "crires_util_wlcalib_ima.fits") ;

    /* Write the table */
    crires_table_save(set,
            parlist,
            set,
            tab,
            recipe_name,
            CRIRES_CALPRO_WAVE,
            CRIRES_PROTYPE_WL_POLY,
            NULL,
            (const cpl_propertylist **)qclists,
            PACKAGE "/" PACKAGE_VERSION,
            "crires_util_wlcalib_tab.fits") ;

    if (xctab[0] != NULL) {
        /* Write the table */
        crires_table_save(set,
                parlist,
                set,
                xctab,
                recipe_name,
                CRIRES_WL_XCORR_TAB,
                CRIRES_PROTYPE_XCORR,
                NULL,
                (const cpl_propertylist **)qclists,
                PACKAGE "/" PACKAGE_VERSION,
                "crires_util_wlcalib_xctab.fits") ;
    }

    /* Free and return */
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        cpl_propertylist_delete(qclists[i]) ;
    }
    cpl_free(qclists) ;
    return  0;
}
