/*
 * This file is part of the CRIRES Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <locale.h>
#include "crires_recipe.h"

#include "crires_model_kernel.h"
#include "crires_combine.h"
#include "crires_extract.h"
#include "crires_photom.h"
#include "crires_wlcalib.h"
#include "crires_wlestimate.h"
#include "irplib_utils.h"

/*-----------------------------------------------------------------------------
                                Define
 -----------------------------------------------------------------------------*/

#define RECIPE_STRING "crires_win_jitter"

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int crires_win_jitter_save(const cpl_imagelist **,
        const cpl_imagelist *, const cpl_imagelist *, const cpl_table **, 
        const cpl_imagelist *, const cpl_imagelist *, 
        const cpl_parameterlist *, cpl_frameset *) ;

static char crires_win_jitter_description[] =
"crires_win_jitter -- Observation recipe in Windowing mode\n"
"The files listed in the Set Of Frames (sof-file) must be tagged:\n"
"   raw-file.fits "CRIRES_WIN_NODDING_OBJECT_RAW" or\n"
"   raw-file.fits "CRIRES_WIN_NODDING_SKY_RAW" or\n"
"   raw-file.fits "CRIRES_WIN_NODDING_RAW" or\n"
"   raw-file.fits "CRIRES_WIN_NODDING_J_RAW" or\n"
"   raw-file.fits "CRIRES_WIN_NODDING_STD_RAW" or\n"
"   raw-file.fits "CRIRES_WIN_NODDING_J_STD_RAW" or\n"
"   flat-file.fits "CRIRES_CALPRO_FLAT_WIN" or\n" 
"   bpm-file.fits "CRIRES_CALPRO_BPM_WIN" or\n" 
"   detlin-file.fits "CRIRES_CALPRO_COEFFS_CUBE" or\n" 
"   dark-file.fits "CRIRES_CALPRO_DARK_WIN" or\n" 
"   wavecal-file.fits "CRIRES_CALPRO_WAVE" or\n" 
"   stdstar-file.fits "CRIRES_CALPRO_STD_PHOTOFLUX" or\n"
"   model-config-file.fits "CRIRES_CALPRO_MODEL_CONFIG".\n"
"\n"
"In the case of a nodding observation (currently, this recipe only\n"
"   supports nodding observations), in order not to degrade the \n"
"   instrument high resolution, the combined images using only NODA\n"
"   or NODB nodding positions can be produced on request. (see --onlyA/B)\n"
"   In this case, the following spectrum extraction can be applied \n"
"   either on the usual combined image or on those NODA/B combined\n"
"   images (see --comb_used).\n"
"\n"
"This recipe produces 6 to 11 files:\n"
"   The combined image      (PRO TYPE = "CRIRES_PROTYPE_COMBINED")\n"
"   The contribution map    (PRO TYPE = "CRIRES_PROTYPE_CONTRIB")\n"
"   The combined image using only Nodding A frames (optional)\n"
"                           (PRO TYPE = "CRIRES_PROTYPE_COMBINED")\n"
"   The contribution map using only Nodding A frames (optional)\n"
"                           (PRO TYPE = "CRIRES_PROTYPE_CONTRIB")\n"
"   The combined image using only Nodding B frames (optional)\n"
"                           (PRO TYPE = "CRIRES_PROTYPE_COMBINED")\n"
"   The contribution map using only Nodding B frames (optional)\n"
"                           (PRO TYPE = "CRIRES_PROTYPE_CONTRIB")\n"
"   The table with the extracted spectrum\n"
"                           (PRO TYPE = "CRIRES_PROTYPE_SPEC_WL") or\n"
"                           (PRO TYPE = "CRIRES_PROTYPE_SENSIT") or\n"
"                           (PRO TYPE = "CRIRES_PROTYPE_CONVERS")\n"
"   The profile image       (PRO TYPE = "CRIRES_PROTYPE_PROFILE")\n"
"   The background map      (PRO TYPE = "CRIRES_PROTYPE_BGD_MAP")\n"
"   The wavelength map      (PRO TYPE = "CRIRES_PROTYPE_WL_MAP")\n"
"   The wavelength map from the model (optional)\n"
"                           (PRO TYPE = "CRIRES_PROTYPE_WL_MAP")\n" ;

CRIRES_RECIPE_DEFINE(crires_win_jitter,
        CRIRES_PARAM_DISPLAY            |
        CRIRES_PARAM_REFINE             |
        CRIRES_PARAM_ONLYA              |
        CRIRES_PARAM_ONLYB              |
        CRIRES_PARAM_COMB_USED          |
        CRIRES_PARAM_HOR_SIZE           |
        CRIRES_PARAM_SPEC_HSIZE         |
        CRIRES_PARAM_KAPPA              |
        CRIRES_PARAM_CLOSING_HSIZE      |
        CRIRES_PARAM_CLEAN_RATE         |
        CRIRES_PARAM_REJECT             |
        CRIRES_PARAM_Y_SPEC_ZONE_CHIP1  |
        CRIRES_PARAM_Y_SPEC_ZONE_CHIP2  |
        CRIRES_PARAM_Y_SPEC_ZONE_CHIP3  |
        CRIRES_PARAM_Y_SPEC_ZONE_CHIP4  |
        CRIRES_PARAM_DEGREE,
        "Observation recipe in Windowing mode",
        crires_win_jitter_description) ;

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    int                 comb_refine ;
    int                 comb_onlyA ;
    int                 comb_onlyB ;
    crires_comb_method  comb_used ;
    int                 extr_box_hor_size ;
    int                 extr_spec_hsize ;
    double              extr_kappa ;
    int                 extr_closing_hs ;
    int                 extr_clean_rate ;
    int                 extr_rej_left ;
    int                 extr_rej_right ;
    const char      *   extr_y_spec_zone_c1 ;
    const char      *   extr_y_spec_zone_c2 ;
    const char      *   extr_y_spec_zone_c3 ;
    const char      *   extr_y_spec_zone_c4 ;
    int                 display ;
    /* Outputs */
    int                 std_mode ;
    double              qc_wlcent[CRIRES_NB_DETECTORS] ;
    double              qc_wldisp[CRIRES_NB_DETECTORS] ;
    int                 qc_specpos[CRIRES_NB_DETECTORS] ;
    int                 qc_specwrec[CRIRES_NB_DETECTORS] ;
    int                 qc_specwopt[CRIRES_NB_DETECTORS] ;
    double              qc_specoptmed[CRIRES_NB_DETECTORS] ;
    double              qc_s2nmed[CRIRES_NB_DETECTORS] ;
    double              qc_sensmed[CRIRES_NB_DETECTORS] ;
    double              qc_convmed[CRIRES_NB_DETECTORS] ;
    double              qc_thromed[CRIRES_NB_DETECTORS] ;
    double              qc_fwhm_comb_pix[CRIRES_NB_DETECTORS] ;
    double              qc_fwhm_comb_as[CRIRES_NB_DETECTORS] ;
    double              qc_fwhm_prof_pix[CRIRES_NB_DETECTORS] ;
    double              qc_fwhm_prof_as[CRIRES_NB_DETECTORS] ;
    double              qc_fwhm_diff[CRIRES_NB_DETECTORS] ;
} crires_win_jitter_config ;

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    parlist     the parameters list
  @param    frameset   the frames list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int crires_win_jitter(
        cpl_frameset            *   frameset,
        const cpl_parameterlist *   parlist)
{
    const char          *   sval ;
    const char          *   y_pos ;
    cpl_frameset        *   rawframes ;
    const char          *   fname ;
    cpl_frame           *   fr ;
    double                  tot_ndit ;
    const char          *   flat ;
    const char          *   dark ;
    const char          *   bpm ;
    const char          *   detlin ;
    const char          *   wavecal ;
    const char          *   std_star ;
    const char          *   cfg_model ;
    cpl_table           *   std_star_tab ;
    cpl_bivector        *   std_star_biv ;
    cpl_imagelist       **  comblist ;
    int                     comblist_offset ;
    int                     pix, extr_spec_starty, extr_spec_stopy ;
    cpl_polynomial      *   wave_poly ;
    cpl_table           *   wave_tab[CRIRES_NB_DETECTORS] ;
    cpl_table           *   extr_tab[CRIRES_NB_DETECTORS] ;
    cpl_image           *   profiles[CRIRES_NB_DETECTORS] ;
    cpl_image           *   bg_maps[CRIRES_NB_DETECTORS] ;
    cpl_imagelist       *   prof_list ;
    cpl_imagelist       *   bgmap_list ;
    cpl_image           *   profile_empty ;
    cpl_image           *   bgmap_empty ;
    cpl_table           *   extr_tab_empty ;
    cpl_imagelist       *   wl_map ;
    cpl_imagelist       *   wl_map_model ;
    cpl_vector          **  wavelengths ;
    int                     i, j ;

    /* Needed for sscanf() */
    setlocale(LC_NUMERIC, "C");

    /* Initialise */
    rawframes = NULL ;
    comblist_offset = -1 ;
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        crires_win_jitter_config.qc_wlcent[i] = -1 ;
        crires_win_jitter_config.qc_wldisp[i] = -1 ;
        crires_win_jitter_config.qc_specpos[i] = -1 ;
        crires_win_jitter_config.qc_specwrec[i] = -1 ;
        crires_win_jitter_config.qc_specwopt[i] = -1 ;
        crires_win_jitter_config.qc_specoptmed[i] = -1.0 ;
        crires_win_jitter_config.qc_s2nmed[i] = -1.0 ;
        crires_win_jitter_config.qc_sensmed[i] = -1.0 ;
        crires_win_jitter_config.qc_convmed[i] = -1.0 ;
        crires_win_jitter_config.qc_thromed[i] = -1.0 ;
        crires_win_jitter_config.qc_fwhm_comb_pix[i] = -1.0 ;
        crires_win_jitter_config.qc_fwhm_comb_as[i] = -1.0 ;
        crires_win_jitter_config.qc_fwhm_prof_pix[i] = -1.0 ;
        crires_win_jitter_config.qc_fwhm_prof_as[i] = -1.0 ;
        crires_win_jitter_config.qc_fwhm_diff[i] = -1.0 ;
    }
 
    /* Retrieve input parameters */
    crires_win_jitter_config.display = crires_parameterlist_get_int(parlist,
            RECIPE_STRING, CRIRES_PARAM_DISPLAY) ;
    crires_win_jitter_config.comb_refine = crires_parameterlist_get_bool(
            parlist, RECIPE_STRING, CRIRES_PARAM_REFINE) ;
    crires_win_jitter_config.comb_onlyA = crires_parameterlist_get_bool(
            parlist, RECIPE_STRING, CRIRES_PARAM_ONLYA) ;
    crires_win_jitter_config.comb_onlyB = crires_parameterlist_get_bool(
            parlist, RECIPE_STRING, CRIRES_PARAM_ONLYB) ;
    sval = crires_parameterlist_get_string(parlist, RECIPE_STRING,
            CRIRES_PARAM_COMB_USED) ;
    if (!strcmp(sval, "NODA"))
        crires_win_jitter_config.comb_used = CRIRES_COMB_METHOD_NODA ;
    else if (!strcmp(sval, "NODB"))
        crires_win_jitter_config.comb_used = CRIRES_COMB_METHOD_NODB ;
    else if (!strcmp(sval, "COMB"))
        crires_win_jitter_config.comb_used = CRIRES_COMB_METHOD_COMB ;
    else {
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
        cpl_msg_error(__func__, "Invalid combination method specified");
        return -1;
    }
    crires_win_jitter_config.extr_box_hor_size = crires_parameterlist_get_int(
            parlist, RECIPE_STRING, CRIRES_PARAM_HOR_SIZE) ;
    crires_win_jitter_config.extr_spec_hsize = crires_parameterlist_get_int(
            parlist, RECIPE_STRING, CRIRES_PARAM_SPEC_HSIZE) ;
    crires_win_jitter_config.extr_kappa = crires_parameterlist_get_double(
            parlist, RECIPE_STRING, CRIRES_PARAM_KAPPA) ;
    crires_win_jitter_config.extr_closing_hs = crires_parameterlist_get_int(
            parlist, RECIPE_STRING, CRIRES_PARAM_CLOSING_HSIZE) ;
    crires_win_jitter_config.extr_clean_rate = crires_parameterlist_get_double(
            parlist, RECIPE_STRING, CRIRES_PARAM_CLEAN_RATE) ;
    sval = crires_parameterlist_get_string(parlist, RECIPE_STRING,
            CRIRES_PARAM_REJECT) ;
    if (sscanf(sval, "%d,%d",
                    &crires_win_jitter_config.extr_rej_left,
                    &crires_win_jitter_config.extr_rej_right)!=2) {
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
        return -1 ;
    }
    crires_win_jitter_config.extr_y_spec_zone_c1 =
        crires_parameterlist_get_string(parlist, RECIPE_STRING,
                CRIRES_PARAM_Y_SPEC_ZONE_CHIP1) ;
    crires_win_jitter_config.extr_y_spec_zone_c2 =
        crires_parameterlist_get_string(parlist, RECIPE_STRING,
                CRIRES_PARAM_Y_SPEC_ZONE_CHIP2) ;
    crires_win_jitter_config.extr_y_spec_zone_c3 =
        crires_parameterlist_get_string(parlist, RECIPE_STRING,
                CRIRES_PARAM_Y_SPEC_ZONE_CHIP3) ;
    crires_win_jitter_config.extr_y_spec_zone_c4 =
        crires_parameterlist_get_string(parlist, RECIPE_STRING,
                CRIRES_PARAM_Y_SPEC_ZONE_CHIP4) ;

    /* Identify the RAW and CALIB frames in the input frameset */
    if (crires_dfs_set_groups(frameset, "crires_win_jitter")) {
        cpl_msg_error(__func__, "Cannot identify RAW and CALIB frames") ;
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
        return -1 ;
    }

    /* Retrieve calibration data */
    flat        = crires_extract_filename(frameset, CRIRES_CALPRO_FLAT_WIN) ;
    dark        = crires_extract_filename(frameset, CRIRES_CALPRO_DARK_WIN) ;
    bpm         = crires_extract_filename(frameset, CRIRES_CALPRO_BPM_WIN) ;
    detlin      = crires_extract_filename(frameset, CRIRES_CALPRO_COEFFS_CUBE) ;
    wavecal     = crires_extract_filename(frameset, CRIRES_CALPRO_WAVE) ;
    std_star    = crires_extract_filename(frameset,CRIRES_CALPRO_STD_PHOTOFLUX);
    cfg_model   = crires_extract_filename(frameset, CRIRES_CALPRO_MODEL_CONFIG);
   
    /* Retrieve raw frames */
    if ((rawframes = crires_extract_frameset(frameset,
                    CRIRES_WIN_NODDING_OBJECT_RAW)) != NULL) {
        crires_win_jitter_config.std_mode = 0 ;
    } else if ((rawframes = crires_extract_frameset(frameset,
                    CRIRES_WIN_NODDING_RAW)) != NULL) {
        crires_win_jitter_config.std_mode = 0 ;
    } else if ((rawframes = crires_extract_frameset(frameset,
                    CRIRES_WIN_NODDING_J_RAW)) != NULL) {
        crires_win_jitter_config.std_mode = 0 ;
    } else if ((rawframes = crires_extract_frameset(frameset,
                    CRIRES_WIN_NODDING_STD_RAW)) != NULL) {
        crires_win_jitter_config.std_mode = 1 ;
    } else if ((rawframes = crires_extract_frameset(frameset,
                    CRIRES_WIN_NODDING_J_STD_RAW)) != NULL) {
        crires_win_jitter_config.std_mode = 1 ;
    } else {
        cpl_msg_error(__func__, "No raw frame in input") ;
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
        return -1 ;
    }

   /* Verify the STRIPE keys conformity */
    fname = cpl_frame_get_filename(cpl_frameset_get_position(rawframes,0)) ;
    if (flat != NULL) {
        if (crire_stripe_keys_mismatch(fname, flat)) {
            cpl_msg_error(__func__,
                    "Mismatch of STRIPE keys with the flat frame") ;
            cpl_frameset_delete(rawframes) ;
            cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
            return -1 ;
        }
    }
    if (bpm != NULL) {
        if (crire_stripe_keys_mismatch(fname, bpm)) {
            cpl_msg_error(__func__,
                    "Mismatch of STRIPE keys with the bpm frame") ;
            cpl_frameset_delete(rawframes) ;
            cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
            return -1 ;
        }
    }
    if (dark != NULL) {
        if (crire_stripe_keys_mismatch(fname, dark)) {
            cpl_msg_error(__func__,
                    "Mismatch of STRIPE keys with the dark frame") ;
            cpl_frameset_delete(rawframes) ;
            cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
            return -1 ;
        }
    }

    /* Checks on the parameters validity */
    if ((crires_win_jitter_config.comb_used == CRIRES_COMB_METHOD_NODA)
        && (crires_win_jitter_config.comb_onlyA == 0)) {
        cpl_msg_warning(__func__,
                "You forgot to require the NODA image to be produced !") ;
        crires_win_jitter_config.comb_onlyA = 1 ;
    }
    if ((crires_win_jitter_config.comb_used == CRIRES_COMB_METHOD_NODB)
        && (crires_win_jitter_config.comb_onlyB == 0)) {
        cpl_msg_warning(__func__,
                "You forgot to require the NODB image to be produced !") ;
        crires_win_jitter_config.comb_onlyB = 1 ;
    }

    /* Set comblist_offset */
    if (crires_win_jitter_config.comb_used == CRIRES_COMB_METHOD_COMB)
        comblist_offset = 0 ;
    else if (crires_win_jitter_config.comb_used == CRIRES_COMB_METHOD_NODA)
        comblist_offset = 1 ;
    else if (crires_win_jitter_config.comb_used == CRIRES_COMB_METHOD_NODB)
        comblist_offset = 2 ;

    /* Get the total number of NDIT */
    fr = cpl_frameset_get_position(rawframes, 0);
    tot_ndit = crires_get_totndit(cpl_frame_get_filename(fr)) ;
    if (tot_ndit < 0) {
        cpl_msg_error(__func__, "Cannot get the total number of NDIT") ;
        cpl_frameset_delete(rawframes) ;
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
        return -1 ;
    }
    tot_ndit *= cpl_frameset_get_size(rawframes) ;
    
    /* Images recombination */
    cpl_msg_info(__func__, "Images combination") ;
    cpl_msg_indent_more() ;
    if ((comblist = crires_combine_imagelist_win(rawframes, 
                    flat, dark, bpm, detlin,
                    crires_win_jitter_config.comb_refine,
                    crires_win_jitter_config.comb_onlyA,
                    crires_win_jitter_config.comb_onlyB)) == NULL) {
        cpl_msg_error(__func__, "Cannot combine the images") ;
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
        cpl_frameset_delete(rawframes) ;
        cpl_msg_indent_less() ;
        return -1 ;
    }
    cpl_msg_indent_less() ;
      
    /* Spectrum extraction */
    cpl_msg_info(__func__, "Spectrum extraction") ;
    cpl_msg_indent_more() ;
    for (i=1 ; i<CRIRES_NB_DETECTORS-1 ; i++) {
        cpl_msg_info(__func__, "Chip number %d extraction", i+1) ;
        cpl_msg_indent_more() ;
    
        /* Where is the spectrum extraction zone ? */
        y_pos = "" ;
        if (i == 1) y_pos = crires_win_jitter_config.extr_y_spec_zone_c1 ;
        if (i == 2) y_pos = crires_win_jitter_config.extr_y_spec_zone_c2 ;
        if (i == 3) y_pos = crires_win_jitter_config.extr_y_spec_zone_c3 ;
        if (i == 4) y_pos = crires_win_jitter_config.extr_y_spec_zone_c4 ;
        if (sscanf(y_pos,"%d,%d", &extr_spec_starty, &extr_spec_stopy)!=2) {
            cpl_msg_warning(__func__, "Wrong Spectral Zone specified: %s",
                    y_pos) ;
            extr_spec_starty = extr_spec_stopy = -1 ;
        }
        if ((extr_tab[i] = crires_extract_spectrum(
                        cpl_imagelist_get(comblist[0+2*comblist_offset], i),
                        cpl_imagelist_get(comblist[1+2*comblist_offset], i),
                        crires_win_jitter_config.extr_box_hor_size,
                        crires_win_jitter_config.extr_spec_hsize,
                        crires_win_jitter_config.extr_kappa,
                        crires_win_jitter_config.extr_closing_hs,
                        crires_win_jitter_config.extr_clean_rate,
                        crires_win_jitter_config.extr_rej_left,
                        crires_win_jitter_config.extr_rej_right,
                        extr_spec_starty, extr_spec_stopy, i+1, tot_ndit,
                        CRIRES_ILLUM_FULL_DETECTOR,
                        &(crires_win_jitter_config.qc_specpos[i]),
                        &(crires_win_jitter_config.qc_specwrec[i]),
                        &(crires_win_jitter_config.qc_specwopt[i]),
                        &(crires_win_jitter_config.qc_specoptmed[i]),
                        &(crires_win_jitter_config.qc_s2nmed[i]),
                        &(profiles[i]),
                        &(bg_maps[i]))) == NULL) {
            cpl_msg_error(__func__, "Cannot extract the spectrum") ;
            cpl_msg_indent_less() ;
            cpl_msg_indent_less() ;
            for (j=1 ; j<i ; j++) 
                cpl_table_delete(extr_tab[j]) ;
            for (j=1 ; j<i ; j++) 
                cpl_image_delete(profiles[j]) ;
            for (j=1 ; j<i ; j++) 
                cpl_image_delete(bg_maps[j]) ;
            cpl_imagelist_delete(comblist[0]) ;
            cpl_imagelist_delete(comblist[1]) ;
            if (crires_win_jitter_config.comb_onlyA) {
                cpl_imagelist_delete(comblist[2]) ;
                cpl_imagelist_delete(comblist[3]) ;
            }
            if (crires_win_jitter_config.comb_onlyB) {
                cpl_imagelist_delete(comblist[4]) ;
                cpl_imagelist_delete(comblist[5]) ;
            }
            cpl_free(comblist) ;
            cpl_frameset_delete(rawframes) ;
            return -1 ;
        }
        cpl_msg_info(__func__, "Chip number %d FWHM Computation", i+1) ;
        if (crires_extract_qc_fwhm(
                    cpl_imagelist_get(comblist[0+2*comblist_offset], i),
                    profiles[i],
                    &(crires_win_jitter_config.qc_fwhm_comb_pix[i]),
                    &(crires_win_jitter_config.qc_fwhm_comb_as[i]),
                    &(crires_win_jitter_config.qc_fwhm_prof_pix[i]),
                    &(crires_win_jitter_config.qc_fwhm_prof_as[i]),
                    &(crires_win_jitter_config.qc_fwhm_diff[i])) == -1) {
            cpl_msg_warning(__func__, "Failed for FWHM computation") ;
            crires_win_jitter_config.qc_fwhm_comb_pix[i] = -1.0 ;
            crires_win_jitter_config.qc_fwhm_comb_as[i] = -1.0 ;
            crires_win_jitter_config.qc_fwhm_prof_pix[i] = -1.0 ;
            crires_win_jitter_config.qc_fwhm_prof_as[i] = -1.0 ;
            crires_win_jitter_config.qc_fwhm_diff[i] = -1.0 ;
        }
        cpl_msg_indent_less() ;
    }

    /* Create the profile and bg maps */
    prof_list = cpl_imagelist_new() ;
    bgmap_list = cpl_imagelist_new() ;
    profile_empty = cpl_image_duplicate(profiles[1]) ;
    bgmap_empty = cpl_image_duplicate(bg_maps[1]) ;
    cpl_image_multiply_scalar(profile_empty, 0.0) ;
    cpl_image_multiply_scalar(bgmap_empty, 0.0) ;
    cpl_imagelist_set(prof_list, cpl_image_duplicate(profile_empty), 0) ;
    cpl_imagelist_set(bgmap_list, cpl_image_duplicate(bgmap_empty), 0) ;
    for (i=1 ; i<CRIRES_NB_DETECTORS-1 ; i++) {
        cpl_imagelist_set(prof_list, profiles[i], i) ;
        cpl_imagelist_set(bgmap_list, bg_maps[i], i) ;
    }
    cpl_imagelist_set(prof_list, cpl_image_duplicate(profile_empty), 
            CRIRES_NB_DETECTORS-1) ;
    cpl_imagelist_set(bgmap_list, cpl_image_duplicate(bgmap_empty), 
            CRIRES_NB_DETECTORS-1) ;
    cpl_image_delete(profile_empty) ;
    cpl_image_delete(bgmap_empty) ;
    
    /* Test that the spectrum is at the same place in all detectors */
    if (crires_win_jitter_config.qc_specpos[1] > 0 && 
            crires_win_jitter_config.qc_specpos[2] > 0 &&
            fabs(crires_win_jitter_config.qc_specpos[1] -
                crires_win_jitter_config.qc_specpos[2]) > 
            CRIRES_SPEC_POS_TOLERANCE) {
        cpl_msg_warning(__func__,
"The spectrum positions in chip 1 and chip 2 are too different: %d -> %d",
                crires_win_jitter_config.qc_specpos[1], 
                crires_win_jitter_config.qc_specpos[2]) ;
    }
    cpl_msg_indent_less() ;

    /* Wavelength calibration */
    cpl_msg_info(__func__, "Wavelength Calibration") ;
    cpl_msg_indent_more() ;
    if (wavecal != NULL) {
        /* Wavelength solution is provided */
        cpl_msg_info(__func__, "Use the provided solution") ;
        for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
            if ((wave_tab[i] = crires_load_table_check(wavecal, i+1,
                            CRIRES_PROTYPE_WL_POLY, -1, -1, 0)) == NULL) {
                cpl_msg_error(__func__, "Cannot load the wavelength table") ;
                cpl_msg_indent_less() ;
                cpl_frameset_delete(rawframes) ;
                cpl_imagelist_delete(comblist[0]) ;
                cpl_imagelist_delete(comblist[1]) ;
                if (crires_win_jitter_config.comb_onlyA) {
                    cpl_imagelist_delete(comblist[2]) ;
                    cpl_imagelist_delete(comblist[3]) ;
                }
                if (crires_win_jitter_config.comb_onlyB) {
                    cpl_imagelist_delete(comblist[4]) ;
                    cpl_imagelist_delete(comblist[5]) ;
                }
                cpl_free(comblist) ;
                for (j=1 ; j<CRIRES_NB_DETECTORS-1 ; j++)
                    cpl_table_delete(extr_tab[j]) ;
                cpl_imagelist_delete(prof_list) ;
                cpl_imagelist_delete(bgmap_list) ;
                return -1 ;
            }
        }
    } else {
        /* Calibrate from the keywords */
        cpl_msg_info(__func__, "Use the keywords for the wavelength") ;

        /* Reduce the first raw frame */
        fname = cpl_frame_get_filename(cpl_frameset_get_position(rawframes,0)) ;

        /* Loop on the detectors */
        for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
            cpl_msg_info(__func__, "Calibrate chip number %d", i+1) ;
            cpl_msg_indent_more() ;
            wave_poly = crires_wlestimate_get(fname, i+1) ;
            wave_tab[i] = crires_wlcalib_gen_wltab_one(wave_poly, -1, -1.0) ;
            cpl_polynomial_delete(wave_poly) ;
            cpl_msg_indent_less() ;
        }
    }

    /* Create the wave map */
    if ((wl_map = crires_wlcalib_gen_wlmap((const cpl_table **)wave_tab))
            == NULL) {
        cpl_msg_error(__func__, "Cannot compute the Wavelength Map") ;
        cpl_frameset_delete(rawframes) ;
        cpl_imagelist_delete(comblist[0]) ;
        cpl_imagelist_delete(comblist[1]) ;
        if (crires_win_jitter_config.comb_onlyA) {
            cpl_imagelist_delete(comblist[2]) ;
            cpl_imagelist_delete(comblist[3]) ;
        }
        if (crires_win_jitter_config.comb_onlyB) {
            cpl_imagelist_delete(comblist[4]) ;
            cpl_imagelist_delete(comblist[5]) ;
        }
        cpl_free(comblist) ;
        cpl_imagelist_delete(prof_list) ;
        cpl_imagelist_delete(bgmap_list) ;
        for (j=1 ; j<CRIRES_NB_DETECTORS-1 ; j++) 
            cpl_table_delete(extr_tab[j]) ;
        for (j=0 ; j<CRIRES_NB_DETECTORS ; j++) 
            if (wave_tab[j] != NULL) cpl_table_delete(wave_tab[j]);
        return -1 ;
    }
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++)
        if (wave_tab[i] != NULL) cpl_table_delete(wave_tab[i]);

    /* Compute the QC parameters */
    for (i=1 ; i<CRIRES_NB_DETECTORS-1 ; i++) {
        crires_win_jitter_config.qc_wlcent[i] =
            cpl_image_get(cpl_imagelist_get(wl_map, i),
                    512, crires_win_jitter_config.qc_specpos[i], &pix) ;
        crires_win_jitter_config.qc_wldisp[i] =
            ((cpl_image_get(cpl_imagelist_get(wl_map, i), 1024,
                            crires_win_jitter_config.qc_specpos[i], &pix)) -
             (cpl_image_get(cpl_imagelist_get(wl_map, i), 1,
                            crires_win_jitter_config.qc_specpos[i], &pix)))
            / 1023 ;
    }

    /* Get the wl map from the model */
    fname  = cpl_frame_get_filename(cpl_frameset_get_position(rawframes,0)) ;
    if ((cfg_model != NULL) && (!crires_model_off()) && 
            (crires_model_config_check(cfg_model, fname) == 0)) {
        cpl_msg_info(__func__, "Call the model to get the wavelength map") ;
        cpl_msg_indent_more() ;
        wl_map_model = crires_model_wavpix(fname, cfg_model, -1) ;
        if (wl_map_model == NULL) {
            cpl_msg_warning(__func__, "Model function returns NULL") ;
            cpl_error_reset() ;
        }
        cpl_msg_indent_less() ;
    } else {
        wl_map_model = NULL ;
    }
    
    /* Apply the wavelength */
    for (i=1 ; i<CRIRES_NB_DETECTORS-1 ; i++) {
        cpl_table_new_column(extr_tab[i], CRIRES_COL_WAVELENGTH, 
                CPL_TYPE_DOUBLE) ;
        for (j=0 ; j<cpl_table_get_nrow(extr_tab[i]) ; j++) {
            cpl_table_set_double(extr_tab[i], CRIRES_COL_WAVELENGTH, j, 
                    cpl_image_get(cpl_imagelist_get_const(wl_map, i), j+1,
                        crires_win_jitter_config.qc_specpos[i], &pix));
        }
    }

    /* Add the Model Wavelength and Call the model to fill it */
    for (i=1 ; i<CRIRES_NB_DETECTORS-1 ; i++) {
        cpl_table_new_column(extr_tab[i], CRIRES_COL_WAVELENGTH_MODEL, 
                CPL_TYPE_DOUBLE) ;
        cpl_table_fill_column_window_double(extr_tab[i],
                CRIRES_COL_WAVELENGTH_MODEL, 0,
                cpl_table_get_nrow(extr_tab[i]), -1.0) ;
    }
    if ((cfg_model != NULL) && (!crires_model_off()) && (1)) {
        cpl_msg_info(__func__, "Call the model to get the wavelengths") ;
        cpl_msg_indent_more() ;
        wavelengths = crires_model_wavelengths(
                cpl_frame_get_filename(cpl_frameset_get_position(rawframes,0)),
                cfg_model, -1,
                (double)(crires_win_jitter_config.qc_specpos[0]),
                wl_map_model) ;
        if (wavelengths != NULL) {
            /* Loop on the detectors */
            for (i=1 ; i<CRIRES_NB_DETECTORS-1 ; i++) {
                /* Loop on the x values */
                for (j=0 ; j<cpl_vector_get_size(wavelengths[i]) ; j++) {
                    cpl_table_set_double(extr_tab[i], 
                            CRIRES_COL_WAVELENGTH_MODEL, j, 
                            cpl_vector_get(wavelengths[i], j)) ;
                }
            }
            for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) 
                cpl_vector_delete(wavelengths[i]) ;
            cpl_free(wavelengths) ;
        } else {
            cpl_msg_warning(__func__, "Model function returns NULL") ;
            cpl_error_reset() ;
        }
        cpl_msg_indent_less() ;
    }

    /* Conversion factor / Sensitivity */
    if (crires_win_jitter_config.std_mode) {
        cpl_msg_info(__func__, 
                "Sensitivity / Conversion / Throughput computation") ;
        cpl_msg_indent_more() ;
        /* Load std star */
        if ((std_star_tab = crires_load_table_check(std_star, 1,
                        CRIRES_PROTYPE_PHO_FLUX, -1, -1, 0)) == NULL) {
            cpl_msg_error(__func__, "Cannot load the std star flux") ;
        } else {
            /* Get the wished std star */
            if ((std_star_biv = crires_photom_conv_get_star(std_star_tab, 
            cpl_frame_get_filename(cpl_frameset_get_position(rawframes,0)))) 
                    == NULL) {
                cpl_msg_error(__func__, "Cannot find the star flux") ;
                cpl_table_delete(std_star_tab) ;
            } else {
                cpl_table_delete(std_star_tab) ;
                /* Apply the conversion  */
                for (i=1 ; i<CRIRES_NB_DETECTORS-1 ; i++) {
                    if (crires_photom_conv_engine(extr_tab[i], std_star_biv,
                            i+1, (i+1==crires_win_jitter_config.display))==0){
                        crires_win_jitter_config.qc_convmed[i] =
                            cpl_table_get_column_median(extr_tab[i],
                                    CRIRES_COL_CONVERSION_RECT) ;
                        crires_win_jitter_config.qc_thromed[i] =
                            cpl_table_get_column_median(extr_tab[i],
                                    CRIRES_COL_THROUGHPUT) ;
                        /* Apply the sensitivity */
                        if (crires_photom_sens_engine(extr_tab[i], 
                            cpl_frame_get_filename(cpl_frameset_get_position(
                                        rawframes,0)), -1.0, 
                                (i+1==crires_win_jitter_config.display))>0) {
                            crires_win_jitter_config.qc_sensmed[i] =
                                cpl_table_get_column_median(extr_tab[i], 
                                        CRIRES_COL_SENSITIVITY) ; ;
                        }
                    }
                }
                cpl_bivector_delete(std_star_biv) ;
            }
        }
        cpl_msg_indent_less() ;
    }
    cpl_frameset_delete(rawframes) ;

    /* Create empty extr_tab for non-processed chips  */
    extr_tab_empty = cpl_table_duplicate(extr_tab[1]) ;
    cpl_table_set_size(extr_tab_empty, 0) ;
    extr_tab[0] = cpl_table_duplicate(extr_tab_empty) ;
    extr_tab[CRIRES_NB_DETECTORS-1] = cpl_table_duplicate(extr_tab_empty) ;
    cpl_table_delete(extr_tab_empty) ;
 
    /* Save the product */
    cpl_msg_info(__func__, "Save the product") ;
    cpl_msg_indent_more() ;
    if (crires_win_jitter_save((const cpl_imagelist **)comblist, 
                prof_list, bgmap_list,
                (const cpl_table **)extr_tab, wl_map, wl_map_model, parlist, 
                frameset)) {
        cpl_msg_error(__func__, "Cannot save the product") ;
        cpl_imagelist_delete(prof_list) ;
        cpl_imagelist_delete(bgmap_list) ;
        for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) 
            if (extr_tab[i] != NULL) cpl_table_delete(extr_tab[i]) ;
        cpl_imagelist_delete(comblist[0]) ;
        cpl_imagelist_delete(comblist[1]) ;
        if (crires_win_jitter_config.comb_onlyA) {
            cpl_imagelist_delete(comblist[2]) ;
            cpl_imagelist_delete(comblist[3]) ;
        }
        if (crires_win_jitter_config.comb_onlyB) {
            cpl_imagelist_delete(comblist[4]) ;
            cpl_imagelist_delete(comblist[5]) ;
        }
        cpl_free(comblist) ;
        cpl_imagelist_delete(wl_map) ;
        if (wl_map_model) cpl_imagelist_delete(wl_map_model) ;
        cpl_msg_indent_less() ;
        return -1 ;
    }
    cpl_imagelist_delete(wl_map) ;
    if (wl_map_model) cpl_imagelist_delete(wl_map_model) ;
    cpl_imagelist_delete(comblist[0]) ;
    cpl_imagelist_delete(comblist[1]) ;
    if (crires_win_jitter_config.comb_onlyA) {
        cpl_imagelist_delete(comblist[2]) ;
        cpl_imagelist_delete(comblist[3]) ;
    }
    if (crires_win_jitter_config.comb_onlyB) {
        cpl_imagelist_delete(comblist[4]) ;
        cpl_imagelist_delete(comblist[5]) ;
    }
    cpl_free(comblist) ;
    cpl_imagelist_delete(prof_list) ;
    cpl_imagelist_delete(bgmap_list) ;
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) 
        if (extr_tab[i] != NULL) cpl_table_delete(extr_tab[i]) ;
    cpl_msg_indent_less() ;

    /* Return */
    if (cpl_error_get_code()) return -1 ;
    else return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the products
  @param    images      6 image lists : combined, contrib, combined
            nodding A, contrib nodding A, combined nodding B, contrib nodding B
  @param    prof        the profile images
  @param    prof        the background maps images
  @param    extr_tab    the extracted tables
  @param    wl_map      the wavelength map
  @param    wl_map_model the wavelength map from the model
  @param    parlist     the parameters list
  @param    set         the frames set
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int crires_win_jitter_save(
        const cpl_imagelist     **  images,
        const cpl_imagelist     *   prof,
        const cpl_imagelist     *   bgmap,
        const cpl_table         **  extr_tab,
        const cpl_imagelist     *   wl_map,
        const cpl_imagelist     *   wl_map_model,
        const cpl_parameterlist *   parlist,
        cpl_frameset            *   set)
{
    cpl_propertylist    **  qclists ;
    const cpl_frame     *   ref_frame ;
    const char          *   procat ;
    const char          *   protype ;
    cpl_propertylist    *   inputlist ;
    const char          *   recipe_name = "crires_win_jitter" ;
    int                     i ;

    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW) ;

    /* Create the QC lists */
    qclists = cpl_malloc(CRIRES_NB_DETECTORS * sizeof(cpl_propertylist*)) ;
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        qclists[i] = cpl_propertylist_new() ;
        cpl_propertylist_append_int(qclists[i], "ESO QC SPECPOS",
                crires_win_jitter_config.qc_specpos[i]) ;
        cpl_propertylist_append_int(qclists[i], "ESO QC SPECWREC",
                crires_win_jitter_config.qc_specwrec[i]) ;
        cpl_propertylist_append_int(qclists[i], "ESO QC SPECWOPT",
                crires_win_jitter_config.qc_specwopt[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC SIGNAL MED",
                crires_win_jitter_config.qc_specoptmed[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC S2NMED",
                crires_win_jitter_config.qc_s2nmed[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC CENTWL",
                crires_win_jitter_config.qc_wlcent[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC DISPWL",
                crires_win_jitter_config.qc_wldisp[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC SENSMED",
                crires_win_jitter_config.qc_sensmed[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC CONVMED",
                crires_win_jitter_config.qc_convmed[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC THROMED",
                crires_win_jitter_config.qc_thromed[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC FWHMPIX COMBINED",
                crires_win_jitter_config.qc_fwhm_comb_pix[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC FWHMARC COMBINED",
                crires_win_jitter_config.qc_fwhm_comb_as[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC FWHMPIX PROFILE",
                crires_win_jitter_config.qc_fwhm_prof_pix[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC FWHMARC PROFILE",
                crires_win_jitter_config.qc_fwhm_prof_as[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC FWHM DIFF",
                crires_win_jitter_config.qc_fwhm_diff[i]) ;
        /* Propagate some keywords from input raw frame extensions */
        inputlist = cpl_propertylist_load_regexp(
                cpl_frame_get_filename(ref_frame), i+1,
                CRIRES_HEADER_EXT_FORWARD, 0) ;
        cpl_propertylist_copy_property_regexp(qclists[i], inputlist, 
                CRIRES_HEADER_EXT_FORWARD, 0) ;
        cpl_propertylist_delete(inputlist) ;
    }

    /* PRO.CATG */
    if (crires_win_jitter_config.std_mode == 1) {
        procat = CRIRES_STD_WIN_COMBINED_IMA ;
    } else {
        procat = CRIRES_OBS_WIN_COMBINED_IMA ;
    }

    /* Write the combined image */
    crires_image_save(set,
            parlist,
            set, 
            images[0], 
            recipe_name,
            procat, 
            CRIRES_PROTYPE_COMBINED,
            CRIRES_ILLUM_FULL_DETECTOR,
            NULL,
            (const cpl_propertylist **)qclists, 
            PACKAGE "/" PACKAGE_VERSION,
            "crires_win_jitter_comb.fits") ;

    /* PRO.CATG */
    if (crires_win_jitter_config.std_mode == 1) {
        procat = CRIRES_STD_WIN_CONTRIBUTION_IMA ;
    } else {
        procat = CRIRES_OBS_WIN_CONTRIBUTION_IMA ;
    }

    /* Write the contribution map */
    crires_image_save(set,
            parlist,
            set, 
            images[1], 
            recipe_name,
            procat, 
            CRIRES_PROTYPE_CONTRIB,
            CRIRES_ILLUM_FULL_DETECTOR,
            NULL,
            (const cpl_propertylist **)qclists, 
            PACKAGE "/" PACKAGE_VERSION,
            "crires_win_jitter_contrib.fits") ;

    /* Nodded A support */
    if (crires_win_jitter_config.comb_onlyA) {
        /* PRO.CATG */
        if (crires_win_jitter_config.std_mode == 1) {
            procat = CRIRES_STD_WIN_COMBINED_NA_IMA ;
        } else {
            procat = CRIRES_OBS_WIN_COMBINED_NA_IMA ;
        }

        /* Write the combined Nodded A image */
        crires_image_save(set,
                parlist,
                set,
                images[2],
                recipe_name,
                procat,
                CRIRES_PROTYPE_COMBINED,
                CRIRES_ILLUM_FULL_DETECTOR,
                NULL,
                (const cpl_propertylist **)qclists,
                PACKAGE "/" PACKAGE_VERSION,
                "crires_win_jitter_comb_noddedA.fits") ;

         /* PRO.CATG */
        if (crires_win_jitter_config.std_mode == 1) {
            procat = CRIRES_STD_WIN_CONTRIBUTION_NA_IMA ;
        } else {
            procat = CRIRES_OBS_WIN_CONTRIBUTION_NA_IMA ;
        }

        /* Write the contribution Nodded A image */
        crires_image_save(set,
                parlist,
                set,
                images[3],
                recipe_name,
                procat,
                CRIRES_PROTYPE_CONTRIB,
                CRIRES_ILLUM_FULL_DETECTOR,
                NULL,
                (const cpl_propertylist **)qclists,
                PACKAGE "/" PACKAGE_VERSION,
                "crires_win_jitter_contrib_noddedA.fits") ;
    }

    /* Nodded B support */
    if (crires_win_jitter_config.comb_onlyB) {
        /* PRO.CATG */
        if (crires_win_jitter_config.std_mode == 1) {
            procat = CRIRES_STD_WIN_COMBINED_NB_IMA ;
        } else {
            procat = CRIRES_OBS_WIN_COMBINED_NB_IMA ;
        }

        /* Write the combined Nodded B image */
        crires_image_save(set,
                parlist,
                set,
                images[4],
                recipe_name,
                procat,
                CRIRES_PROTYPE_COMBINED,
                CRIRES_ILLUM_FULL_DETECTOR,
                NULL,
                (const cpl_propertylist **)qclists,
                PACKAGE "/" PACKAGE_VERSION,
                "crires_win_jitter_comb_noddedB.fits") ;

         /* PRO.CATG */
        if (crires_win_jitter_config.std_mode == 1) {
            procat = CRIRES_STD_WIN_CONTRIBUTION_NB_IMA ;
        } else {
            procat = CRIRES_OBS_WIN_CONTRIBUTION_NB_IMA ;
        }

        /* Write the contribution Nodded B image */
        crires_image_save(set,
                parlist,
                set,
                images[5],
                recipe_name,
                procat,
                CRIRES_PROTYPE_CONTRIB,
                CRIRES_ILLUM_FULL_DETECTOR,
                NULL,
                (const cpl_propertylist **)qclists,
                PACKAGE "/" PACKAGE_VERSION,
                "crires_win_jitter_contrib_noddedB.fits") ;
    }

    /* PRO.CATG */
    if (crires_win_jitter_config.std_mode == 1) {
        procat = CRIRES_STD_WIN_EXTRACT_PROFILE_IMA ;
    } else {
        procat = CRIRES_OBS_WIN_EXTRACT_PROFILE_IMA ;
    }

    /* Write the profile image */
    crires_image_save(set,
            parlist,
            set, 
            prof, 
            recipe_name,
            procat, 
            CRIRES_PROTYPE_PROFILE,
            CRIRES_ILLUM_FULL_DETECTOR,
            NULL,
            (const cpl_propertylist **)qclists, 
            PACKAGE "/" PACKAGE_VERSION,
            "crires_win_jitter_prof.fits") ;
 
    /* PRO.CATG */
    if (crires_win_jitter_config.std_mode == 1) {
        procat = CRIRES_STD_WIN_EXTRACT_BGMAP_IMA ;
    } else {
        procat = CRIRES_OBS_WIN_EXTRACT_BGMAP_IMA ;
    }

    /* Write the background image */
    crires_image_save(set,
            parlist,
            set, 
            bgmap, 
            recipe_name,
            procat, 
            CRIRES_PROTYPE_BGD_MAP,
            CRIRES_ILLUM_FULL_DETECTOR,
            NULL,
            (const cpl_propertylist **)qclists, 
            PACKAGE "/" PACKAGE_VERSION,
            "crires_win_jitter_bgmap.fits") ;

    /* PRO.CATG */
    if (crires_win_jitter_config.std_mode == 1) {
        procat = CRIRES_STD_WIN_WL_MAP_IMA ;
    } else {
        procat = CRIRES_OBS_WIN_WL_MAP_IMA ;
    }
 
    /* Write the map */
    crires_image_save(set,
            parlist,
            set,
            wl_map,
            recipe_name,
            procat,
            CRIRES_PROTYPE_WL_MAP,
            CRIRES_ILLUM_FULL_DETECTOR,
            NULL,
            (const cpl_propertylist **)qclists,
            PACKAGE "/" PACKAGE_VERSION,
            "crires_win_jitter_wlmap.fits") ;

    if (wl_map_model != NULL) {
        /* PRO.CATG */
        if (crires_win_jitter_config.std_mode == 1) {
            procat = CRIRES_STD_WIN_WL_MAP_MODEL_IMA ;
        } else {
            procat = CRIRES_OBS_WIN_WL_MAP_MODEL_IMA ;
        }

        /* Write the model map */
        crires_image_save(set,
                parlist,
                set,
                wl_map_model,
                recipe_name,
                procat,
                CRIRES_PROTYPE_WL_MAP,
                CRIRES_ILLUM_FULL_DETECTOR,
                NULL,
                (const cpl_propertylist **)qclists,
                PACKAGE "/" PACKAGE_VERSION,
                "crires_win_jitter_wlmap_model.fits") ;
    }
    
    /* Write the extracted spectra */
    /* Get the PRO.CATG */
    if (cpl_table_has_column(extr_tab[0], CRIRES_COL_SENSITIVITY)) {
        procat = CRIRES_WIN_EXTRACT_SENS_TAB ;
        protype = CRIRES_PROTYPE_SENSIT ;
    } else if (cpl_table_has_column(extr_tab[0],CRIRES_COL_CONVERSION_OPT)){
        procat = CRIRES_WIN_EXTRACT_CONV_TAB ;
        protype = CRIRES_PROTYPE_CONVERS ;
    } else if (crires_win_jitter_config.std_mode == 1) {
        procat = CRIRES_STD_WIN_EXTRACT_WL_TAB ;
        protype = CRIRES_PROTYPE_SPEC_WL ;
    } else {
        procat = CRIRES_OBS_WIN_EXTRACT_WL_TAB ;
        protype = CRIRES_PROTYPE_SPEC_WL ;
    }
    crires_table_save(set, 
            parlist, 
            set, 
            extr_tab, 
            recipe_name, 
            procat, 
            protype,
            NULL, 
            (const cpl_propertylist **)qclists,
            PACKAGE "/" PACKAGE_VERSION,
            "crires_win_jitter_extracted.fits") ;

    /* Free and return */
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        cpl_propertylist_delete(qclists[i]) ;
    }
    cpl_free(qclists) ;
    return  0;
}

